# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import atexit
import weakref
import logging
import os
import re
import sys
import warnings
from abc import abstractmethod, ABCMeta
from collections import OrderedDict
from importlib import import_module

import numpy as np
import six as six

from braintech.utils.singleton_app import SingleApplicationInstance

LOGGER = logging.getLogger("BCIAppliance")


def check_blinker():
    if sys.platform.startswith('linux'):
        devices = os.listdir('/dev')
        for d in devices:
            if 'blinker' in d:
                return True
        return False
    return True


if not check_blinker():
    LOGGER.warning("""No blinker device is found! Maybe blinker udev rules are not yet installed?
You can install them by running:

        install_blinker

    """)


class FieldConfig(object):
    enabled = True
    brightness = 1
    buffer = [0, 1]
    frequency = 0
    phase = 0
    channel_indexes = []

    def __init__(self, number):
        self.number = number

    def __str__(self):
        if self.channel_indexes:
            ch = ", ch:" + ', '.join("%s" % c for c in self.channel_indexes)
        else:
            ch = ''
        return "%i: %s, LUM=%.2f, Hz=%.2f, φ=%.2f%s" % (self.number, 'ON' if self.enabled else 'OFF', self.brightness,
                                                        self.frequency, self.phase, ch)


class AbstractBlinker(six.with_metaclass(ABCMeta)):
    BUFFER_SIZE = 64  # type: int
    ROWS = 4
    COLUMNS = 4
    MAX_PHASE = 360.0

    """
    int: blinker device internal buffer size in samples. Each sample is one byte.
    """

    FIELDS_NUMBER = 16  # type: int
    """
    int: number of independent, generated, diode driving channels that can be controlled from blinker device
    """

    def __init__(self, port, preview=False, reset=True):
        """
        ``port`` is a device name: depending on operating system. e.g. ``/dev/ttyUSB0``
        on Linux or ``COM3`` on Windows.

        NOTE: this class uses frequencies in HZ (Hz) units

        :param port: serial port device name
        :param preview: display a qt-based blinker emulator
        :param reset: Set all fields to maximum brightness and enabled, blinking stopped
        :raises serial.SerialException: In case the device can not be found or can not be configured.
        """
        super(AbstractBlinker, self).__init__()
        self._fields = [FieldConfig(i) for i in range(self.FIELDS_NUMBER)]
        self._device = None
        if preview:
            self._preview = self._create_preview()
        else:
            self._preview = None
        if port:
            self._serial_lock = SingleApplicationInstance(flavor_id=port, basename='obci')
        self._init_device(port)
        if reset:
            self.reset()
        weak_close = weakref.WeakMethod(self.close)

        def close_at_exit():
            close = weak_close()
            if close is not None:
                close()
        atexit.register(close_at_exit)

    @abstractmethod
    def _create_preview(self):
        pass

    @abstractmethod
    def _init_device(self, port):
        pass

    def set_rectangular_function(self, field_number, frequency, duty_cycle, phase_offset=0, luminosity=1):
        """
        Set specified channel to repeat rectangle signal with ``frequency`` specified in Hz (HZ).

        ``duty_cycle`` specifies how many samples are filled with ``0xFF``, remaining
        samples are filled with `0x00`.

        ``phase_offset`` shifts whole signal circularly by specified number of angles.

        :param field_number: unsigned integer
        :param frequency: float, must be in range [0.0, 6553.5] (in Hz units)
        :param buffer: float, exactly ``self.BUFFER_SIZE`` bytes with samples value string of bytes
        :param duty_cycle: float, must be in range [0.0, 1.0],
        :param phase_offset:float, signal phase offset in angles 0 to 360
        :param luminosity: float, must be in range [0 to 1], 0: the darkest, 1: the brightest
        """
        duty_cycle = int(duty_cycle * self.BUFFER_SIZE)
        buffer = [1.0 if i < duty_cycle else 0 for i in range(self.BUFFER_SIZE)]
        self._update_field(field_number=field_number,
                           buffer=buffer,
                           frequency=frequency,
                           phase_offset=phase_offset,
                           luminosity=luminosity
                           )

    def set_sine_function(self, field_number, frequency, phase_offset=0, luminosity=1):
        """
        Set specified channel to repeat sine-filled buffer with ``frequency`` specified in Hz (HZ).

        :param field_number: 8-bit unsigned integer
        :param frequency: float, must be in range [0.0, 6553.5] (in Hz units)
        :param phase_offset:float, signal phase offset in angles 0 to 360
        :param luminosity: float, must be in range [0 to 1], 0: the darkest, 1: the brightest
        """
        buffer = (np.sin(np.linspace(0, 2 * np.pi, self.BUFFER_SIZE)) + 1) / 2.
        self._update_field(field_number=field_number,
                           buffer=buffer,
                           frequency=frequency,
                           phase_offset=phase_offset,
                           luminosity=luminosity
                           )

    def set_buffer(self, field_number, frequency, buffer, phase_offset=0, luminosity=1):
        """
        Specify ``buffer`` with sample values to be repeatedly played on specified channel.

        Whole buffer will be repeated with ``frequency`` specified in Hz (HZ).

        :param field_number: 8-bit unsigned integer
        :param frequency: float, must be in range [0.0, 6553.5] (in Hz units)
        :param buffer: float, exactly ``self.BUFFER_SIZE`` bytes with samples value string of bytes
        :param phase_offset:float, signal phase offset in angles 0 to 360
        :param luminosity: float, must be in range [0 to 1], 0: the darkest, 1: the brightest
        """
        assert isinstance(buffer, bytes)
        assert len(buffer) == self.BUFFER_SIZE
        self._update_field(field_number=field_number,
                           buffer=buffer,
                           frequency=frequency,
                           phase_offset=phase_offset,
                           luminosity=luminosity
                           )

    def set_still(self, field_number, luminosity):
        """
        Disable channel specified by ``field_number``.

        Disabled channels do not blink and are constantly lit by background diodes with specified ``luminosity``.

        :param field_number: 8-bit unsigned integer
        :param luminosity: float, must be in range [0 to 1], 0: the darkest, 1: the brightest
        """
        self._update_field(field_number=field_number,
                           buffer=np.ones(self.BUFFER_SIZE, float),
                           frequency=1,
                           luminosity=luminosity
                           )

    def set_luminosity(self, field_number, luminosity):
        """
        Change luminosity of a single channel. Default luminosity is 1.

        :param field_number: 8-bit unsigned integer
        :param luminosity: float, must be in range [0 to 1], 0: the darkest, 1: the brightest
        """
        self._update_field(field_number=field_number,
                           luminosity=luminosity
                           )

    def set_phase_offset(self, field_number, phase_offset):
        """
        Change luminosity of a single channel. Default luminosity is 1.

        :param field_number: 8-bit unsigned integer
        :param phase_offset:float, signal phase offset in angles 0 to 360
        """
        self._update_field(field_number=field_number,
                           phase_offset=phase_offset
                           )

    def set_frequency(self, field_number, frequency, offset=None):
        """
        Change luminosity of a single channel. Default luminosity is 1.

        :param field_number: 8-bit unsigned integer
        :param frequency: float, must be in range [0.0, 6553.5] (in Hz units)
        :param offset: float, as in set_phase_offset - signal phase offset in angles 0 to 360
        """
        self._update_field(field_number=field_number,
                           frequency=frequency,
                           phase_offset=offset
                           )

    @abstractmethod
    def _update_field(self, field_number, buffer=None, frequency=None, phase_offset=None, luminosity=None):
        pass

    # blinkers which dont support synchro will just do nothing
    def set_synchro_field(self, field: int, flush=True):
        pass

    @abstractmethod
    def start(self):
        """Starts blinking. Disabled diodes are unaffected."""
        pass

    @abstractmethod
    def stop(self):
        """Stops blinking. Disabled diodes are unaffected."""
        pass

    @abstractmethod
    def enable(self, field_number):
        """
        Enable channel specified by ``field_number``.

        Enabled channels blink periodically with specified pattern.

        :param field_number: 8-bit unsigned integer
        """
        self._fields[field_number].enabled = True

    @abstractmethod
    def disable(self, field_number, luminosity):
        """
        Disable channel specified by ``field_number``.

        Disabled channels do not blink and are constantly lit by background diodes with specified ``luminosity``.

        :param field_number: 8-bit unsigned integer
        :param luminosity: float, must be in range [0 to 1], 0: the darkest, 1: the brightest
        """
        self._fields[field_number].enabled = False
        self._fields[field_number].brightness = luminosity

    def reset(self):
        """Disables blinking and sets all channels to maximum brightness."""
        for field_number in range(self.fields_num):
            self.disable(field_number, 1.0)
        self.stop()
        for field_number in range(self.fields_num):
            self.enable(field_number)

    clear = reset  # backward compatibility with psychopy

    def close(self):
        if self._device:
            self.reset()

    @property
    def fields_num(self):
        return len(self._fields)

    # noinspection PyPep8Naming
    @property
    def CHANNELS_NUMBER(self):
        warnings.warn("CHANNELS_NUMBER is deprecated. Use `fields_num` instead.", DeprecationWarning)
        return self.fields_num


class WrongDeviceNameException(Exception):
    pass


_KNOWN_DEVICES = OrderedDict((
    (r'fat_blinker(\d+)?', ('..fat_blinker', 'FatBlinker')),
    (r'slim_blinker(\d+)?', ('..slim_blinker', 'SlimBlinker')),
    (r'fuw_blinker(\d+)?', ('..fuw_blinker', 'FUWBlinker')),
    (r'slim16_blinker(\d+)?', ('..slim_blinker', 'Slim16Blinker')),
    (r'fat16_blinker(\d+)?', ('..fat_blinker', 'Fat16Blinker')),
    # we want to check ttyUSB* last
    (r'ttyUSB(\d+)?', ('..fuw_blinker', 'FUWBlinker')),
))


def get_blinker(port=None, preview=False):
    """
    Create blinker which is designed to work with port
    :param port: blinker name
    :param preview: show preview or not
    :return: Blinker instance
    """

    if not port or port == 'auto':
        device = _find_any_device(preview)
        if device:
            return device
    else:
        for name, (module_name, class_name) in _KNOWN_DEVICES.items():
            match = re.search(name + '$', port)
            if match:
                if "dummy" in port:
                    port = None
                elif sys.platform.startswith('win'):
                    try:
                        port = "COM" + match.group(1)
                    except TypeError:
                        raise WrongDeviceNameException("You have to provide COM number in device name!")
                klass = _get_class(module_name, class_name)
                return klass(port, preview)
    raise WrongDeviceNameException("Unknown blinker device: %s!" % port)


def _get_class(module_name, class_name):
    module = import_module(module_name, __name__)
    return getattr(module, class_name)


def _find_any_device(preview=True):
    if sys.platform.startswith('win'):
        raise WrongDeviceNameException("You have to provide COM number on Windows!")
    devices = os.listdir('/dev')
    # we want to check ttyUSB* last
    for name, (module_name, class_name) in _KNOWN_DEVICES.items():
        for dev_name in devices:
            if re.match(name, dev_name):
                klass = _get_class(module_name, class_name)
                return klass(os.path.join('/dev', dev_name), preview)
    if preview:
        return _get_class('..fat_blinker', 'Fat16Blinker')(None, preview)
    else:
        return


Blinker = get_blinker  # Backward compatibility
CHANNELS_NUMBER = 16  # Backward compatibility


def run_demo(blinker, sleep):
    blinker.stop()
    import time
    for i in range(blinker.FIELDS_NUMBER):
        blinker.enable(i)

    # parameter vectors for programming function Hz

    luminosity = [1.0] * blinker.FIELDS_NUMBER

    freqlist = [0.5] * blinker.FIELDS_NUMBER
    step = int(blinker.MAX_PHASE / blinker.FIELDS_NUMBER)
    k = step * blinker.COLUMNS
    phase_offset = []
    for i in range(blinker.ROWS):
        row = range(i * k, (i + 1) * k, step)
        if i % 2 == 1:
            row = reversed(row)
        phase_offset.extend(row)
    duty_cycle = 0.7

    # programming function(sine wave)
    for i in range(blinker.FIELDS_NUMBER):
        blinker.set_sine_function(i, freqlist[i], phase_offset[i], luminosity[i])
    # begin the demo blurred-snake and stop it after 5.5 sec
    print("Blurred snake...")
    blinker.start()
    time.sleep(sleep)  # required
    blinker.stop()
    print("Blurred stop")

    # begin the demo sharp-snake and stop it after 5.5 sec

    for i in range(blinker.FIELDS_NUMBER):
        blinker.set_rectangular_function(i, freqlist[i], duty_cycle, phase_offset[i], luminosity[i])
    print("Snake ..")
    blinker.start()
    time.sleep(sleep)  # required
    print("Snake stop")
    blinker.stop()
