# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import os
import pickle
import signal
import subprocess
import sys

import PySide2
from PySide2.QtCore import QThread, Signal, QTimer, Slot
from PySide2.QtGui import QPainter, QPen, QColor
from PySide2.QtWidgets import QApplication, QWidget, QLabel, QGridLayout, QVBoxLayout, QFrame, QSizePolicy

from braintech.drivers.blinker import blinker

__all__ = 'Emulator'

PYTHON_3 = sys.version_info >= (3, 0)
PY2_PY3_COMPATIBLE_PROTOCOL = 2

if PYTHON_3:
    try:
        stdin = sys.stdin.buffer
    except AttributeError:
        stdin = sys.stdin
else:
    stdin = sys.stdin


class BCIState(object):
    """AbstractBlinker state, shared by threads."""

    def __init__(self, rows, columns):
        self.blinking = False
        self.set_fields(rows * columns)
        self.rows = rows
        self.columns = columns
        self.no = 0  # increased on state change
        self.intensity = 100

    def set_fields(self, fields_num):
        self.fields = [blinker.FieldConfig(i) for i in range(fields_num)]


class SimpleLineChart(QWidget):
    """A simple qt bar/line chart."""

    def __init__(self, series, parent=None):
        super(SimpleLineChart, self).__init__(parent)
        self._series = series

    def paintEvent(self, event):
        """Paint the chart."""
        if not self._series:
            return
        painter = QPainter()
        painter.begin(self)
        width = self.geometry().width()
        height = self.geometry().height()
        delta_x = width / len(self._series)
        painter.setPen(QPen(QColor(0, 0, 255), 2))
        x = 0
        last_y = None
        for value in self._series:
            y = height * (1 - value / 256.0)
            if last_y:
                painter.drawLine(x, y, x, last_y)  # connect lines
            painter.drawLine(x, y, x + delta_x, y)  # chart the value
            x += delta_x
            last_y = y
        painter.end()

    def update(self, series):
        """Update the chart series and repaint the chart."""
        self._series = series
        super(SimpleLineChart, self).update()


class _BlinkFieldWidget(QWidget):
    def __init__(self, number):
        self.number = number
        super(_BlinkFieldWidget, self).__init__()
        box = QVBoxLayout()
        self.label = QLabel(str(self.number))
        self.label.setWordWrap(True)
        self.label.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        box.addWidget(self.label)
        self.chart = SimpleLineChart([])
        box.addWidget(self.chart)
        box.setStretch(0, 0)
        box.setStretch(1, 1)
        self.setLayout(box)

    def set_config(self, config):
        self.config = c = config
        self.label.setText(str(config))
        self.label.setFrameShape(QFrame.Box if c.enabled else QFrame.NoFrame)
        self.chart.update(c.buffer)


class QtBCI(QWidget):
    """4x4 window showing BCI state."""
    COLUMN_WIDTH = 150
    ROW_HEIGHT = 150
    TOP = 20

    def __init__(self, state):
        """
        :param state: shared BCI state to show.
        """
        super(QtBCI, self).__init__()
        self.title = 'BCI'
        self.state = state
        self.last_state_no = None
        self.init_UI(self.state.rows, self.state.columns)
        self.updateTimer = QTimer()
        self.updateTimer.setSingleShot(True)
        self.updateTimer.setInterval(50)
        self.updateTimer.timeout.connect(self.update)

    def init_UI(self, rows, columns):
        """Build the window."""
        self.setWindowTitle(self.title)
        screen_size = QApplication.instance().desktop().screenGeometry()
        self.setGeometry(0, self.TOP,
                         min(self.COLUMN_WIDTH * columns, screen_size.width()),
                         min(self.ROW_HEIGHT * rows, screen_size.height()))
        self.grid = QGridLayout()
        self.fields = []
        self.setLayout(self.grid)
        self.init_fields(columns, rows * columns)
        self.show()

        self.setLayout(self.grid)
        self.show()

    def init_fields(self, columns, fields_num):
        for f in self.fields:
            self.grid.removeWidget(f)
            f.deleteLater()
        self.fields = []
        for i in range(fields_num):
            field = _BlinkFieldWidget(i)
            self.fields.append(field)
            self.grid.addWidget(field, i / columns, i % columns)

    @Slot(object)
    def show_dev_state(self, state):
        """
        Update the window new BCI state.
        :param channel: optional, channel # to update
        """
        self.state = state
        if not self.updateTimer.isActive():
            self.updateTimer.start()

    def update(self):
        s = self.state
        if len(s.fields) != len(self.fields):
            self.init_fields(s.columns, len(s.fields))
        for i, f in enumerate(s.fields):
            self.fields[i].set_config(f)
        self.setWindowTitle('Blinker {enabled}, state #{state}, Intensity: {intensity:.0f}%'.format(
            enabled='enabled' if s.blinking else 'disabled',
            state=s.no,
            intensity=s.intensity))


class Emulator(object):
    """QT-based AbstractBlinker emulator."""

    def __init__(self, rows, columns):
        self.s = BCIState(rows, columns)
        self.proc = subprocess.Popen([sys.executable, __file__, 'qt', str(rows), str(columns)], stdin=subprocess.PIPE,
                                     bufsize=0, executable=sys.executable)

    def _send_state(self):
        try:
            self.s.no += 1
            pickle.dump(self.s, self.proc.stdin, protocol=PY2_PY3_COMPATIBLE_PROTOCOL)
        except IOError:
            pass

    def close(self):
        self.proc.terminate()


class _ReceiveThread(QThread):
    new_data = Signal(object)

    def __init__(self, app, gui):
        self.app = app
        self.gui = gui
        super(_ReceiveThread, self).__init__(app)
        self.new_data.connect(self._update_state)

    @Slot(object)
    def _update_state(self, new_state):
        # invoked in main thread
        if new_state is False:
            self.app.quit()
        else:
            self.gui.show_dev_state(new_state)

    def run(self):
        while True:
            try:
                state = pickle.load(stdin)
                self.new_data.emit(state)
            except (EOFError, ValueError):
                self.quit()
                return

    def quit(self, *args, **kwargs):
        self.app.quit()


if __name__ == '__main__':
    if sys.platform.startswith('win'):
        os.environ['QT_QPA_PLATFORM_PLUGIN_PATH'] = os.path.dirname(PySide2.__file__) + '/plugins/platforms'

    if 'qt' in sys.argv:
        app = QApplication([])
        u = QtBCI(BCIState(int(sys.argv[-2]), int(sys.argv[-1])))
        u.show()
        th = _ReceiveThread(app, u)
        th.start()
        signal.signal(signal.SIGINT, th.quit)
        signal.signal(signal.SIGTERM, th.quit)
        app.exec_()
        stdin.close()
    else:
        from braintech.drivers.blinker.fuw_blinker import FUWBlinker

        blinker.run_demo(FUWBlinker(None, True), 10)  # non0 sleep to avoid occasional ?qt segfault
