# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import ctypes
import sys
import time

from braintech.drivers.blinker.commands import MessageType, BlinkerState, Command, receivable
from braintech.drivers.blinker.slim_blinker import SlimBlinker, SlimBlinkerDevice, Blinker16Mixin, Segment


@receivable
class TemperatureRequestCommand(Command):
    TYPE = MessageType.TEMP_REQ
    _fields_ = [
        ('temp1', ctypes.c_float),
        ('temp2', ctypes.c_float),
    ]


class FatBlinkerDevice(SlimBlinkerDevice):
    def get_temperatures(self):
        command = self._prepare_bytes(TemperatureRequestCommand())
        self.write(command)
        resp = self._get_response()
        assert isinstance(resp, TemperatureRequestCommand)
        return resp.temp1, resp.temp2


def _calculate_fat_blinker_segments(screen_width, screen_height):
    segments = []
    rows = 4
    columns = 10
    diodes_per_segment = 7
    segments_per_field = 8
    diode_width = screen_width / (columns * diodes_per_segment)
    diode_height = screen_height / (rows * diodes_per_segment)
    for r in range(rows):
        for c in range(columns):
            for i in range(diodes_per_segment):
                segments.append(
                    Segment((r * columns + c) * segments_per_field + i,
                            c * (diode_width * diodes_per_segment),
                            (r * diodes_per_segment + i) * diode_height,
                            diode_width * (diodes_per_segment - 1),
                            diode_height))
            segments.append(Segment(
                (r * columns + c) * segments_per_field + diodes_per_segment,
                ((c + 1) * diodes_per_segment - 1) * diode_width,
                (r * diodes_per_segment) * diode_height,
                diode_width,
                diode_height * (diodes_per_segment - 1)
            ))
    return segments


class FatBlinker(SlimBlinker):
    DEVICE_CLASS = FatBlinkerDevice
    SEGMENTS = _calculate_fat_blinker_segments(*SlimBlinker.RESOLUTION)
    SEGMENT_COUNT = len(SEGMENTS)
    SEGMENTS_PER_FIELD = 8
    MAX_INTENSITY = 60
    FIRMWARE_INITIAL_INTENSITY = 30

    def __init__(self, *args, **kwargs):
        super(FatBlinker, self).__init__(*args, **kwargs)
        self.set_intensity(self.FIRMWARE_INITIAL_INTENSITY / self.MAX_INTENSITY)

    def get_temperatures(self):
        return self._device.get_temperatures()

    def close(self):
        if self._device:
            self.set_intensity(self.FIRMWARE_INITIAL_INTENSITY / self.MAX_INTENSITY)
        super().close()


class Fat16Blinker(Blinker16Mixin, FatBlinker):
    FIELDS_NUMBER = 16
    ROWS = 4
    COLUMNS = 4


if __name__ == '__main__':
    port = sys.argv[1] if len(sys.argv) > 1 else None
    blinker = FatBlinker(port, preview=port is None, reset=True)
    blinker.stop()
    if blinker.version >= 4:
        print(blinker.get_temperatures())
    for i in range(blinker.fields_num):
        blinker.disable(i, i / blinker.fields_num)
    blinker.flush()
    time.sleep(5)
    if blinker.version > 4:
        blinker.set_state(BlinkerState.OFF)
        time.sleep(1)
        blinker.set_state(BlinkerState.DEMO)
        time.sleep(1)
        blinker.set_state(BlinkerState.ON)
        time.sleep(1)
        blinker.set_intensity(0.5)
        time.sleep(1)
    blinker.set_number_of_fields(blinker.SEGMENT_COUNT)
    for i in range(blinker.SEGMENT_COUNT):
        blinker.define_field(i, [i])
        blinker.set_frequency(i, 0.1, i * blinker.MAX_PHASE / blinker.SEGMENT_COUNT)
        blinker.enable(i)
    blinker.start()
    blinker.flush()
    print(blinker.version)
    for i in range(3):
        if blinker.version >= 4:
            print(blinker.get_temperatures())
        time.sleep(1)
    for i in range(blinker.fields_num):
        blinker.disable(i, 0)
    blinker.flush()
    blinker.reset()
    from braintech.drivers.blinker.blinker import run_demo
    run_demo(blinker, 5)
    del blinker
