# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.


import struct
import sys

import numpy as np
import serial

from braintech.drivers.blinker.blinker import AbstractBlinker, run_demo
from braintech.drivers.blinker.blinker_preview import Emulator


class ResponseError(Exception):
    """Raised when device hadn't answered with ``OK`` message."""


class _CMD:
    """Possible blinker's commands."""
    UART_BAUD_RATE = 57600
    START_FRAME = 0x02
    END_FRAME = 0x01
    WRITE = 0x11
    START = 0x12
    STOP = 0x13
    ENABLE = 0x14
    DISABLE = 0x15
    FIRMWARE = 0xFF


class EmptyBlinkerSerial(object):
    """Useful when no hardware is attached."""

    def __init__(self):
        self.is_open = True

    def write(self, data):
        """Send cmd to the blinker."""
        pass

    def flush(self):
        """Make sure that the blinker got the cmd."""
        pass

    def readline(self):
        """Get the ACK from blinker."""
        return b'OK'


class FUWEmulator(Emulator):
    def write(self, data):
        """Emulate the blinker / update the virtual blinker state."""
        data = struct.unpack(len(data) * 'B', data)
        cmd = data[3]
        if cmd == _CMD.START:
            self.s.blinking = True
            assert (data[4] == 1)
        elif cmd == _CMD.STOP:
            self.s.blinking = False
            assert (data[4] == 1)
        else:
            field_config = self.s.fields[data[4]]
            if cmd == _CMD.ENABLE:
                field_config.enabled = True
                assert (data[5] == 1)
            elif cmd == _CMD.DISABLE:
                field_config.brightness = float(data[5] / 255.0)
                field_config.enabled = False
                assert (data[6] == 1)
            elif cmd == _CMD.WRITE:
                freq = data[5] * 256 + data[6]
                field_config.frequency = freq / 10.0
                field_config.buffer = list(data[7:-1])
                assert (data[-1] == 1)
            else:
                raise ValueError
        self._send_state()

    def set_phase(self, channel, phase):
        """Keep the user-api supplied phase for channel."""
        self.s.fields[channel].phase = phase
        self._send_state()

    def set_brightness(self, channel, brightness):
        """Keep the user-api supplied brightness for channel."""
        self.s.fields[channel].brightness = brightness
        self._send_state()


class FUWBlinker(AbstractBlinker):
    BUFFER_SIZE = 64  # type: int
    ROWS = 4
    COLUMNS = 4

    """
    int: blinker device internal buffer size in samples. Each sample is one byte.
    """

    FIELDS_NUMBER = 16  # type: int
    """
    int: number of independent, generated, diode driving channels that can be controlled from blinker device
    """

    FULL_ANGLE = 360.0  # type: float

    def _create_preview(self):
        return FUWEmulator(self.ROWS, self.COLUMNS)

    def _init_device(self, port):
        if port:
            self._device = serial.Serial(
                port=port,
                baudrate=57600,
                bytesize=serial.EIGHTBITS,
                parity=serial.PARITY_NONE,
                stopbits=serial.STOPBITS_TWO,
                timeout=1,
                exclusive=True
            )
        else:
            self._device = EmptyBlinkerSerial()
        assert self._device.is_open

    def start(self):
        """Starts blinking. Disabled diodes are unaffected."""
        self._send_cmd(_CMD.START)  # start CMD

    def stop(self):
        """Stops blinking. Disabled diodes are unaffected."""
        self._send_cmd(_CMD.STOP)  # stop CMD

    def enable(self, field_number):
        super(FUWBlinker, self).enable(field_number)
        field_number = self._prepare_uint8(field_number, 'field_number')
        self._send_cmd(
            _CMD.ENABLE,  # enable CMD
            struct.pack("B", field_number)
        )

    def disable(self, field_number, brightness):
        """
        Disable channel specified by ``field_number``.

        Disabled channels do not blink and are constantly lit by background diodes with specified ``brightness``.

        :param field_number: 8-bit unsigned integer
        :param brightness: float, must be in range [0 to 1], 0: the darkest, 1: the brightest
        """
        super(FUWBlinker, self).disable(field_number, brightness)
        field_number = self._prepare_uint8(field_number, 'field_number')
        brightness = self._prepare_uint8(brightness * 255, 'brightness')
        self._send_cmd(
            _CMD.DISABLE,  # disable CMD
            struct.pack("BB", field_number, brightness)
        )

    def _update_field(self, field_number, buffer=None, frequency=None, phase_offset=None, luminosity=None):
        field_number = self._prepare_uint8(field_number, 'field_number')
        channel = self._fields[field_number]
        if phase_offset is not None:
            channel.phase = phase_offset
            if self._preview:
                self._preview.set_phase(field_number, phase_offset)
        if frequency is not None:
            channel.frequency = frequency
        if luminosity is not None:
            channel.brightness = luminosity
            if self._preview:
                self._preview.set_brightness(field_number, luminosity)
        if buffer is not None:
            channel.buffer = np.array(buffer)
        phase_offset = self._prepare_uint16(int(channel.phase / self.FULL_ANGLE * self.BUFFER_SIZE), "phase_offset")
        frequency = self._prepare_uint16(int(channel.frequency * 10.0), 'frequency')
        tmp_buffer = np.roll(channel.buffer * channel.brightness * 0xFF, -phase_offset).astype(dtype=np.uint8)
        _buffer = tmp_buffer.tobytes()

        self._send_cmd(
            _CMD.WRITE,  # write CMD
            struct.pack(">BH", field_number, frequency) + _buffer
        )

    def _send_cmd(self, cmd_code, extra_data=None):
        """
        Send ``cmd_code`` command to this device. Some commands require nonempty ``extra_data`` parameter.

        :param cmd_code: command number
        :param extra_data: if not ``None``
        :raises ResponseError: raised if device response was not ``OK``
        """
        assert extra_data is None or isinstance(extra_data, bytes)

        cmd_code = self._prepare_uint8(cmd_code, 'cmd_code')
        total_length = self._prepare_uint16(5 + (0 if extra_data is None else len(extra_data)),
                                            'Total data length')

        data = struct.pack(">BHB",
                           _CMD.START_FRAME,  # SOF (Start Of Frame)
                           total_length,
                           cmd_code  # CMD code
                           )
        if extra_data is not None:
            data += extra_data

        data += struct.pack("B", _CMD.END_FRAME)  # EOF
        if self._preview:
            self._preview.write(data)
        self._device.write(data)
        self._device.flush()

        response = self._device.readline().strip()

        if response != b'OK':
            raise ResponseError("Response was not 'OK' (response: {}).".format(repr(response)))

    @staticmethod
    def _prepare_uint8(value, name=None):
        """
        Convert ``value`` to 8-bit unsigned integer.

        :param value: number to convert
        :param name: value name to include in error message
        :return: 8-bit unsigned integer
        :raises ValueError: if ``value`` cannot be converted to 8-bit integer or ``value`` is
                            outside of 8-bit unsigned integer range
        """
        value = int(value)
        if value < 0 or value > 255:
            raise ValueError('{} must be in range [0, 255]'
                             .format('Value' if name is None else name))
        return value

    @staticmethod
    def _prepare_uint16(value, name=None):
        """
        Convert ``value`` to 16-bit unsigned integer.

        :param value: number to convert
        :param name: value name to include in error message
        :return: 16-bit unsigned integer
        :raises ValueError: if ``value`` cannot be converted to 16-bit integer or ``value`` is
                            outside of 16-bit unsigned integer range
        """
        value = int(value)
        if value < 0 or value > 65535:
            raise ValueError('{} must be in range [0.0, 6553.5]'
                             .format('Value' if name is None else name))
        return value

    def reset(self):
        for field_number in range(self.FIELDS_NUMBER):
            # Disable command is not working as expected
            self.set_still(field_number, 1.0)
        self.start()
        self.stop()
        for field_number in range(self.FIELDS_NUMBER):
            self.enable(field_number)


if __name__ == '__main__':
    port = sys.argv[1] if len(sys.argv) > 1 else None
    run_demo(FUWBlinker(port, port is None), 5)
