# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

"""Module provides single :class 'BCIAppliance'."""
from braintech.obci.experiment.peer.configured_peer import ConfiguredPeer
from braintech.drivers.blinker.blinker import get_blinker
from braintech.obci.core.broker import messages
from braintech.obci.core.broker.message_handler_mixin import register_message_handler
from braintech.obci.core.broker.messages import Field, BaseMessage
from braintech.obci.core.utils.properties import param_property

VALID_KEYS = {'action', 'field_number', 'frequency', 'phase_offset', 'luminosity', 'duty_cycle'}


class FieldsField(Field):
    """Helper class for :class 'OBCIApplianceConfigurationMsg'."""

    def __init__(self):
        """Validate dictionary from :class 'OBCIApplianceConfigurationMsg'."""
        super(FieldsField, self).__init__(dict)

    def validate(self, value):
        """Validation method of dictionary, checks keys names."""
        super(FieldsField, self).validate(value)
        for channel in value:
            if not (set(value[channel].keys()) <= VALID_KEYS):
                raise TypeError("Field keys '{}' not in valid keys ({})"
                                .format(",".join(value[channel].keys()),
                                        ",".join(VALID_KEYS)))


class OBCIApplianceConfigurationMsg(BaseMessage):
    """OBCI Appliance configuration Message."""

    fields_config = FieldsField()


__all__ = ('BCIAppliance',)


class BCIAppliance(ConfiguredPeer):
    """Class provides methods for direct control of BCIAppliance."""

    async def _connections_established(self):
        await super()._connections_established()
        self.blinker = get_blinker(self.port, self.preview)

    port = param_property('port')
    preview = param_property('preview', deserializer=lambda s: bool(int(s)))

    async def _start(self):
        await super()._start()
        self.blinker.start()

    async def _stop(self):
        await super()._stop()
        self.blinker.stop()

    @register_message_handler(OBCIApplianceConfigurationMsg)
    async def _handle_configuration_msg(self, msg):
        """
        Message data contains specified parameters.

        function, field_number, frequency, phase_offset, luminosity, duty_cycle
        :return OkMsg
        """
        for _, config in msg.fields_config.items():
            if config['action'] == 'sine':
                self.blinker.set_sine_function(field_number=config['field_number'],
                                               frequency=config['frequency'],
                                               phase_offset=config['phase_offset'],
                                               luminosity=config['luminosity']
                                               )
            elif config['action'] == 'rect':
                self.blinker.set_rectangular_function(field_number=config['field_number'],
                                                      frequency=config['frequency'],
                                                      phase_offset=config['phase_offset'],
                                                      luminosity=config['luminosity'],
                                                      duty_cycle=config['duty_cycle']
                                                      )
            elif config['action'] == 'still':
                self.blinker.set_still(field_number=config['field_number'], luminosity=config['luminosity'])

        return messages.OkMsg()

    def _cleanup(self):
        self.blinker.reset()
        super()._cleanup()
