# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

"""Package providing small signal verifier with sample_counter and saw."""
import numpy
from braintech.obci.core.broker import messages

from braintech.obci.core.broker.peer import Peer
from braintech.obci.core.drivers.eeg.eeg_amplifier import SamplePacket
from braintech.obci.core.utils.signal_generators import saw_generator

__all__ = ('DummySignalReceiverPeer',)


class DummySignalReceiverPeer(Peer):
    """Receives and verifies signal generated by `DummyAmplifierPeer` and send mean signal across channels with saw.

    Signal receiver peer receives signal generated by `DummyAmplifierPeer`,
    verifies its correctness and sends another signal consisting of received
    mean signal across channels and saw.
    """

    def __init__(self, *args, generator_peer_name='Generator', **kwargs):
        """
        Class which verifies signal and add saw signal with sample_counter.

        :param args: args for Peer class
        :param generator_peer_name: To whom listen for signal
        :param kwargs: kwargs for Peer class
        """
        super().__init__(*args, **kwargs)
        self._saw_gen = saw_generator()
        self._samples_counter = 0
        self._generator_peer_name = generator_peer_name

    async def _connections_established(self):
        await super()._connections_established()
        self.subscribe_for_specific_msg_subtype(messages.SignalMessage,
                                                self._generator_peer_name, self.handle_signal_message)

    async def handle_signal_message(self, msg):
        """
        Coroutine which works forever: accepting signal message and adds two channels: saw and samples_counter.

        :param msg: Message with signal
        :return: mean signal across channels with additional channels: saw and samples_counter.
        """
        input = msg.data  # SamplePacket
        sample_count = len(input.ts)
        samples = numpy.zeros([sample_count, 3])
        samples[:, 1] = numpy.mean(input.samples, axis=1)
        for i in range(sample_count):
            samples[i, 0] = self._samples_counter
            samples[i, 2] = next(self._saw_gen)
            self._samples_counter += 1
        return messages.SignalMessage(sender=self.id, data=SamplePacket(ts=input.ts, samples=samples))
