# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

import collections
import enum

import dlib
import numpy as np

from . import camera
from . import display
from . import eyes
from . import face
from . import head


class HeadState(enum.Enum):
    MOVING = 'moving'
    NOT_MOVING = 'not_moving'
    NONE_PRESENT = 'none_present'


class EyesState(enum.Enum):
    OPEN = 'open'
    CLOSED = 'closed'
    NONE_PRESENT = 'none_present'


State = collections.namedtuple('State', ('head', 'eyes'))


def get_current_state_from_camera(angle_difference_for_movement,
                                  display_preview_window,
                                  eyes_open_borderline_ratio):
    if display_preview_window:
        window = dlib.image_window()
    faces_detector = face.FacesDetector()
    face_shape_predictor = face.FaceShapePredictor()
    movement_detector = MovementDetector(angle_difference_for_movement)
    for image in camera.images_generator():
        faces = faces_detector.detect(image)
        if faces:
            dominant_face = face.pick_closest_to_center(faces)

            face_shape = face_shape_predictor.predict(image, dominant_face)
            if eyes.are_open(face_shape, eyes_open_borderline_ratio):
                eyes_state = EyesState.OPEN
            else:
                eyes_state = EyesState.CLOSED
            pose = head.calculate_pose(face_shape)
            euler_angle = head.calculate_euler_angle(pose)

            if display_preview_window:
                image = display.annotate(image, face_shape, pose, euler_angle)
                window.set_image(image)

            movement_detector.add_angle(euler_angle)
            if movement_detector.has_moved():
                yield State(eyes=eyes_state, head=HeadState.MOVING)
            else:
                yield State(eyes=eyes_state, head=HeadState.NOT_MOVING)
        else:
            yield State(eyes=EyesState.NONE_PRESENT, head=HeadState.NONE_PRESENT)


class MovementDetector:
    USED_FRAMES = 10

    def __init__(self, angle_difference_for_movement):
        self.angle_difference_for_movement = angle_difference_for_movement
        self._previous = collections.deque(maxlen=self.USED_FRAMES)

    def add_angle(self, euler_angle):
        self._previous.append(euler_angle)

    def has_moved(self):
        return (len(self._previous) == self.USED_FRAMES
                and self._compute_denoised_angle_change() >= self.angle_difference_for_movement)

    def _compute_denoised_angle_change(self):
        historic_angles = list(self._previous)[:-1]
        current_angle = self._previous[-1]
        summed = 0
        for angle in historic_angles:
            summed += _angles_difference(angle, current_angle)
        count = len(historic_angles)
        return summed / count


def _angles_difference(angle1, angle2):
    return np.absolute(angle1 - angle2).sum()
