# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

import collections

import cv2
import numpy as np

from . import settings

assert cv2.__version__.startswith('3.4'), cv2.__version__

# fill in 3D ref points(world coordinates), model referenced
# from http:# aifi.isr.uc.pt/Downloads/OpenGL/glAnthropometric3DModel.cpp
OBJECT_PTS = np.array([
    [6.825897, 6.760612, 4.402142],  # #33 left brow left corner
    [1.330353, 7.122144, 6.903745],  # #29 left brow right corner
    [-1.330353, 7.122144, 6.903745],  # #34 right brow left corner
    [-6.825897, 6.760612, 4.402142],  # #38 right brow right corner
    [5.311432, 5.485328, 3.987654],  # #13 left eye left corner
    [1.789930, 5.393625, 4.413414],  # #17 left eye right corner
    [-1.789930, 5.393625, 4.413414],  # #25 right eye left corner
    [-5.311432, 5.485328, 3.987654],  # #21 right eye right corner
    [2.005628, 1.409845, 6.165652],  # #55 nose left corner
    [-2.005628, 1.409845, 6.165652],  # #49 nose right corner
    [2.774015, -2.080775, 5.048531],  # #43 mouth left corner
    [-2.774015, -2.080775, 5.048531],  # #39 mouth right corner
    [0.000000, -3.116408, 6.097667],  # #45 mouth central bottom corner
    [0.000000, -7.415691, 4.070434],  # #6 chin corner
])

# reproject 3D points world coordinate axis to verify result pose
REPROJECTSRC = np.array([
    [10.0, 10.0, 10.0],
    [10.0, 10.0, -10.0],
    [10.0, -10.0, -10.0],
    [10.0, -10.0, 10.0],
    [-10.0, 10.0, 10.0],
    [-10.0, 10.0, -10.0],
    [-10.0, -10.0, -10.0],
    [-10.0, -10.0, 10.0],
])


Pose = collections.namedtuple('Pose', ('rotation_vector', 'translation_vector'))


def calculate_pose(face_shape):
    # 2D ref points(image coordinates), referenced from detected facial feature
    # fill in 2D ref points, annotations follow https:# ibug.doc.ic.ac.uk/resources/300-W/
    image_pts = np.array([
        [face_shape.part(17).x, face_shape.part(17).y],  # 17 left brow left corner
        [face_shape.part(21).x, face_shape.part(21).y],  # 21 left brow right corner
        [face_shape.part(22).x, face_shape.part(22).y],  # 22 right brow left corner
        [face_shape.part(26).x, face_shape.part(26).y],  # 26 right brow right corner
        [face_shape.part(36).x, face_shape.part(36).y],  # 36 left eye left corner
        [face_shape.part(39).x, face_shape.part(39).y],  # 39 left eye right corner
        [face_shape.part(42).x, face_shape.part(42).y],  # 42 right eye left corner
        [face_shape.part(45).x, face_shape.part(45).y],  # 45 right eye right corner
        [face_shape.part(31).x, face_shape.part(31).y],  # 31 nose left corner
        [face_shape.part(35).x, face_shape.part(35).y],  # 35 nose right corner
        [face_shape.part(48).x, face_shape.part(48).y],  # 48 mouth left corner
        [face_shape.part(54).x, face_shape.part(54).y],  # 54 mouth right corner
        [face_shape.part(57).x, face_shape.part(57).y],  # 57 mouth central bottom corner
        [face_shape.part(8).x, face_shape.part(8).y],  # 8 chin corner
    ]).astype(np.float64)
    assert image_pts.dtype == OBJECT_PTS.dtype
    assert len(image_pts) == len(OBJECT_PTS) == 14, (len(image_pts), len(OBJECT_PTS))
    _, rotation_vec, translation_vec = cv2.solvePnP(OBJECT_PTS, image_pts,
                                                    settings.CAMERA_INTRISICS,
                                                    settings.DISTORTION_COEFFICIENTS)
    return Pose(rotation_vector=rotation_vec,
                translation_vector=translation_vec)


def calculate_head_box(pose: Pose):
    reprojectdst = _reproject_points(pose)
    reprojectdst = reprojectdst.astype(np.int64)
    return [
        [tuple(reprojectdst[0][0]), tuple(reprojectdst[1][0])],
        [tuple(reprojectdst[1][0]), tuple(reprojectdst[2][0])],
        [tuple(reprojectdst[2][0]), tuple(reprojectdst[3][0])],
        [tuple(reprojectdst[3][0]), tuple(reprojectdst[0][0])],
        [tuple(reprojectdst[4][0]), tuple(reprojectdst[5][0])],
        [tuple(reprojectdst[5][0]), tuple(reprojectdst[6][0])],
        [tuple(reprojectdst[6][0]), tuple(reprojectdst[7][0])],
        [tuple(reprojectdst[7][0]), tuple(reprojectdst[4][0])],
        [tuple(reprojectdst[0][0]), tuple(reprojectdst[4][0])],
        [tuple(reprojectdst[1][0]), tuple(reprojectdst[5][0])],
        [tuple(reprojectdst[2][0]), tuple(reprojectdst[6][0])],
        [tuple(reprojectdst[3][0]), tuple(reprojectdst[7][0])],
    ]


def _reproject_points(pose) -> np.array:
    points, _ = cv2.projectPoints(REPROJECTSRC, pose.rotation_vector,
                                  pose.translation_vector,
                                  settings.CAMERA_INTRISICS,
                                  settings.DISTORTION_COEFFICIENTS)
    return points


def calculate_euler_angle(pose: Pose):
    rotation_mat, _ = cv2.Rodrigues(pose.rotation_vector)
    pose_mat = cv2.hconcat([rotation_mat, pose.translation_vector])
    output = cv2.decomposeProjectionMatrix(pose_mat)
    euler_angle = output[-1]
    return euler_angle
