# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http:# www.braintech.pl>
# All rights reserved.

import datetime
import collections
import pathlib
import sys

import cv2
import dlib
import numpy as np

# Source in C++: https://github.com/lincolnhard/head-pose-estimation

resolution = (1920 / 3, 1080 / 3)

# Intrisics can be calculated using opencv sample code under opencv/sources/samples/cpp/tutorial_code/calib3d
# Approximates are used here:
approximated_K = [resolution[0], 0.0, resolution[0] / 2,
                  0.0, resolution[0], resolution[1] / 2,
                  0.0, 0.0, 1.0]
camera_intrisics = np.array(approximated_K, dtype=np.float64).reshape(3, 3)
example_D = [7.0834633684407095e-002, 6.9140193737175351e-002, 0.0, 0.0,
             -1.3073460323689292e+000]
distortion_coefficients = np.array(example_D, dtype=np.float64).reshape(5, 1)

# fill in 3D ref points(world coordinates), model referenced
# from http:# aifi.isr.uc.pt/Downloads/OpenGL/glAnthropometric3DModel.cpp
object_pts = np.array([
    [6.825897, 6.760612, 4.402142],  # #33 left brow left corner
    [1.330353, 7.122144, 6.903745],  # #29 left brow right corner
    [-1.330353, 7.122144, 6.903745],  # #34 right brow left corner
    [-6.825897, 6.760612, 4.402142],  # #38 right brow right corner
    [5.311432, 5.485328, 3.987654],  # #13 left eye left corner
    [1.789930, 5.393625, 4.413414],  # #17 left eye right corner
    [-1.789930, 5.393625, 4.413414],  # #25 right eye left corner
    [-5.311432, 5.485328, 3.987654],  # #21 right eye right corner
    [2.005628, 1.409845, 6.165652],  # #55 nose left corner
    [-2.005628, 1.409845, 6.165652],  # #49 nose right corner
    [2.774015, -2.080775, 5.048531],  # #43 mouth left corner
    [-2.774015, -2.080775, 5.048531],  # #39 mouth right corner
    [0.000000, -3.116408, 6.097667],  # #45 mouth central bottom corner
    [0.000000, -7.415691, 4.070434],  # #6 chin corner
])

# reproject 3D points world coordinate axis to verify result pose
reprojectsrc = np.array([
    [10.0, 10.0, 10.0],
    [10.0, 10.0, -10.0],
    [10.0, -10.0, -10.0],
    [10.0, -10.0, 10.0],
    [-10.0, 10.0, 10.0],
    [-10.0, 10.0, -10.0],
    [-10.0, -10.0, -10.0],
    [-10.0, -10.0, 10.0],
])


Pose = collections.namedtuple('Pose', ('rotation_vector', 'translation_vector'))


def benchmark_detecting_faces():
    start = datetime.datetime.now()
    frames_iterations = 100
    detect_faces_from_camera(frames_iterations)
    end = datetime.datetime.now()
    duration_in_seconds = (end - start).seconds
    print("Done. Frames per second: ", frames_iterations / duration_in_seconds)


def detect_faces_from_camera(frames_iterations):
    window = dlib.image_window()
    generator = _annotated_heads_camera_images_generator()
    for _ in range(frames_iterations):
        image = next(generator)
        window.set_image(image)


def _annotated_heads_camera_images_generator():
    detector = dlib.get_frontal_face_detector()
    predictor = _get_predictor()
    camera_images_generator = _camera_images_generator()
    while True:
        image = next(camera_images_generator)
        for face in _find_faces(detector, image):
            try:
                image = _annotate_face_on_image(face, image, predictor)
            except OverflowError:
                print("Overflow error!")
        yield image


def _get_predictor():
    predictor_path = _get_predictor_path()
    predictor = dlib.shape_predictor(predictor_path)
    return predictor


def _get_predictor_path():
    if len(sys.argv) == 2:
        return sys.argv[1]
    else:
        this_directory = pathlib.Path(__file__).parent
        possible_path = this_directory / 'shape_predictor_68_face_landmarks.dat'
        if possible_path.exists():
            return str(possible_path)
        else:
            raise FileNotFoundError(
                "Give the path to the trained shape predictor model as the first "
                "argument.\n"
                "For example, if you are in the python_examples folder then "
                "execute this program by running:\n"
                "    ./face_landmark_detection.py shape_predictor_68_face_landmarks.dat\n"
                "You can download a trained facial shape predictor from:\n"
                "    http://dlib.net/files/shape_predictor_68_face_landmarks.dat.bz2"
            )


def _camera_images_generator():
    camera = cv2.VideoCapture(0)
    camera.set(cv2.CAP_PROP_FRAME_WIDTH, resolution[0])
    camera.set(cv2.CAP_PROP_FRAME_HEIGHT, resolution[1])
    while True:
        success, image = camera.read()
        assert success, "There was a problem getting image from the camera."
        yield cv2.cvtColor(src=image, code=cv2.COLOR_BGR2GRAY)


def _find_faces(detector, image):
    return detector(image, 1)


def _annotate_face_on_image(face, source_image, predictor):
    image = source_image.copy()
    shape = predictor(image, face)
    color = (0, 0, 255)
    image = _draw_feature_circles(image, shape, color)
    pose = _calculate_pose(shape)
    image = _draw_head_box_on_image(image, pose, color)
    euler_angle = _calculate_euler_angle(pose)
    outtext = "X: " + str(euler_angle[0])
    cv2.putText(image, outtext, (50, 40), cv2.FONT_HERSHEY_SIMPLEX, 0.45, (0, 0, 0))
    outtext = "Y: " + str(euler_angle[1])
    cv2.putText(image, outtext, (50, 60), cv2.FONT_HERSHEY_SIMPLEX, 0.45, (0, 0, 0))
    outtext = "Z: " + str(euler_angle[2])
    cv2.putText(image, outtext, (50, 80), cv2.FONT_HERSHEY_SIMPLEX, 0.45, (0, 0, 0))
    return image


def _draw_feature_circles(image, shape, color):
    for part in shape.parts():
        cv2.circle(image, (part.x, part.y), 2, color, -1)
    return image


def _calculate_pose(shape):
    # 2D ref points(image coordinates), referenced from detected facial feature
    # fill in 2D ref points, annotations follow https:# ibug.doc.ic.ac.uk/resources/300-W/
    image_pts = np.array([
        [shape.part(17).x, shape.part(17).y],  # 17 left brow left corner
        [shape.part(21).x, shape.part(21).y],  # 21 left brow right corner
        [shape.part(22).x, shape.part(22).y],  # 22 right brow left corner
        [shape.part(26).x, shape.part(26).y],  # 26 right brow right corner
        [shape.part(36).x, shape.part(36).y],  # 36 left eye left corner
        [shape.part(39).x, shape.part(39).y],  # 39 left eye right corner
        [shape.part(42).x, shape.part(42).y],  # 42 right eye left corner
        [shape.part(45).x, shape.part(45).y],  # 45 right eye right corner
        [shape.part(31).x, shape.part(31).y],  # 31 nose left corner
        [shape.part(35).x, shape.part(35).y],  # 35 nose right corner
        [shape.part(48).x, shape.part(48).y],  # 48 mouth left corner
        [shape.part(54).x, shape.part(54).y],  # 54 mouth right corner
        [shape.part(57).x, shape.part(57).y],  # 57 mouth central bottom corner
        [shape.part(8).x, shape.part(8).y],  # 8 chin corner
    ]).astype(np.float64)
    assert image_pts.dtype == object_pts.dtype
    assert len(image_pts) == len(object_pts) == 14, (len(image_pts), len(object_pts))
    _, rotation_vec, translation_vec = cv2.solvePnP(object_pts, image_pts, camera_intrisics, distortion_coefficients)
    return Pose(rotation_vector=rotation_vec, translation_vector=translation_vec)


def _draw_head_box_on_image(image, pose: Pose, color):
    reprojectdst = _reproject_points(pose)
    reprojectdst = reprojectdst.astype(np.int64)
    cv2.line(image, tuple(reprojectdst[0][0]), tuple(reprojectdst[1][0]), color)
    cv2.line(image, tuple(reprojectdst[1][0]), tuple(reprojectdst[2][0]), color)
    cv2.line(image, tuple(reprojectdst[2][0]), tuple(reprojectdst[3][0]), color)
    cv2.line(image, tuple(reprojectdst[3][0]), tuple(reprojectdst[0][0]), color)
    cv2.line(image, tuple(reprojectdst[4][0]), tuple(reprojectdst[5][0]), color)
    cv2.line(image, tuple(reprojectdst[5][0]), tuple(reprojectdst[6][0]), color)
    cv2.line(image, tuple(reprojectdst[6][0]), tuple(reprojectdst[7][0]), color)
    cv2.line(image, tuple(reprojectdst[7][0]), tuple(reprojectdst[4][0]), color)
    cv2.line(image, tuple(reprojectdst[0][0]), tuple(reprojectdst[4][0]), color)
    cv2.line(image, tuple(reprojectdst[1][0]), tuple(reprojectdst[5][0]), color)
    cv2.line(image, tuple(reprojectdst[2][0]), tuple(reprojectdst[6][0]), color)
    cv2.line(image, tuple(reprojectdst[3][0]), tuple(reprojectdst[7][0]), color)
    return image


def _reproject_points(pose: Pose) -> np.array:
    points, _ = cv2.projectPoints(reprojectsrc, pose.rotation_vector,
                                  pose.translation_vector, camera_intrisics,
                                  distortion_coefficients)
    return points


def _calculate_euler_angle(pose: Pose):
    rotation_mat, _ = cv2.Rodrigues(pose.rotation_vector)
    pose_mat = cv2.hconcat([rotation_mat, pose.translation_vector])
    output = cv2.decomposeProjectionMatrix(pose_mat)
    euler_angle = output[-1]
    return euler_angle


if __name__ == '__main__':
    benchmark_detecting_faces()
