# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved
import asyncio
import os

import numpy as np
from PySide2 import QtCore
from PySide2.QtCore import QEvent, QObject, QJsonValue
from PySide2.QtWidgets import QApplication

from braintech.drivers.blinker.blinker import get_blinker, WrongDeviceNameException
from bci_demo_ssvep.simple_ssvep.frontend import Frontend
from obci_demo.gui.bci_frontend import BCIFrontend
from obci_demo.gui.web_view import unwrap_json


class BlinkerController:

    def __init__(self, fields_number):
        try:
            self.blinker = get_blinker(os.environ.get('BLINKER_DEVICE', 'auto'))  # type:FatBlinker
        except WrongDeviceNameException:
            self.blinker = get_blinker('dummy_fat_blinker', preview=True)  # type:FatBlinker
        self.blinker.set_intensity(1)
        self.stim_freqs = [0] * fields_number
        self._phase_randomistation = True
        self._setup_fields(fields_number)
        self._segments = {}

    def _setup_fields(self, fields_number):
        self._fields_number = fields_number
        self._status_bar_field_id = fields_number
        self._remaining_field_id = fields_number + 1
        self._segments = {}
        self.blinker.set_number_of_fields(self._remaining_field_id + 1)
        fields = []
        if self._fields_number == 2:
            fields.append((0.2, 0.5, 0.2, 0.25))
            fields.append((0.6, 0.5, 0.2, 0.25))
        elif self._fields_number == 3:
            fields.append((0.1, 0.5, 0.2, 0.25)),
            fields.append((0.4, 0.5, 0.2, 0.25)),
            fields.append((0.6, 0.5, 0.2, 0.25)),
        fields.append((0, 0, 1.0, 0.25))
        for field_id, rect in enumerate(fields):
            self.configure_field(field_id, *rect)
        self.fullbright()

    def _get_remaining(self):
        remaining = set(range(self.blinker.SEGMENT_COUNT))
        for segments in self._segments.values():
            remaining -= set(s.id for s in segments)
        return list(remaining)

    def configure_field(self, field_id, x, y, width, height):
        segments = self.blinker.get_segments_for_rect(x, y, width, height)
        self._segments[field_id] = segments
        if segments:
            self.blinker.define_field(field_id, segments)
        self.blinker.define_field(self._remaining_field_id, self._get_remaining())
        self.blinker.flush()

    def fullbright(self):
        self.blinker.stop(flush=False)
        for field in range(self._remaining_field_id + 1):
            self.blinker.disable(field_number=field, brightness=1.0)
        self.blinker.flush()

    def start_stimulation(self):
        self.blinker.disable(field_number=self._remaining_field_id, brightness=0)
        self.blinker.disable(field_number=self._status_bar_field_id, brightness=1.0)
        for f in range(self._fields_number):
            self.blinker.enable(f)
        self.blinker.start(flush=True)

    def update_stim_freqs(self, stim_freqs):
        assert len(stim_freqs) == len(self.stim_freqs)
        self.stim_freqs = stim_freqs[::]

        for field, freq in enumerate(stim_freqs):
            if self._phase_randomistation:
                phase_offset_per_freq = np.random.uniform(0, 360)
            else:
                phase_offset_per_freq = 0
            self.blinker.set_rectangular_function(field, freq, duty_cycle=0.5, luminosity=1,
                                                  phase_offset=phase_offset_per_freq)

    @property
    def channels_number(self):
        return 0


class SSVEPFrontend(BCIFrontend, Frontend):
    TITLE = "SSVEP Stimuli"
    _calibration_finished = QtCore.Signal()

    def __init__(self, field_labels, blinker: BlinkerController):
        self._blinker = blinker
        super().__init__(field_labels)
        self.view.installEventFilter(self)

    def eventFilter(self, obj, event):
        if event.type() == QEvent.Move:
            self.toggle('update_fields')
            return True
        else:
            # standard event processing
            return QObject.eventFilter(self, obj, event)

    @QtCore.Slot(QJsonValue)
    @unwrap_json
    def set_field_sizes(self, fields_def):
        p = self.view.pos()
        v_x, v_y = p.x(), p.y()
        s = self.view.size()
        v_w, v_h = s.width(), s.height()
        desktop = QApplication.instance().desktop()
        screen_rect = desktop.screenGeometry(self.view)
        for field_id, rect in fields_def.items():
            x = (v_x - screen_rect.x() + rect['x'] * v_w) / screen_rect.width()
            y = (v_y - screen_rect.y() + rect['y'] * v_h) / screen_rect.height()
            width = (rect['width'] * v_w) / screen_rect.width()
            height = (rect['height'] * v_h) / screen_rect.height()
            self._blinker.configure_field(int(field_id), x, y, width, height)

    @QtCore.Slot(int, float, float, float, float)
    def set_field_size(self, field_id, x, y, width, height):
        # x,y,width,height are in precentages relative to view size
        self.set_field_sizes({field_id: {'x': x, 'y': y, 'width': width, 'height': height}})

    def _handle_key(self, key):
        super(SSVEPFrontend, self)._handle_key(key)

    async def calibration_finished(self):
        self.instruction = ("Frequencies selected. You can finish calibration now, by closing this window "
                            "or press space to continue to collect more statistics for better intention detection."
                            )
        await self._space_pressed.wait()

    async def display_prediction(self, field, feedback_duration, is_correct=True):
        if field is None:
            self.instruction = "No answer decoded"
        else:
            self.instruction = "'%s' selected" % self._field_labels[field]
            self.select(field)
        await super(SSVEPFrontend, self).display_prediction(field, feedback_duration, is_correct)
        self.clear()

    async def set_pause_text(self):
        self.instruction = 'BREAK'

    async def display_no_prediction(self):
        await super(SSVEPFrontend, self).display_no_prediction()
        await  self.display_prediction(None, 1.0, False)

    async def ask_question(self):
        self.instruction = "Focus your attention on field with your answer"
        await asyncio.sleep(1.0)

    async def calibrate_for(self, field_id):
        if field_id == -1:
            self.instruction = "Please focus your attention on this text."
        else:
            self.instruction = "Please focus your attention on field %s" % self._field_labels[field_id]
        self.toggle('update_fields')
        await asyncio.sleep(1.0)

    async def prepare_calibration(self, wait_on_space):
        self._space_pressed.clear()
        if wait_on_space:
            text = "Get ready for calibration. Press space to continue."
        else:
            text = ''
        self.instruction = text
