# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved
import numpy as np
from scipy.signal import iirnotch, butter

from bci_panel.web_bci_panel import RemoteWebBCIPanel
from bci_ssvep.logic.ssvep_calibrator import CalibrationStartedDebugMsg
from bci_ssvep.utils.visualization import SsvepVisualizationDispatcher

from obci_p300.helper_functions import get_line_filters
from braintech.obci.core.broker.message_handler_mixin import subscribe_message_handler
from braintech.obci.core.broker.messages import SignalMessage
from braintech.obci.core.broker.peer import Peer
from braintech.obci.signal_processing.signal.containers import SamplePacket
from obci_p300.common.signal_visualiser import SignalVisualiser, SignalSpectrumVisualiser
from obci_p300.utils import ArrayCascadeFilter
from .analysis_proxy import SSVEPAnalysisMsg, SSVEPCalibrationResult, SSVEPDecision
from ..simple_ssvep.analysis import SimpleSSVEPAnalysis
from obci_p300.common.signal_preprocessor import SignalPreprocessor


class SSVEPAnalysisWithPeer(SimpleSSVEPAnalysis):

    def __init__(self, peer, visualiser, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._peer = peer
        self._visualiser = visualiser
        self._visualiser.reset_feature_means(stimulation_freqs=self.stimulation_freqs)

    def _handle_calibration_done(self):
        super()._handle_calibration_done()
        self._visualiser.handle_cv_result(self.calibration_result)

        self.send_calibration_result()

    def send_calibration_result(self):
        self._peer.send_message(SSVEPCalibrationResult(
            result=self.calibration_result
        ))

    def handle_prediction(self, prediction, timestamp):
        if prediction:
            field = self.stimulation_freqs.index(prediction)
        else:
            field = None
        self._peer.send_message(SSVEPDecision(field=field, timestamp=timestamp))

    def _handle_calibration_start(self, calibration_type):
        self._peer.send_message(CalibrationStartedDebugMsg(frequency=float(calibration_type)))
        super()._handle_calibration_start(calibration_type)

    def _add_features_for_learning(self, features):
        super()._add_features_for_learning(features)

        x_tick_labels = []
        for f in self.stimulation_freqs:
            x_tick_labels.append('{:0.2f}'.format(f))

        x_tick_locations = np.arange(0, len(features))

        self._visualiser.handle_features(features=features, calibration_type=self.calibration_type,
                                         x_ticks=[x_tick_locations, x_tick_labels])


class SSVEPAnalysisPeer(Peer):
    def __init__(self, *args, **kwargs):
        self._config = kwargs.pop('config')
        self._analysis = None
        self._preprocessor = None
        super(SSVEPAnalysisPeer, self).__init__(*args, **kwargs)

    def get_base_filters_def(self, power_line_freq):
        nyquist_rate = self._config['sampling_rate'] / 2
        filters = get_line_filters(nyquist_rate, line_freq=power_line_freq)
        # 125 Hz notch for perun amp
        if nyquist_rate > 125:
            w0 = 125 / nyquist_rate
            Q = w0 / (5 / nyquist_rate)
            b, a = iirnotch(w0, Q)
            filters.append({'b': b, 'a': a})

        b, a = butter(2, Wn=[20 / nyquist_rate, 60 / nyquist_rate], btype="bandpass")
        filters.append({'b': b, 'a': a})

        return filters

    def get_visualising_filters_def(self, power_line_freq):
        nyquist_rate = self._config['sampling_rate'] / 2
        filters = get_line_filters(nyquist_rate, line_freq=power_line_freq)

        # 125 Hz notch for perun amp
        if nyquist_rate > 125:
            w0 = 125 / nyquist_rate
            Q = w0 / (5 / nyquist_rate)
            b, a = iirnotch(w0, Q)
            filters.append({'b': b, 'a': a})

        b, a = butter(2, Wn=0.5 / nyquist_rate, btype="highpass")
        filters.append({'b': b, 'a': a})

        b, a = butter(2, Wn=40 / nyquist_rate, btype="lowpass")
        filters.append({'b': b, 'a': a})

        return filters

    async def _initialized(self):
        await super(SSVEPAnalysisPeer, self)._initialized()
        self._visualizer = SSVEPVisualizer(self, self._config)
        filters = ArrayCascadeFilter(filters_def=self.get_base_filters_def(self._config['power_line_freq']))
        self._preprocessor = SignalPreprocessor(
            self._config['channel_names'],
            self._config['channels_for_classification'],
            self._config['channels_for_montage'],
            self._config['sampling_rate'],
            self._config['exclude_from_montage'],
            notch_filters=[self._config['power_line_freq']],
            channels_gains=self._config['channel_gains'],
            channels_offsets=self._config['channel_offsets'],
            filters=filters,
        )
        filters = ArrayCascadeFilter(filters_def=self.get_visualising_filters_def(self._config['power_line_freq']))
        self._visualizing_preprocessor = SignalPreprocessor(
            self._config['channel_names'],
            self._config['channels_for_classification'],
            self._config['channels_for_montage'],
            self._config['sampling_rate'],
            self._config['exclude_from_montage'],
            notch_filters=[self._config['power_line_freq']],
            channels_gains=self._config['channel_gains'],
            channels_offsets=self._config['channel_offsets'],
            filters=filters,
        )

    def create_analysis(self, params):
        params.update({'sample_rate': self._config['sampling_rate'],
                       'channels_for_classification': self._config['channels_for_classification'],
                       'visualiser': self._visualizer
                       })
        self._analysis = SSVEPAnalysisWithPeer(self, **params)

    @subscribe_message_handler(SignalMessage)
    def handle_signal(self, message):
        self._visualizer.handle_impedance(message.data)
        preprocessed = self._preprocessor.get_preprocessed_signal(message.data.samples)
        preprocessed_packet = SamplePacket(ts=message.data.ts, samples=preprocessed)

        preprocessed_for_visualisation = self._visualizing_preprocessor.get_preprocessed_signal(message.data.samples)
        preprocessed_packet_for_visualisation = SamplePacket(ts=message.data.ts, samples=preprocessed_for_visualisation)
        self._visualizer.handle_raw_signal(preprocessed_packet_for_visualisation)
        if self._analysis:
            self._analysis.handle_sample_packet(preprocessed_packet)

    @subscribe_message_handler(SSVEPAnalysisMsg)
    def handle_ssvep_message(self, message):
        req_type = message.req_type
        if req_type == SSVEPAnalysisMsg.NEW_ANALYSIS:
            self.create_analysis(message.data)
        elif req_type == SSVEPAnalysisMsg.START_CALIBRATION:
            self._analysis.start_calibration(message.data)
        elif req_type == SSVEPAnalysisMsg.STOP_CALIBRATION:
            self._analysis.end_calibration()
        elif req_type == SSVEPAnalysisMsg.SAVE_CALIBRATION:
            self.save_calibration(path=message.data)
        elif req_type == SSVEPAnalysisMsg.LOAD_CALIBRATION:
            self.load_calibration(path=message.data)

    def set_calibration(self, calibration):
        pass

    def save_calibration(self, path):
        self._analysis.save(path)

    def load_calibration(self, path):
        self._analysis.load_wisdom(path)
        self._analysis.send_calibration_result()


class SSVEPVisualizer(SsvepVisualizationDispatcher):
    def __init__(self, peer, config):
        super(SSVEPVisualizer, self).__init__(channel_names=config['channel_names'],
                                              channels_after_montage=config['channels_for_classification'],
                                              panel=RemoteWebBCIPanel(peer),
                                              sampling_rate=config['sampling_rate'],
                                              stimulation_freqs=[], bci_panel_class=RemoteWebBCIPanel,
                                              )
        self.signal_visualiser = SignalVisualiser(self.bci_panel,
                                                  chnames=config['channels_for_classification'],
                                                  every=self.bci_panel.SIGNAL_UPDATE_RATE,
                                                  last_s=self.bci_panel.SIGNAL_CHUNK_LENGTH,
                                                  fs=self.sampling_rate,
                                                  feature_name='raw_signal')
        self.signal_fft_visualiser = SignalSpectrumVisualiser(
            bci_panel=self.bci_panel,
            fs=self.sampling_rate,
            chnames=config['channels_for_classification'],
            feature_name='raw_fft',
            every=self.bci_panel.SIGNAL_UPDATE_RATE,
            last_s=self.bci_panel.SIGNAL_CHUNK_LENGTH,
            freq_mask=config['field_freqs_search'],
        )

    def handle_raw_signal(self, sample_packet):
        self.signal_visualiser.handle_packet(sample_packet)
        self.signal_fft_visualiser.handle_packet(sample_packet)

    def handle_impedance(self, sample_packet):
        self.signal_impedance_visualiser.handle_packet(sample_packet)

    def reset_best_freqs(self):
        self.bci_panel.update_text(
            field_id='None',
            feature='calibration_progress',
            data={"Selected best freqs": 'None', }
        )

    def reset_calibration_progress(self):
        self.bci_panel.update_text(
            field_id='None',
            feature='calibration_progress',
            data={"Mean AUC": ' 0.0',
                  "Mean example count per freq": '0.0'},
            title='calibration_progress',
        )

    def handle_best_frequencies_selected(self, best_freqs):
        freqs_strings = ['{:0.2f}'.format(i) for i in best_freqs]
        freqs_string = ', '.join(freqs_strings)
        self.bci_panel.update_text(
            field_id='None',
            feature='calibration_progress',
            data={"Selected best freqs": freqs_string, }
        )

    def _send_update_calibration_progress(self, mean_auc, freqs_string, example_count):
        self.bci_panel.update_text(
            field_id='None',
            feature='calibration_progress',
            data={"Mean AUC": mean_auc,
                  "Mean example count per freq": str(example_count)},
            title='calibration_progress',
        )
