# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved
import pickle
import threading
import time
from logging import getLogger
from queue import Queue, Empty

from braintech.obci.core.broker.messages import BaseMessage, Field

LOGGER = getLogger(__name__)
LOGGER.setLevel('DEBUG')


class SSVEPAnalysisMsg(BaseMessage):
    req_type = Field(int)
    data = Field(None, object)
    NEW_ANALYSIS = 1
    START_CALIBRATION = 2
    STOP_CALIBRATION = 3
    CALIBRATION_FINISHED = 4
    SAVE_CALIBRATION = 5
    LOAD_CALIBRATION = 6


class SSVEPCalibrationResult(BaseMessage):
    result = Field(object)

    def serialize_data(self):
        return pickle.dumps(self.result)

    @classmethod
    def deserialize_data(cls, data: bytes):
        return {'result': pickle.loads(data)}


class SSVEPDecision(BaseMessage):
    field = Field(int, None)
    timestamp = Field(float)


class AnalysisProxy:
    VALID_PREDICTIONS_THRESHOLD = 0.5

    def __init__(self, peer):
        self._peer = peer  # type: Peer
        self._peer.subscribe_for_all_msg_subtype(SSVEPDecision, self._handle_decision)
        self._peer.subscribe_for_all_msg_subtype(SSVEPCalibrationResult, self._handle_calibration)
        self.calibration_result = None
        self._calibration_result_event = threading.Event()
        self.clear_prediction_queue()
        self.stimulation_freqs = []

    def clear_prediction_queue(self):
        self._valid_prediction_timestamp = time.time() + self.VALID_PREDICTIONS_THRESHOLD
        self._queue = Queue()
        LOGGER.debug("Clear")

    def _send(self, req_type, data=None):
        self._peer.send_message(SSVEPAnalysisMsg(req_type=req_type,
                                                 data=data))

    def _handle_decision(self, message: SSVEPDecision):
        if message.timestamp > self._valid_prediction_timestamp:
            LOGGER.debug('Handle decision: %s %s %s', message.field, message.timestamp,
                         self._valid_prediction_timestamp)
            self._queue.put(message.field)

    def start_calibration(self, calibration_type):
        self._calibration_result_event.clear()
        self.calibration_result = None
        self._send(SSVEPAnalysisMsg.START_CALIBRATION, calibration_type)

    def end_calibration(self):
        self._send(SSVEPAnalysisMsg.STOP_CALIBRATION)

    def _handle_calibration(self, msg):
        self.calibration_result = msg.result
        self._calibration_result_event.set()

    def get_prediction_field(self, block=True, timeout=None):
        try:
            return self._queue.get(block=block, timeout=timeout)
        except Empty:
            return None

    def new_analysis(self, **params):
        self.stimulation_freqs = params['stimulation_frequencies']
        self._send(SSVEPAnalysisMsg.NEW_ANALYSIS, params)

    def wait_for_calibration_result(self):
        self._calibration_result_event.wait()
        return self.calibration_result

    def save(self, path):
        self._send(SSVEPAnalysisMsg.SAVE_CALIBRATION, path)

    def load_wisdom(self, path):
        self.calibration_result = None
        self._calibration_result_event.clear()
        self._send(SSVEPAnalysisMsg.LOAD_CALIBRATION, path)
        self.wait_for_calibration_result()
