# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved

from bci_ssvep.logic.ssvep_metacalibrator import FieldFreqsProvider

from braintech.obci.core.broker.message_handler_mixin import NotRegisteredException
from braintech.obci.core.broker.peer import Peer
from .analysis import SSVEPVisualizer
from .analysis_proxy import AnalysisProxy
from ..simple_ssvep.calibrator import SSVEPCalibratorLogic
from ..simple_ssvep.meta_calibrator import SSVEPMetaCalibrator


class FrontendSSVEPCalibratorLogic(SSVEPCalibratorLogic):
    async def _prepare_calibration(self):
        await self._frontend.prepare_calibration(self.pause_on_start)
        await self.pause(space_ends=self.pause_on_start, pause_text=False)


class PeerMetaCalibrator(SSVEPMetaCalibrator):
    CalibratorClass = FrontendSSVEPCalibratorLogic

    def __init__(self, peer, *args, **kwargs):
        super(PeerMetaCalibrator, self).__init__(*args, **kwargs)
        self._peer = peer  # type: SSVEPLogic
        self._analysis_proxy = AnalysisProxy(peer)

    def _create_analysis(self):
        self._analysis_proxy.new_analysis(
            stimulation_frequencies=self.field_freqs,
            sample_rate=self.sample_rate,
            window=self.window,
            feedback_every=self.feedback_every,
            learning_every=self.learning_every,
            channels_for_classification=self._channels_for_classification
        )
        return self._analysis_proxy

    async def _find_best_frequencies(self):
        self._visualizer.handle_frequency_preference_function(
            self.field_freqs_provider.get_available_freqs(),
            self.field_freqs_provider.get_freq_preference_function(),
            self.field_freqs_provider.get_base_frequency_preference_function())
        await super(PeerMetaCalibrator, self)._find_best_frequencies()
        await self._peer.initial_calibration_finished()


class SSVEPLogic(Peer):
    def __init__(self, *args, **kwargs):
        self._config = kwargs.pop('config')
        kwargs['autoshutdown'] = False
        kwargs['autostart'] = False
        self._calibration_mode = True
        self._frontend = kwargs.pop('frontend')  # type: SSVEPFrontend
        self._blinker = kwargs.pop('blinker')
        self.is_calibrated = False
        super().__init__(*args, **kwargs)

    async def _initialized(self):
        await super()._initialized()

        self._frontend.set_loop(self._loop)
        self._visualizer = SSVEPVisualizer(self, self._config)
        await self._create_calibrator()

    async def _create_calibrator(self):
        try:
            self.unsubscribe_message_handler('ssvep_decision')
        except NotRegisteredException:
            pass
        try:
            self.unsubscribe_message_handler('ssvep_calibration_result')
        except NotRegisteredException:
            pass
        self._freqs_provider = FieldFreqsProvider(
            min_freq=self._config['field_freqs_search'][0],
            max_freq=self._config['field_freqs_search'][1],
            step=2,
            n=2)
        self._calibrator = PeerMetaCalibrator(self,
                                              frontend=self._frontend,
                                              field_freqs_provider=self._freqs_provider,
                                              sample_rate=self._config['sampling_rate'],
                                              blinker_controller=self._blinker,
                                              visualizer=self._visualizer,
                                              channels_for_classification=self._config['channels_for_classification'],
                                              )

    async def _start(self):
        await super(SSVEPLogic, self)._start()
        if self._calibration_mode:
            await self._create_calibrator()
            task = self._calibrator.calibrate()
        else:
            task = self._calibrator.endless_feedback()
        self._task = self.create_task(task)

    async def _calibration_finished(self):
        self.is_calibrated = True
        await self._frontend.calibration_finished()

    def start_calibration(self):
        self._calibration_mode = True
        self.is_calibrated = False
        self.start()

    def start_bci(self):
        self._calibration_mode = False
        self.start()

    async def _stop(self):
        self._task.cancel()
        if self._calibration_mode and self.is_calibrated:
            self._calibrator.save(self._config['wisdom_path'])

    async def initial_calibration_finished(self):
        await self._calibration_finished()

    def load_calibration(self, path):
        self._calibrator.load_wisdom(path)
        self.is_calibrated = self._calibrator.is_calibrated()
