# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved
import asyncio
import time
from logging import getLogger

from bci_demo_ssvep.simple_ssvep.frontend import Frontend
from bci_ssvep.utils.ssvep_blinker_controller import SSVEPBlinkerController
from obci_ugm.utils.sequence_provider import RandomSequentialMgr

LOGGER = getLogger(__name__)


class SSVEPCalibratorLogic:
    TRIAL_TIME = 6
    PREPARE_TIME = 0.5
    PAUSE_TIME = 2
    QUESTION_TIME = 5
    FEEDBACK_TIME = 2

    def __init__(self, ssvep_blinker_controller: SSVEPBlinkerController,
                 analysis,
                 frontend: Frontend,
                 pause_on_start=True,
                 calibration_fields=(0, 1, 2),
                 minimum_example_count=3,
                 feedback=None,
                 ):
        self.pause_on_start = pause_on_start
        self.analysis = analysis
        self.ssvep_blinker_controller = ssvep_blinker_controller
        self.calibration_fields = calibration_fields
        self.side_manager = RandomSequentialMgr(0, len(self.calibration_fields))
        self.calibration_fields_counter = {i: 0 for i in self.calibration_fields}
        self._running = False
        self._running_finished = None
        self._last_answer_timestamp = 0
        self._minimum_example_count = minimum_example_count
        self._feedback = feedback
        self._frontend = frontend

    async def calibrate(self, infinite=False):
        await self._prepare_calibration()
        while True:
            await self._prepare_trial()
            calibration_field_id = self.calibration_fields[self.side_manager.get_value()]

            LOGGER.info("Starting trial for field {}".format(calibration_field_id))
            await self.pretrial(calibration_field_id)
            await self.trial(calibration_field_id)
            space_ends = self._frontend.was_space_pressed()
            await self.pause(space_ends=space_ends)
            LOGGER.info('Trial counter: {}'.format(self.calibration_fields_counter))
            self.calibration_fields_counter[calibration_field_id] += 1
            if not infinite and self.is_calibrated():
                return

    async def _prepare_trial(self):
        LOGGER.debug("Prepare trial")

    async def _prepare_calibration(self):
        await self.pause(space_ends=self.pause_on_start, pause_text=False)

    async def pause(self, space_ends=False, pause_text=True):
        LOGGER.info('Pause')
        if pause_text:
            await self._frontend.set_pause_text()
        if space_ends:
            await self._frontend.wait_for(Frontend.SPACE)
        else:
            await asyncio.sleep(self.PAUSE_TIME)

    async def pretrial(self, calibration_field):
        await self._frontend.calibrate_for(calibration_field)

    async def trial(self, calibration_field):
        self.start_trial_stimuli()
        try:
            await asyncio.sleep(self.PREPARE_TIME)
            trial_start_time = time.time()
            await self.mark_trial_start(trial_start_time, calibration_field)
            LOGGER.debug("Trial for %s", calibration_field)
            await asyncio.sleep(self.TRIAL_TIME - (time.time() - trial_start_time))
            await self.mark_trial_end(time.time(), calibration_field)
        finally:
            self.ssvep_blinker_controller.fullbright()

    def is_calibrated(self):
        result = self.analysis.calibration_result
        return result and result.get_minimum_example_count() >= self._minimum_example_count

    def start_trial_stimuli(self):
        self.ssvep_blinker_controller.start_stimulation()

    async def mark_trial_start(self, trial_start_time, field_id):
        self.analysis.start_calibration(field_id)
        pass

    async def mark_trial_end(self, trial_start_time, field_id):
        self.analysis.end_calibration()
        pass
