# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved
import operator
from logging import getLogger

from bci_demo_ssvep.peers.analysis import SSVEPVisualizer
from .analysis import SimpleSSVEPAnalysis
from .calibrator import SSVEPCalibratorLogic
from .feedback import SimpleFeedbackLogic

LOGGER = getLogger(__name__)


class SSVEPMetaCalibrator:
    MINIMUM_EXAMPLE_COUNT = 5
    DEFAULT_WINDOW = 1.0
    DEFAULT_FEEDBACK_EVERY = 0.5
    DEFAULT_LEARNING_EVERY = 1.0
    DEFAULT_CRITICAL_AUC = 0.8
    CalibratorClass = SSVEPCalibratorLogic

    def __init__(self,
                 field_freqs_provider,
                 sample_rate,
                 blinker_controller,
                 visualizer: SSVEPVisualizer,
                 frontend,
                 channels_for_classification,
                 window=None,
                 feedback_every=None,
                 learning_every=None,
                 critical_auc=None,
                 start_with_pause=True,
                 use_harmonics=True,

                 ):

        self.crit_auc = critical_auc or self.DEFAULT_CRITICAL_AUC
        self.sample_rate = sample_rate
        self.window = window or self.DEFAULT_WINDOW
        self.feedback_every = feedback_every or self.DEFAULT_FEEDBACK_EVERY
        self.learning_every = learning_every or self.DEFAULT_LEARNING_EVERY
        self.start_with_pause = start_with_pause
        self.field_freqs_provider = field_freqs_provider
        self.field_freqs = []
        self.statuses = {}
        self.frontend = frontend

        # to keep best ROC
        self.rocs_per_calibrations = {}  # per frequencies sets
        self.rocs_per_frequencies = {}  # per frequencies
        self._latest_calibration_result = None

        self._visualizer = visualizer

        # for pause on first calib
        self._first_calib = True
        self._current_analysis = None
        self._current_calibrator = None
        self._current_blinker_controller = blinker_controller
        self._use_harmonics = use_harmonics
        self._current_feedback = None
        self._channels_for_classification = channels_for_classification

    async def calibrate(self, infinite=True):
        self.frontend.clear()
        await self._find_best_frequencies()
        if infinite:
            await self._infinite_calibration_on_best_freqs()

    async def _find_best_frequencies(self):
        self._visualizer.reset_calibration_progress()
        self._visualizer.reset_best_freqs()
        while not self.are_best_frequencies_found():
            self._start_new_calib()
            await self._current_calibrator.calibrate()
            self._latest_calibration_result = await self.get_calibration_result()
            self.rocs_per_calibrations[tuple(self.field_freqs)] = self._latest_calibration_result.get_minimum_roc()
            for f in self.field_freqs:
                roc = self._latest_calibration_result.roc_for_freq(f)
                self.rocs_per_frequencies[f] = roc
            self.field_freqs_provider.update_frequency_preference_from_auc(self.rocs_per_frequencies,
                                                                           self.crit_auc)
            self._visualizer.handle_frequency_preference_function(
                self.field_freqs_provider.get_available_freqs(),
                self.field_freqs_provider.get_freq_preference_function(),
                self.field_freqs_provider.get_base_frequency_preference_function(),
            )

    async def get_calibration_result(self):
        return self._current_analysis.calibration_result

    def are_best_frequencies_found(self):
        return any(auc > self.crit_auc
                   for auc in self.rocs_per_calibrations.values())

    def _start_new_calib(self):
        field_freqs = self.field_freqs_provider.get()
        LOGGER.info("Starting new calibration for freqs: {}".format(field_freqs))
        self._create_calibration_components(field_freqs, self._first_calib)

        if self._first_calib is True:
            self._first_calib = False

    async def _infinite_calibration_on_best_freqs(self):
        try:
            best_freqs = sorted(self.rocs_per_calibrations.items(), key=operator.itemgetter(1), reverse=True)[0][0]
        except IndexError:
            LOGGER.exception(
                "No best frequencies something is wrong!!!!. Choosing random frequencies for now")
            best_freqs = [39.0, 41.0, 43.0]

        self._handle_found_best_freqs(best_freqs)
        if set(best_freqs) != set(self._current_analysis.stimulation_freqs):
            self._create_calibration_components(best_freqs, self._first_calib)
        self._current_calibrator.pause_on_start = False
        await self._current_calibrator.calibrate(infinite=True)

    def _handle_found_best_freqs(self, best_freqs):
        LOGGER.info("Found best frequencies. Starting "
                    "infinite calibration on freqs {}.".format(best_freqs)
                    )
        self._visualizer.handle_best_frequencies_selected(best_freqs)

    def _create_calibration_components(self, stim_freqs, pause_on_start):
        self.field_freqs = stim_freqs
        self._current_analysis = self._create_analysis()
        self._visualizer.reset_feature_means(stim_freqs)
        self._current_blinker_controller.update_stim_freqs(self.field_freqs)
        self._current_feedback = self._create_feedback_controller()
        self._current_calibrator = self._create_calibrator(pause_on_start)

    def _create_analysis(self) -> SimpleSSVEPAnalysis:
        return SimpleSSVEPAnalysis(
            stimulation_frequencies=self.field_freqs,
            sample_rate=self.sample_rate,
            window=self.window,
            feedback_every=self.feedback_every,
            learning_every=self.learning_every,
            channels_for_classification=self._channels_for_classification,
        )

    def _create_calibrator(self, pause_on_start) -> SSVEPCalibratorLogic:
        return self.CalibratorClass(
            self._current_blinker_controller,
            self._current_analysis,
            pause_on_start=pause_on_start and self.start_with_pause,
            calibration_fields=[-1] + list(range(len(self.field_freqs))),
            minimum_example_count=self.MINIMUM_EXAMPLE_COUNT,
            frontend=self.frontend,
            feedback=self._current_feedback
        )

    def _create_feedback_controller(self):
        return SimpleFeedbackLogic(
            analysis=self._current_analysis,
            frontend=self.frontend,
            blinker_controller=self._current_blinker_controller
        )

    async def endless_feedback(self):
        self.frontend.clear()
        await self._current_feedback.endless_feedback()

    def handle_sample_packet(self, packet):
        if self._current_analysis is not None:
            self._current_analysis.handle_sample_packet(packet)

    def handle_tag_msg(self, tagmsg):
        if self._current_analysis is not None:
            self._current_analysis.handle_tag_msg(tagmsg)

    def load_wisdom(self, path):
        # to get calibrations freqs from wisdom
        self.field_freqs = self.field_freqs_provider.get()
        temp_analysis = self._create_analysis()
        temp_analysis.load_wisdom(path)
        # proper load
        wisdom_freqs = temp_analysis.calibration_result.get_available_freqs()
        self._create_calibration_components(wisdom_freqs, True)
        self._current_analysis.load_wisdom(path)

    def save(self, path):
        if self._current_analysis:
            self._current_analysis.save(path)

    def is_calibrated(self):
        return self._current_calibrator.is_calibrated() and self._current_analysis.calibration_result.is_valid()
