# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved
import warnings
try:
    import _tkinter
except (RuntimeError, ModuleNotFoundError) as e:
    warnings.warn(str(e))
import math
import os
import traceback
from collections.__init__ import defaultdict

try:
    from matplotlib import pyplot as plt
except (RuntimeError, ModuleNotFoundError) as e:
    warnings.warn(str(e))


class BCIPanel:
    MAX_COLS = 5
    LEGEND_FONTSIZE = 5
    CUSTOM_TICK_FONTSIZE = 7
    CUSTOM_TICK_LIMIT = 10

    def __init__(self):
        """Create dynamic window with updatable plots."""
        plt.ion()
        self.figure = plt.figure('BCI Control Panel®')
        self.axes = {}  # keys are 'feature' 'id' - feature - name of the feature, id - id of the field or None
        self.plots = defaultdict(list)
        self.texts = defaultdict(list)
        self.n_rows = 1
        self.n_cols = 1
        self.pos = 1
        self.figure.canvas.mpl_connect('resize_event', self.on_resize)

    def on_resize(self, event):
        if len(self.axes) > 0:
            try:
                self.figure.tight_layout()
            except ValueError:
                pass

    def close(self):
        plt.close(self.figure)

    def _expand(self):
        n_plots = len(self.axes.keys()) + 1

        self.n_cols = n_plots
        self.n_rows = 1
        if n_plots > self.MAX_COLS:
            self.n_cols = self.MAX_COLS
            self.n_rows = math.ceil(n_plots / self.MAX_COLS)

    def _sort_plots(self):
        for nr, i in enumerate(sorted(self.axes.keys(), key=lambda x: str(x[0]) + str(x[1]))):
            ax = self.axes[i]
            ax.change_geometry(self.n_rows, self.n_cols, nr + 1)

    def _create_ax(self, field_id, feature, full_title, xlabel, ylabel):
        self._expand()
        ax = self.figure.add_subplot(self.n_rows, self.n_cols, self.pos)
        self.axes[(field_id, feature)] = ax
        self.pos += 1

        ax.set_title(full_title, fontsize=6)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        self._sort_plots()
        try:
            self.figure.tight_layout()
        except ValueError:
            pass
        return ax

    def _clear_ax(self, field_id, feature):
        """Clear plots and legends.
        Mostly used to update histograms.
        """
        ax = self.axes[(field_id, feature)]
        full_title = ax.get_title()
        xlabel = ax.get_xlabel()
        ylabel = ax.get_ylabel()
        ax.clear()
        ax.set_title(full_title, fontsize=6)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        try:
            del self.plots[(field_id, feature)]
        except KeyError:
            pass

    def _create_plot(self, field_id, feature, full_title, xlabel, ylabel, ylim, x_data, y_data, labels, plot_type,
                     legend_loc):
        ax = self._create_ax(field_id, feature, full_title, xlabel, ylabel)
        self._create_plot_on_existing_axes(field_id, feature, ylim, x_data, y_data, labels, plot_type, ax, legend_loc)

    def _create_plot_on_existing_axes(self, field_id, feature, ylim, x_data, y_data, labels, plot_type, ax, legend_loc):
        if ylim is not None:
            ax.set_ylim(*ylim)
        if plot_type == 'plot' or plot_type == 'scatter':
            for nr, (x, y) in enumerate(zip(x_data, y_data)):
                try:
                    label = labels[nr]
                except IndexError:
                    label = ''
                if plot_type == 'plot':
                    plots = ax.plot(x, y, label=label)
                if plot_type == "scatter":
                    plots = ax.plot(x, y, '-o', label=label)
                self.plots[(field_id, feature)].append(plots[0])
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(handles, labels, loc=legend_loc, fontsize=self.LEGEND_FONTSIZE)

    def _update_plot(self, field_id, feature, ylim, x_data, y_data, title, full_title):
        for nr, (x, y) in enumerate(zip(x_data, y_data)):
            self.plots[(field_id, feature)][nr].set_data(x, y)
        ax = self._update_ax(field_id, feature, title, full_title)
        if ylim is not None:
            ax.set_ylim(*ylim)

    def _update_ax(self, field_id, feature, title, full_title):
        ax = self.axes[(field_id, feature)]
        ax.relim()
        ax.autoscale_view(True, True, True)
        if title is not None:
            ax.set_title(full_title, fontsize=6)
        return ax

    @staticmethod
    def _prepare_title(field_id, feature, title):
        return 'field {}, feature {}\n{}'.format(field_id, feature, title)

    def update_text(self, field_id, feature, data: {str: str}, title):
        key = (field_id, feature)
        if key in self.axes:
            for text in self.texts[key]:
                text.remove()
            self.texts[key] = []
            ax = self.axes[key]
        else:
            full_title = self._prepare_title(field_id, feature, title)
            ax = self._create_ax(field_id, feature, full_title,
                                 xlabel='', ylabel='')
            ax.axis('off')

        start_x = 0
        start_y = -1
        for i, data_item in enumerate(data.items()):
            label, value = data_item
            text = '{}: {}'.format(label, value)
            reference = ax.text(0, start_y - i, text)
            self.texts[key].append(reference)

        approx_max_x = start_x + 1
        end_y = start_y - (len(data) - 1)
        padding = 0.5
        ax.set_xlim(left=start_x - padding, right=approx_max_x + padding)
        ax.set_ylim(top=start_y + padding, bottom=end_y - padding)

    def update_plot(self, field_id=None, feature='', xlabel='', ylabel='', title=None,
                    x_data=[[], ], y_data=[[], ], plot_type='plot', labels=[''], ylim=None, x_ticks=None, legend_loc=2,
                    **kwargs):
        """"Create plot or change plot data/title.

        Creates plot for unique id and feature pairs (strings or other hashable types).
        If used with the same feature and id pair only title, x_data, y_data are used - to update existing plot.

        When creating plot - you can plot more than one line:
        x_data, y_data - list of lists (or numpy 1D arrays) to be plotted
        labels - list of string for name the curves given by x_data and y_data, if shorter than len(x_data) - unlabeled
            plots will have empty string in the legend.
        type - string - type of plot, for now only 'plot' or "scatter" is suported.
        ylim - 2 element list - if you want to set plot ylim
        """
        assert len(x_data) == len(y_data)

        full_title = self._prepare_title(field_id, feature, title)
        if (field_id, feature) in self.axes:
            try:
                self._update_plot(field_id, feature, ylim, x_data, y_data, title, full_title)
            except IndexError:
                self._clear_ax(field_id, feature)
                self._create_plot_on_existing_axes(field_id, feature, ylim, x_data, y_data, labels,
                                                   plot_type, self.axes[(field_id, feature)], legend_loc=legend_loc)
        else:
            self._create_plot(field_id, feature, full_title, xlabel, ylabel, ylim, x_data, y_data, labels,
                              plot_type, legend_loc)
        if x_ticks is not None:
            tick_labels = x_ticks[1]
            tick_locations = x_ticks[0]
            if len(tick_labels) > self.CUSTOM_TICK_LIMIT:
                tick_labels = tick_labels

            skip_every = int(len(tick_labels) / self.CUSTOM_TICK_LIMIT)
            if skip_every > 0:
                tick_labels = tick_labels[::skip_every]
                tick_locations = tick_locations[::skip_every]

            ax = self.axes[field_id, feature]
            ax.set_xticks(tick_locations)
            ax.set_xticklabels(tick_labels, rotation='vertical', fontsize=self.CUSTOM_TICK_FONTSIZE)

    def update_hist(self, field_id='', feature='', data=[[], ], bins=10, title='', xlabel='', ylabel='',
                    labels=[], normed=False, vline=None, xlim=None, **kwargs):
        """"Create histogram or change histogram data/title.

        Creates histogram for unique id and feature pairs (strings or other hashable types).
        If used with the same feature and id pair only title, x_data, y_data are used - to update existing
        plot/histogram.

        When creating histogram - you can bin more than one data:
        data - list of lists (or numpy 1D arrays) to be plotted
        labels - list of string for name the curves given by x_data and y_data, if shorter than len(x_data) - unlabeled
            plots will have empty string in the legend.
        bins - compatable with pylab.hist
            """

        full_title = self._prepare_title(field_id, feature, title)

        if (field_id, feature) in self.axes:
            self._update_hist(field_id, feature, data, bins, title, full_title, normed, labels, vline, xlim)
        else:
            self._create_hist(field_id, feature, full_title, xlabel, ylabel, data, bins, labels, normed, vline, xlim)

    def save_plots(self, dir_path):
        dirname = os.path.realpath(os.path.expanduser(dir_path))
        os.makedirs(dirname, exist_ok=True)
        for key, plot in self.axes.items():
            file_path = dirname + '/' + "_".join(map(str, key)) + '.png'
            bbox_inches = plot.get_tightbbox(self.figure.canvas.renderer).transformed(
                self.figure.dpi_scale_trans.inverted())
            self.figure.savefig(file_path, bbox_inches=bbox_inches, dpi=300)

    def _paint_hist(self, field_id, feature, data, bins, labels, normed, vline, xlim):
        ax = self.axes[(field_id, feature)]
        try:
            _, _, patches = ax.hist(data, bins=bins, label=labels, normed=normed)
            if xlim:
                ax.set_xlim(xlim)
        except ValueError:
            traceback.print_exc()
            _, _, patches = ax.hist([], bins=bins, label=labels, normed=normed)
            if xlim:
                ax.set_xlim(xlim)
        if vline is not None:
            ax.axvline(vline, color='black')
        self.plots[(field_id, feature)] = patches

        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles, labels, loc=2, fontsize=self.LEGEND_FONTSIZE)

    def _create_hist(self, field_id, feature, full_title, xlabel, ylabel, data, bins, labels, normed, vline, xlim):
        self._create_ax(field_id, feature, full_title, xlabel, ylabel)
        self._paint_hist(field_id, feature, data, bins, labels, normed, vline, xlim)

    def _update_hist(self, field_id, feature, data, bins, title, full_title, normed, labels, vline, xlim):
        self._update_ax(field_id, feature, title, full_title)
        self._clear_ax(field_id, feature)
        self._paint_hist(field_id, feature, data, bins, labels, normed, vline, xlim)

    def process_events(self):
        # todo: rename this module
        try:
            self.figure.canvas.flush_events()
        except _tkinter.TclError:
            pass
