# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import logging
import os
import pickle
from collections import defaultdict
from collections import deque
from concurrent.futures import ThreadPoolExecutor
from copy import deepcopy
from functools import partial
from queue import Queue

import numpy as np
from scipy.signal import iirnotch, butter

from braintech.obci.core.broker.messages import TagMsg
from braintech.obci.signal_processing.buffers.auto_ring_buffer import AutoRingBuffer
from braintech.obci.signal_processing.signal.data_generic_write_proxy import SamplePacket
from obci_p300.helper_functions import get_line_filters
from obci_p300.utils import ArrayCascadeFilter
from .ssvep_decider import SSVEPDecider
from .ssvep_frequency_finder import SSVEPFrequencyFinder
from ..utils.ssvep_calibration_result import SSVEPCalibrationResult


class SignalGoodnessDetector:
    def __init__(self, detector_list=None):
        self._detector_list = detector_list or ()

    def is_good_chunk(self, timestamped_chunk):
        start_ts = timestamped_chunk[-1, 0]
        stop_ts = timestamped_chunk[-1, -1]
        for detector in self._detector_list:
            good = detector(start_ts, stop_ts)
            if not good:
                return False
        return True


class SSVEPAnalysis:
    "Hardcoded channel names and order: M1;M2;O1;O2;Pz;Cz"
    CALIBRATE_EVERY = 5
    CALIBRATE_USING_LAST_N_BUFFERS = 200000
    MINIMUM_SAMPLES_PER_TYPE = 5

    def __init__(self, stimulation_frequencies=None, sample_rate=512,
                 window=2.0, feedback_every=0.25, learning_every=2.0,
                 channels_for_classification=None, channels_for_montage=None,
                 channel_names=None, signal_goodness_reporters=None, use_harmonics=True,
                 estimator=None, classifier=None, decision_buffer_len=2,
                 power_line_freq=50, exclude_from_montage=()):
        if not isinstance(signal_goodness_reporters, SignalGoodnessDetector):
            self._signal_goodness_detector = SignalGoodnessDetector(signal_goodness_reporters)
        else:
            self._signal_goodness_detector = signal_goodness_reporters
        if estimator is None:
            self._estimator = SSVEPFrequencyFinder(sampling_freq=sample_rate, stimulation_freqs=stimulation_frequencies,
                                                   use_harmonics=use_harmonics)
        else:
            self._estimator = estimator

        if classifier is None:
            self._classifier = SSVEPDecider(stimulation_freqs=stimulation_frequencies)
        else:
            self._classifier = classifier

        self.stimulation_freqs = self._estimator.stimulation_freqs

        self._logger = logging.getLogger('ssvep_analysis.' + self.__class__.__name__)
        self.sample_rate = sample_rate
        self.nyquist_rate = sample_rate / 2
        self.window = window
        self.feedback_every = feedback_every
        self.learning_every = learning_every

        buffor_size_samples = int(sample_rate * window)
        self.buffor_size_samples = buffor_size_samples
        return_length_samples = int(sample_rate * window)
        return_feedback_every_samples = int(sample_rate * feedback_every)
        return_learning_every_samples = int(sample_rate * learning_every)
        self.calibration_type = None

        self.calibration_buffer = defaultdict(partial(deque, maxlen=self.CALIBRATE_USING_LAST_N_BUFFERS))
        self.calibration_counter = 0

        self.baseline_filter = ArrayCascadeFilter(filters_def=self.get_base_filters_def(power_line_freq))
        if channels_for_classification is None:
            self.channels_for_classification = 'O1;O2;Pz;Cz'.split(';')
        else:
            self.channels_for_classification = channels_for_classification

        if channels_for_montage is None:
            self.channels_for_montage = ['M1', 'M2']
        else:
            self.channels_for_montage = channels_for_montage
        self.exclude_from_montage = exclude_from_montage

        if channel_names is None:
            self.chanel_names = self.channels_for_montage + self.channels_for_classification
        else:
            self.chanel_names = channel_names  # in the order of amplifier
        buffer_channel_count = len(self.channels_for_classification) + 1  # timestamps channel
        self.learning_buffer = AutoRingBuffer(from_sample=buffor_size_samples,
                                              samples_count=return_length_samples,
                                              every=return_learning_every_samples,
                                              num_of_channels=buffer_channel_count,
                                              ret_func=self._handle_learning_buffer_callback,
                                              copy_on_ret=True,
                                              )

        self.feedback_buffer = AutoRingBuffer(from_sample=buffor_size_samples,
                                              samples_count=return_length_samples,
                                              every=return_feedback_every_samples,
                                              num_of_channels=buffer_channel_count,
                                              ret_func=self._handle_feedback_buffer_callback,
                                              copy_on_ret=True,
                                              )

        self.calibration_future = None
        self.calibration_executor = ThreadPoolExecutor(max_workers=1)
        self.calibration_result = None

        self.decision_buffer = deque(maxlen=decision_buffer_len)
        self.reset_buffers()

    def save(self, path):
        path = os.path.realpath(os.path.expanduser(path))
        dir = os.path.dirname(path)
        os.makedirs(dir, exist_ok=True)

        with open(path, 'wb') as f:
            pickle.dump(SSVEPCalibrationResult(self.calibration_buffer, self._classifier), f)

    def load_wisdom(self, path):
        path = os.path.realpath(os.path.expanduser(path))
        with open(path, 'rb') as f:
            ssvep_calibration_result = pickle.load(f)
            self._classifier = ssvep_calibration_result.ssvep_decider
            self.calibration_buffer = ssvep_calibration_result.calibration_buffer
            self.calibration_result = ssvep_calibration_result

    def reset_buffers(self):
        self.prediction_queue = Queue()
        self.decision_buffer.clear()
        self.feedback_buffer.clear()

    def get_base_filters_def(self, power_line_freq):
        filters = get_line_filters(self.nyquist_rate, line_freq=power_line_freq)

        nyquist_rate = self.nyquist_rate

        # 125 Hz notch for perun amp
        if nyquist_rate > 125:
            w0 = 125 / nyquist_rate
            Q = w0 / (5 / nyquist_rate)
            b, a = iirnotch(w0, Q)
            filters.append({'b': b, 'a': a})

        b, a = butter(2, Wn=[20 / nyquist_rate, 60 / nyquist_rate], btype="bandpass")
        filters.append({'b': b, 'a': a})

        return filters

    def _handle_calibration_start(self, calibration_type):
        """Calibration type - float or None - None - no calibration, any number - user has his attention on
        this frequency (in Hz), 0 - User is avoiding all stimulants. This method should be called some time after
        stimulation start to ignore VEPs and only get signal with steady-state response."""
        self._logger.debug("handle_calibration_start")
        self.calibration_type = calibration_type
        if calibration_type is not None:
            self.learning_buffer.clear()

    def _handle_calibration_stop(self):
        self._logger.debug("handle_calibration_stop")
        self.calibration_type = None

    def handle_sample_packet(self, packet: SamplePacket):
        self._logger.debug("handle_sample_packet")
        raw_data = packet.samples
        data = self.ssvep_montage(raw_data)

        filtered_data = self.baseline_filter.get(data)

        timestamps, impedance = packet.ts, packet.impedance

        timestamped_filtered_data = np.concatenate((filtered_data, timestamps[:, None]), axis=1)
        self.learning_buffer.handle_sample_packet(
            SamplePacket(samples=timestamped_filtered_data, ts=timestamps, impedance=impedance)
        )
        self.feedback_buffer.handle_sample_packet(
            SamplePacket(samples=timestamped_filtered_data, ts=timestamps, impedance=impedance)
        )

    def ssvep_montage(self, raw_data):
        sample_count = raw_data.shape[0]
        data = np.empty((sample_count, len(self.channels_for_classification)))
        #   M1;M2;O1;O2;Pz;Cz

        montage_base = np.zeros(sample_count)
        try:
            norm = 1 / len(self.channels_for_montage)
        except ZeroDivisionError:
            norm = 0

        for chan in self.channels_for_montage:
            montage_base += (norm * raw_data[:, self.chanel_names.index(chan)])

        for final_nr, chan in enumerate(self.channels_for_classification):
            chan_id = self.chanel_names.index(chan)
            if chan_id in self.exclude_from_montage:
                data[:, final_nr] = raw_data[:, chan_id]
            else:
                data[:, final_nr] = raw_data[:, chan_id] - montage_base
        return data

    def _handle_learning_buffer_callback(self, timestamped_chunk):
        chunk = timestamped_chunk[:-1, :]
        signal_good = self._run_through_signal_goodness_detectors(timestamped_chunk)
        if signal_good:
            features = self._get_frequency_predictions_from_chunk(chunk)
            self._add_features_for_learning(features)

    def _handle_feedback_buffer_callback(self, timestamped_chunk):
        chunk = timestamped_chunk[:-1, :]
        signal_good = self._run_through_signal_goodness_detectors(timestamped_chunk)
        if signal_good:
            features = self._get_frequency_predictions_from_chunk(chunk)
            prediction = self._predict_from_features(features)
            prediction_timestamp = timestamped_chunk[-1, 0]
            self.handle_prediction(prediction, prediction_timestamp)

    def handle_prediction(self, prediction, timestamp):
        self.prediction_queue.put(prediction)

    def _run_through_signal_goodness_detectors(self, timestamped_chunk):
        return self._signal_goodness_detector.is_good_chunk(timestamped_chunk)

    def _get_frequency_predictions_from_chunk(self, chunk):
        features = self._estimator.predict(chunk.T)
        return features

    def is_enough_data_for_calibration(self):
        if len(self.calibration_buffer.keys()) != (1 + len(self._estimator.stimulation_freqs)):
            return False

        for key in self.calibration_buffer.keys():
            if len(self.calibration_buffer[key]) < self.MINIMUM_SAMPLES_PER_TYPE:
                return False
        return True

    def _calibrate(self):
        calibrate_every = (self.calibration_counter % self.CALIBRATE_EVERY == 0)
        calibration_running = (self.calibration_future is not None)

        if calibrate_every and (not calibration_running):
            self.calibration_future = self.calibration_executor.submit(self._calibrate_classifier,
                                                                       deepcopy(self.calibration_buffer),
                                                                       deepcopy(self._classifier))
        elif calibration_running and self.calibration_future.done():
            self._handle_calibration_done()

    def _handle_calibration_done(self):
        self._logger.debug("handle_calibration_done")
        self.calibration_result = self.calibration_future.result()
        self.calibration_future = None
        self._classifier = self.calibration_result.ssvep_decider

    def _calibrate_classifier(self, calibration_buffer, classifier):
        classifier.fit(calibration_buffer)
        return SSVEPCalibrationResult(calibration_buffer, classifier)

    def _add_features_for_learning(self, features):
        if self.calibration_type is not None:
            self.calibration_buffer[self.calibration_type].append(np.copy(features))
            self.calibration_counter += 1
            if self.is_enough_data_for_calibration():
                self._calibrate()

    def _predict_from_features(self, features):
        prediction = self._classifier.predict(features)
        self.decision_buffer.append(prediction)
        if len(set(self.decision_buffer)) == 1 and len(self.decision_buffer) == self.decision_buffer.maxlen:
            prediction = self.decision_buffer[-1]
            self.decision_buffer.clear()
        else:
            prediction = None
        return prediction

    def handle_tag_msg(self, tag: TagMsg):
        if tag.name == 'ssvep_trial_start':
            calibration_type = tag.desc['calibration_type']
            self._handle_calibration_start(calibration_type)
        elif tag.name == 'ssvep_trial_end':
            self._handle_calibration_stop()

    def close(self):
        pass
