# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import numpy as np
from scipy.stats import scoreatpercentile
from sklearn.metrics import auc


class FreqThresholds:
    def __init__(self, freq, abs_thr=0, diff_thr=0, accuracy_abs=0, fpr_abs=None, tpr_abs=None,
                 accuracy_diff=0, fpr_diff=None, tpr_diff=None, abs_auroc=None, diff_auroc=None, diff_thrs=None,
                 diff_accs=None, abs_thrs=None, abs_accs=None):
        self.freq = freq
        self.abs_thr = abs_thr
        self.diff_thr = diff_thr
        self.accuracy_abs = accuracy_abs
        self.fpr_abs = fpr_abs
        self.tpr_abs = tpr_abs
        self.abs_auroc = abs_auroc
        self.accuracy_diff = accuracy_diff
        self.fpr_diff = fpr_diff
        self.tpr_diff = tpr_diff
        self.diff_auroc = diff_auroc
        self.diff_thrs = diff_thrs
        self.diff_accs = diff_accs
        self.abs_thrs = abs_thrs
        self.abs_accs = abs_accs

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.__dict__ == other.__dict__
        return False


class SSVEPDecider:
    def __init__(self, stimulation_freqs=None):
        if stimulation_freqs is None:
            self.stimulation_freqs = [30, 32]
        else:
            self.stimulation_freqs = stimulation_freqs

        self.thresholds = {i: FreqThresholds(i) for i in self.stimulation_freqs}

    def fit(self, calibration_buffor):
        """Calibration buffor with keys '0' - avoided all frequencies and n strings - frequencies of stimulation.
        The values are lists of vectors of len(self.stimulation_freqs) - vectors with contribution degrees
        calculated by Lasso for each stimulation frequency.
        """

        for f in self.stimulation_freqs:
            self._fit_freq(f, calibration_buffor)

    def _fit_freq(self, freq, calibration_buffor):
        freq_id = self.stimulation_freqs.index(freq)

        avoided = np.array(calibration_buffor['0'])[:, freq_id]  # distribution of contribution degrees for this
        # frequency when all fields are being avoided by the user

        this_freq = np.array(calibration_buffor[str(freq)])[:, freq_id]  # distribution of contribution degrees for
        # this frequency when this frequency is in focus of users attention

        # distribution of contribution degrees of every other frequencies when this frequency is in focus of attention
        other_freqs = np.delete(np.array(calibration_buffor[str(freq)]), freq_id, axis=1)

        # looking for abs_thr
        tprs_abs = []
        fprs_abs = []
        thrs_abs = []
        accs_abs = []

        max_search = np.max([np.max(this_freq), np.max(avoided)]) * 1.1
        min_search = 0 - max_search * 0.1

        for thr in np.linspace(min_search, max_search, 1000):
            tpr = np.sum((this_freq > thr)) / this_freq.size
            fpr = np.sum((avoided > thr)) / avoided.size
            acc = (np.sum((this_freq > thr)) + np.sum((avoided <= thr))) / (this_freq.size + avoided.size)

            tprs_abs.append(tpr)
            fprs_abs.append(fpr)
            thrs_abs.append(thr)
            accs_abs.append(acc)

        max_accuracy_abs_id = np.argmax(accs_abs)
        max_accuracy_abs = accs_abs[max_accuracy_abs_id]
        # max_accuracy_abs_thr = thrs_abs[max_accuracy_abs_id]

        min_fpr_abs_id = np.argmin(fprs_abs)
        min_fpr_thr_abs = thrs_abs[min_fpr_abs_id]

        sorting_index = np.argsort(np.array(fprs_abs))
        abs_auroc = auc(np.array(fprs_abs)[sorting_index], np.array(tprs_abs)[sorting_index])

        # looking for diff_thr
        tprs_diff = []
        fprs_diff = []
        thrs_diff = []
        accs_diff = []

        max_other_freqs = scoreatpercentile(other_freqs, 95)  # not max, for robustness against artifacts
        max_search = (np.max(this_freq) - max_other_freqs) * 1.1
        min_search = np.min([np.min(this_freq), np.min(other_freqs)]) - max_other_freqs
        min_search -= np.max([np.abs(min_search) * 0.1, np.abs(max_search) * 0.1])

        for thr in np.linspace(min_search, max_search, 1000):
            tpr = np.sum(((this_freq - max_other_freqs) > thr)) / this_freq.size
            fpr = np.sum(((other_freqs - max_other_freqs) > thr)) / other_freqs.size

            tp = np.sum(((this_freq - max_other_freqs) > thr))
            tn = np.sum(((other_freqs - max_other_freqs) <= thr))
            acc = (tp + tn) / (this_freq.size + other_freqs.size)

            tprs_diff.append(tpr)
            fprs_diff.append(fpr)
            thrs_diff.append(thr)
            accs_diff.append(acc)

        max_accuracy_diff_id = np.argmax(accs_diff)
        max_accuracy_diff = accs_diff[max_accuracy_diff_id]
        # max_accuracy_diff_thr = thrs_diff[max_accuracy_diff_id]

        min_fpr_diff_id = np.argmin(fprs_diff)
        min_fpr_thr_diff = thrs_diff[min_fpr_diff_id]

        sorting_index = np.argsort(np.array(fprs_diff))
        diff_auroc = auc(np.array(fprs_diff)[sorting_index], np.array(tprs_diff)[sorting_index])

        self.thresholds[freq] = FreqThresholds(freq, abs_thr=min_fpr_thr_abs, diff_thr=min_fpr_thr_diff,
                                               accuracy_abs=max_accuracy_abs, fpr_abs=fprs_abs, tpr_abs=tprs_abs,
                                               accuracy_diff=max_accuracy_diff, fpr_diff=fprs_diff, tpr_diff=tprs_diff,
                                               abs_auroc=abs_auroc, diff_auroc=diff_auroc, diff_thrs=thrs_diff,
                                               diff_accs=accs_diff, abs_thrs=thrs_abs, abs_accs=accs_abs,
                                               )

    def predict(self, feature):
        max_freq = self.stimulation_freqs[np.argmax(feature)]
        second_largest_contribution_degree = np.sort(feature)[-2]
        max_freq_contribution_degree = max(feature)

        abs_thr = self.thresholds[max_freq].abs_thr

        if max_freq_contribution_degree <= abs_thr:
            return None

        diff_thr = self.thresholds[max_freq].diff_thr

        if (max_freq_contribution_degree - second_largest_contribution_degree) <= diff_thr:
            return None

        return max_freq
