# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
from _warnings import warn

import numpy as np
from scipy.signal import welch
from sklearn.linear_model import Lasso

from bci_panel.bci_panels import BCIPanelProcess

DEBUG = False


class SSVEP_FFT_SNR_FrequencyFinder:
    MINIMUN_WELCH_WINDOWS = 20
    WELCH_WINDOW_MAXIMUM_LENGTH_SECONDS = 0.5
    LINE_FREQ = 50

    def __init__(self, sampling_freq=512, stimulation_freqs=None, use_harmonics=True):
        if stimulation_freqs is None:
            stimulation_freqs = [30, 32]

        assert max(stimulation_freqs) < sampling_freq / 2
        if use_harmonics:
            assert max(stimulation_freqs) * 2 < sampling_freq / 2

        self._stimulation_freqs = stimulation_freqs
        self._sampling_freq = sampling_freq
        self._use_harmonics = use_harmonics

        if DEBUG:
            self.___bci_panel = BCIPanelProcess()

    @property
    def stimulation_freqs(self):
        return self._stimulation_freqs[::]

    def _get_signal_power_noise_power(self, f, sample_fft, sample_len):
        """sample_fft np.array - [freq, channel]"""
        times = np.linspace(0, sample_len / self._sampling_freq - 1 / self._sampling_freq, sample_len)
        target = np.sin(2 * np.pi * f * times)
        freqs, target_fft_pow = self.do_welch(target)
        target_peak_index = np.argmax(target_fft_pow)
        target_peak = np.max(target_fft_pow)

        sample_fft_pow = np.abs(sample_fft)
        sample_peaks_multiplier = sample_fft_pow[target_peak_index, :] / target_peak

        target_fft_pow_fitted = target_fft_pow[:, None] * sample_peaks_multiplier

        if DEBUG:
            self.___bci_panel.update_plot(field_id='None', feature='test {}'.format(f),
                                          x_data=[freqs] * 2,
                                          y_data=[target_fft_pow_fitted[:, 0],
                                                  sample_fft[:, 0]])

        signal_power = target_fft_pow_fitted.sum()
        noise_power = np.sum(sample_fft_pow)
        return signal_power, noise_power

    def welch_overlap_samples(self, sample_len):
        win_len = self.welch_window_samples(sample_len)
        min_overlap = win_len / 2
        if sample_len / min_overlap < self.MINIMUN_WELCH_WINDOWS:
            return win_len - (sample_len / self.MINIMUN_WELCH_WINDOWS)
        else:
            return min_overlap

    def welch_window_samples(self, sample_len):
        half_sample_len_seconds = sample_len / self._sampling_freq
        if half_sample_len_seconds > self.WELCH_WINDOW_MAXIMUM_LENGTH_SECONDS:
            return int(self._sampling_freq / 2)
        else:
            return int(sample_len / 2)

    def do_welch(self, sample, axis=0):
        sample_len = sample.shape[0]
        freqs, fft = welch(sample, fs=self._sampling_freq, nperseg=self.welch_window_samples(sample_len),
                           axis=0, noverlap=self.welch_overlap_samples(sample_len)
                           )
        for line_f in np.arange(self.LINE_FREQ, self._sampling_freq / 2, self.LINE_FREQ):
            mask = np.array(freqs > line_f - 5) * np.array(freqs < line_f + 5)
            fft[mask] = 0
        return freqs, fft

    def predict(self, sample):
        """
        Get contribution degrees of every stimulation frequency in the sample, based on FFT

        :param sample: ndarray - signal sample (time x channels)
        :return: returns array of contribution degrees, sized as self.stimulation_freqs and in the same order
        """
        if sample.shape[0] < sample.shape[1]:
            warn("SSVEP frequency finder - you've got more channels than timesamples, is your data correct? "
                 "Sample shape: {}".format(sample.shape))

        sample_len = sample.shape[0]

        _, fft = self.do_welch(sample)

        snr_per_freq = []

        for f_stim in self.stimulation_freqs:
            if self._use_harmonics:
                f_list = [f_stim / 2, f_stim, f_stim * 2]
            else:
                f_list = [f_stim, ]
            signal_power = 0
            noise_power = 0
            for f in f_list:
                signal_power_f, noise_power_f = self._get_signal_power_noise_power(f, fft, sample_len)
                signal_power += signal_power_f
                noise_power += noise_power_f
            snr = signal_power / noise_power
            snr_per_freq.append(snr)
        return snr_per_freq


class SSVEPFFTFrequencyFinder:
    def __init__(self, sampling_freq=512, stimulation_freqs=None, use_harmonics=True):
        if stimulation_freqs is None:
            stimulation_freqs = [30, 32]

        assert max(stimulation_freqs) < sampling_freq / 2
        if use_harmonics:
            assert max(stimulation_freqs) * 2 < sampling_freq / 2

        self._stimulation_freqs = stimulation_freqs
        self._sampling_freq = sampling_freq
        self._use_harmonics = use_harmonics

    @property
    def stimulation_freqs(self):
        return self._stimulation_freqs[::]

    def predict(self, sample):
        """
        Get contribution degrees of every stimulation frequency in the sample, based on FFT

        :param sample: ndarray - signal sample (time x channels)
        :return: returns array of contribution degrees, sized as self.stimulation_freqs and in the same order
        """
        if sample.shape[0] < sample.shape[1]:
            warn("SSVEP frequency finder - you've got more channels than timesamples, is your data correct? "
                 "Sample shape: {}".format(sample.shape))

        fft = np.fft.fft(sample, axis=0)
        fft_freq = np.fft.fftfreq(sample.shape[0], 1 / self._sampling_freq)

        # we want at least 3 fft bins

        power_per_freq = []

        for f in self.stimulation_freqs:
            masked_bin_counts = 0
            spacing = 0.2
            while masked_bin_counts < 3:
                fft_mask = (fft_freq >= (f - spacing)) * (fft_freq < (f + spacing))
                masked_bin_counts = np.sum(fft_mask)
                spacing *= 1.20
            power = np.sum(np.abs(fft[fft_mask, :]) ** 2)
            power_per_freq.append(power)
        return power_per_freq


class SSVEPFrequencyFinder:
    def __init__(self, sampling_freq=512, stimulation_freqs=None, use_harmonics=True, max_iter=10000, warm_start=False,
                 selection='cyclic'):
        """DATA MUST BE CENTERED"""
        if stimulation_freqs is None:
            stimulation_freqs = [30, 32]

        assert max(stimulation_freqs) < sampling_freq / 2
        if use_harmonics:
            assert max(stimulation_freqs) * 2 < sampling_freq / 2

        self._stimulation_freqs = stimulation_freqs
        self._sampling_freq = sampling_freq

        self._estimator = Lasso(fit_intercept=False, alpha=1, max_iter=max_iter, warm_start=warm_start,
                                selection=selection)
        self._targets = None  # signal targets (cosines and sines of given frequencies)
        self._targets_freqs = None
        self._use_harmonics = use_harmonics

    def predict(self, sample):
        """
        Get contribution degrees of every stimulation frequency in the sample.

        :param sample: ndarray - signal sample (time x channels), DATA MUST BE CENTERED
        :return: returns array of contribution degrees, sized as self.stimulation_freqs and in the same order
        """
        if sample.shape[0] < sample.shape[1]:
            warn("SSVEP frequency finder - you've got more channels than timesamples, is your data correct? "
                 "Sample shape: {}".format(sample.shape))
        targets, target_freqs = self._get_target_signals(sample)
        # sample = self._scaler.fit_transform(np.abs(sample))
        sample = sample / np.max(sample) * 50  # magic number for LASSO
        # TODO need to test properly of sample vs model amplitude ratios
        self._estimator.fit(sample, targets)
        coef = self._estimator.coef_

        predictions = []
        for f in self.stimulation_freqs:
            freq_mask = (target_freqs == f)
            contribution_degree = np.sum(np.abs(coef[freq_mask]))
            predictions.append(contribution_degree)
        return predictions

    @property
    def stimulation_freqs(self):
        return self._stimulation_freqs[::]

    def _create_targets(self, sample_length):
        timeline = np.linspace(0, sample_length / self._sampling_freq - 1 / self._sampling_freq, sample_length)
        targets = []
        target_freqs = []
        for f in self.stimulation_freqs:
            target_sin = np.sin(2 * np.pi * f * timeline)
            target_cos = np.cos(2 * np.pi * f * timeline)
            targets.append(target_sin)
            target_freqs.append(f)
            targets.append(target_cos)
            target_freqs.append(f)

            if self._use_harmonics:
                targets.append(np.sin(2 * 2 * np.pi * f * timeline))
                target_freqs.append(f)
                targets.append(np.cos(2 * 2 * np.pi * f * timeline))
                target_freqs.append(f)
                targets.append(np.sin(0.5 * 2 * np.pi * f * timeline))
                target_freqs.append(f)
                targets.append(np.cos(0.5 * 2 * np.pi * f * timeline))
                target_freqs.append(f)

        self._targets = np.array(targets).T
        self._targets_freqs = np.array(target_freqs)

    def _get_target_signals(self, sample):
        sample_length = sample.shape[0]
        if self._targets is None:
            self._create_targets(sample_length)
        if self._targets.shape[0] != sample_length:
            self._create_targets(sample_length)
        return self._targets, self._targets_freqs
