# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import asyncio
import time

from bci_ssvep import ssvep_peer
from bci_ssvep.logic.ssvep_metacalibrator import SSVEP_Metacalibrator
from bci_ssvep.ssvep_peer import SSVEPPeer
from bci_ssvep.utils.ssvep_blinker_controller import set_fat_blinker_40_fields_center_status_noncontionious
from braintech.obci.core.broker.message_handler_mixin import subscribe_message_handler
from braintech.obci.core.broker.messages import SignalMessage
from braintech.obci.core.broker.messages import TagMsg
from braintech.obci.core.utils.properties import join_strings
from braintech.obci.core.utils.properties import param_property
from braintech.obci.core.utils.properties import split_floats
from obci_p300.common.bci_reporter import ReporterSummaryMsg
from obci_p300.common.messages import DecisionVerboseMsg
from obci_ugm.gui.ugm import ugm_helper
from obci_ugm.gui.ugm.ugm_keys import UgmKeys
from obci_ugm.utils.messages import UgmKeyEventMsg

__all__ = ('MetaCalibrationSSVEPPeer',)


class UnknownFieldSetupFuncExc(Exception):
    pass


class BaseMetacalibrationSSVEPPeer(ssvep_peer.BaseSSVEPPeer):
    start_with_pause = True

    def __init__(self, *args, **kwargs):
        self.power_line_freq = 50
        self.exclude_from_montage = []
        self.task = None
        super().__init__(*args, **kwargs)

    def _set_static_parameters(self, config, amplifier_config):
        def _deserialize_list(items: str, deserializer):
            if items:
                return [deserializer(i) for i in items.split(';')]
            else:
                return []

        super()._set_static_parameters(config, amplifier_config)
        self.critical_auc = float(config['critical_auc'])
        self.field_freqs_search = _deserialize_list(config['field_freqs_search'], float)
        self.field_freqs_search_min_distance = float(config['field_freqs_search_min_distance'])
        self.field_setup_func = config['field_setup_func']

    @subscribe_message_handler(ReporterSummaryMsg)
    async def handle_reporter_summary(self, msg: ReporterSummaryMsg):
        if self.calibrator:
            self.calibrator.bci_panel.handle_reporter_summary(msg.dict_data)

    def _setup_analysis(self):
        self.analysis = None
        self.reporter = self.create_bci_reporter()

    def create_blinker_controller(self):
        # metacalibrator contains analysis and the calibrator
        return None

    async def logic_process(self):
        ugm_helper.send_config(self, self.ssvep_ugm_config)
        self._soft_start()

    def _soft_start(self):
        assert self.task is None or self.task.done()
        self.task = self.create_task(self._soft_start_task())

    async def _soft_start_task(self):
        self.calibrator = self._create_calibrator()
        if self.calibration_mode:
            await self.calibrator.start_calibration()
        else:
            self.calibrator.load_wisdom(self.wisdom_path)
            await self.calibrator.start_feedback()

    def _create_calibrator(self):
        if self.field_setup_func not in [None, 'set_fat_blinker_40_fields_center_status_noncontionious', "None"]:
            raise UnknownFieldSetupFuncExc(self.field_setup_func)
        if self.field_setup_func == 'set_fat_blinker_40_fields_center_status_noncontionious':
            field_setup_func = set_fat_blinker_40_fields_center_status_noncontionious
        else:
            field_setup_func = None
        signal_goodness_reporters = self._get_signal_goodness_detectors()
        calibrator = SSVEP_Metacalibrator(
            self.field_freqs_search,
            self.field_freqs_search_min_distance,
            self,
            self.ugm_config,
            self.fields_ids_light_indexes,
            self.sampling_rate,
            initial_frontend_parameters=self.frontend_parameters,
            reporter=self.reporter,
            window=1.0,
            feedback_every=0.5,
            learning_every=1.0,
            channels_for_classification=self.channels_for_classification,
            channels_for_montage=self.channels_for_montage,
            channel_names=self.channel_names,
            critical_auc=self.critical_auc,
            signal_goodness_reporters=signal_goodness_reporters,
            use_harmonics=False,
            field_setup_func=field_setup_func,
            bci_panel_class=self.bci_panel_class,
            start_with_pause=self.start_with_pause,
            power_line_freq=self.power_line_freq,
            exclude_from_montage=self.exclude_from_montage,
        )
        return calibrator

    async def _connections_established(self):
        self.calibrator = None
        await super()._connections_established()

    @subscribe_message_handler(SignalMessage)
    def handle_signal(self, msg):
        if self.calibrator is not None:
            self.calibrator.handle_sample_packet(msg.data)

    @subscribe_message_handler(TagMsg)
    def handle_tag(self, msg):
        if self.calibrator is not None:
            self.calibrator.handle_tag_msg(msg)

    @subscribe_message_handler(UgmKeyEventMsg)
    async def handle_ugm_keys(self, msg: UgmKeyEventMsg):
        if msg.getUgmKey() == UgmKeys.Key_Space:
            if self.calibrator is not None:
                self.calibrator.handle_space_pressed()

        if msg.getUgmKey() == UgmKeys.Key_F:
            if self.calibrator is not None:
                self.calibrator.handle_f_pressed()

    async def _stop(self):
        await self._soft_stop()
        await super()._stop()

    async def _soft_stop(self):
        if self.calibrator:
            await self.calibrator.stop()
            self._stop_and_save_analysis()
            if hasattr(self.calibrator, 'bci_panel'):
                self.calibrator.bci_panel.save_plots(self.plots_image_path)
            await self._loop.run_in_executor(executor=None, func=self.calibrator.close)
        if self.task is not None:
            self.task.cancel()
            await asyncio.wait([self.task])

    async def _shutting_down(self):
        if self.calibrator is not None:
            await self.calibrator.stop()
            self._stop_and_save_analysis()
            await self._loop.run_in_executor(executor=None, func=self.calibrator.close)
        await super()._shutting_down()

    def _stop_and_save_analysis(self):
        if self.calibrator is not None:
            self.calibrator.save(self.wisdom_path)

    async def send_decision(self, prediction):
        self.send_message(DecisionVerboseMsg(decision=str(prediction),
                                             score=int(prediction),
                                             decision_type='thetarget',
                                             start_timestamp=time.time(),
                                             stop_timestamp=time.time()))


class MetaCalibrationSSVEPPeer(BaseMetacalibrationSSVEPPeer, SSVEPPeer):
    field_freqs_search = param_property('field_freqs_search', deserializer=split_floats, serializer=join_strings)
    # frequency set should achieve at least this AUC to be accepted for final calibration
    critical_auc = param_property('critical_auc', deserializer=float, serializer=str)
    field_freqs_search_min_distance = param_property('field_freqs_search_min_distance',
                                                     deserializer=float, serializer=str)
    field_setup_func = param_property('field_setup_func')
