# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import time
from collections import defaultdict
from typing import Type

import numpy as np
import scipy.signal

from bci_panel import bci_panels
from obci_p300.common.signal_visualiser import SignalSpectrumVisualiser, SignalImpedanceVisualiser, SignalVisualiser
from obci_p300.helper_functions import FeatureMean, get_line_filters
from obci_p300.utils import SamplePacketCascadeFilter

PLOTS_SAVING_WAIT_SECONDS = 2


class FrequencyPreferenceMixin:
    bci_panel = None

    def handle_frequency_preference_function(self, freqs, preference, preference_base=None):
        if preference_base is not None:
            self.bci_panel.update_plot(
                field_id='all',
                feature='frequency preference function',
                x_data=[freqs, freqs],
                y_data=[preference, preference_base],
                xlabel='Frequency',
                ylabel='Preference',
            )
        else:
            self.bci_panel.update_plot(
                field_id='all',
                feature='frequency preference function',
                x_data=[freqs],
                y_data=[preference],
                xlabel='Frequency',
                ylabel='Preference',
            )


class FeatureMeansMixin:
    def reset_feature_means(self, stimulation_freqs=None):
        if stimulation_freqs is not None:
            self.stimulation_freqs = stimulation_freqs
        self.feature_means = {str(i): FeatureMean() for i in self.stimulation_freqs + [0]}
        self.feature_means[None] = FeatureMean()

    def handle_features(self, features, calibration_type, x_ticks=None):
        self.feature_means[calibration_type].append(features)

        ys = []
        labels = []

        for i in sorted(self.feature_means.keys(), key=str):
            if len(self.feature_means[i]) > 2:
                ys.append(self.feature_means[i].data)
            else:
                ys.append(np.zeros(len(features)))
            labels.append(str(i))

        ys.append(features)
        labels.append('NOW')

        xs = [list(range(len(features))), ] * len(ys)

        self.bci_panel.update_plot(
            field_id='None',
            feature='features',
            x_data=xs,
            y_data=ys,
            title='Signal features',
            labels=labels,
            x_ticks=x_ticks,
            plot_type='scatter',
        )


class SsvepVisualizationDispatcher(FrequencyPreferenceMixin, FeatureMeansMixin):
    MAX_HISTOGRAMS_TO_DRAW = 10

    def __init__(self, channel_names, channels_after_montage, sampling_rate,
                 stimulation_freqs, bci_panel_class: Type[bci_panels.BaseBCIPanelProcess],
                 panel=None, verbose_panel=False):
        if panel is None:
            panel = bci_panel_class()
        self.bci_panel = panel
        self._verbose_panel = verbose_panel

        self.channel_names = channel_names
        self.channels_after_montage = channels_after_montage
        self.sampling_rate = sampling_rate
        self.stimulation_freqs = stimulation_freqs

        every = panel.SIGNAL_UPDATE_RATE
        last_s = panel.SIGNAL_CHUNK_LENGTH
        self.signal_visualiser = RawSignalVisualiser(
            bci_panel=self.bci_panel,
            fs=self.sampling_rate,
            chnames=self.channel_names,
            feature_name='raw_signal',
            every=every,
            last_s=last_s or 3,
        )
        self.signal_fft_visualiser = SignalSpectrumVisualiser(
            bci_panel=self.bci_panel,
            fs=self.sampling_rate,
            chnames=self.channel_names,
            feature_name='raw_fft',
            last_s=last_s or 1,
            every=every,
            freq_mask=[20, 45],
        )
        self.montaged_signal_visualiser = RawSignalVisualiser(
            bci_panel=self.bci_panel,
            fs=self.sampling_rate,
            chnames=self.channels_after_montage,
            feature_name='signal_montage',
            last_s=last_s or 3,
            every=every,
        )
        self.montaged_signal_fft_visualiser = SignalSpectrumVisualiser(
            bci_panel=self.bci_panel,
            fs=self.sampling_rate,
            chnames=self.channels_after_montage,
            feature_name='fft_montage',
            last_s=last_s or 1,
            every=every,
            freq_mask=[20, 45],
        )
        self.signal_impedance_visualiser = SignalImpedanceVisualiser(
            bci_panel=self.bci_panel,
            fs=self.sampling_rate,
            chnames=self.channel_names,
            every=every,
            last_s=last_s or 1,
        )
        self.reset_feature_means()

        self._best_freqs = None
        self._mean_auc = None
        self._example_count = 0
        self.update_calibration_progress()

    def handle_best_frequencies_selected(self, best_freqs):
        self._best_freqs = best_freqs

    def _send_update_calibration_progress(self, mean_auc, freqs_string, example_count):
        # to be overwriten in SSVEP-DEMO (for example)
        self.bci_panel.update_text(
            field_id='None',
            feature='calibration_progress',
            data={"Selected best freqs": '\n' + freqs_string,
                  "Mean AUC": mean_auc,
                  "Mean example count per freq": str(example_count)},
            title='calibration_progress',
        )

    def update_calibration_progress(self):
        if self._best_freqs is not None:
            freqs_string_list = ["{:.2f}".format(i) for i in self._best_freqs]
            freqs_string = ', '.join(freqs_string_list)
        else:
            freqs_string = str(self._best_freqs)
        if self._mean_auc:
            mean_auc = "{:.2f}".format(self._mean_auc)
        else:
            mean_auc = "{}"
        self._send_update_calibration_progress(mean_auc, freqs_string, self._example_count)

    def save_plots(self, path):
        self.bci_panel.save_plots(dir_path=path)
        time.sleep(PLOTS_SAVING_WAIT_SECONDS)

    def handle_reporter_summary(self, data):
        name = 'Accuracy and itr'
        self.bci_panel.update_text(field_id='None', feature=name, data=data, title=name)

    def handle_raw_signal(self, sample_packet):
        self.signal_visualiser.handle_packet(sample_packet)
        self.signal_fft_visualiser.handle_packet(sample_packet)
        self.signal_impedance_visualiser.handle_packet(sample_packet)

    def handle_montaged_signal(self, sample_packet):
        self.montaged_signal_visualiser.handle_packet(sample_packet)
        self.montaged_signal_fft_visualiser.handle_packet(sample_packet)

    def handle_cv_result(self, cv_result):
        calibration_buffer = cv_result.calibration_buffer
        classifier = cv_result.ssvep_decider
        self._mean_auc = cv_result.roc_auc
        self._example_count = cv_result.get_mean_example_count_per_frequency()

        self.update_calibration_progress()

        if len(classifier.stimulation_freqs) > self.MAX_HISTOGRAMS_TO_DRAW:
            return

        for i, f in enumerate(classifier.stimulation_freqs):
            to_hist_dict = defaultdict(list)
            for key in sorted(calibration_buffer.keys()):
                if str(key) == '0':
                    to_hist_key = 'unattended'
                elif str(key) == str(f):
                    to_hist_key = 'attended'
                else:
                    to_hist_key = 'other freqs'

                to_hist_dict[to_hist_key].extend(np.array(calibration_buffer[key])[:, i])

            title_hist = ("F={}, ABS AUC {:.3f}, DIFF AUC {:.3}"
                          .format(f, classifier.thresholds[f].abs_auroc,
                                  classifier.thresholds[f].diff_auroc))

            to_hist = []
            to_hist_labels = []

            for key, value in to_hist_dict.items():
                to_hist_labels.append(key)
                to_hist.append(value)

            max_in_hist = np.max([np.max(i) for i in to_hist])
            self.bci_panel.update_hist(
                field_id=str(f),
                feature='hist_aaa',
                data=to_hist,
                labels=to_hist_labels,
                bins=np.linspace(0, max_in_hist, 10),
                vline=classifier.thresholds[f].abs_thr,
                title=title_hist,
                normed=True,
            )
            if self._verbose_panel:
                title_abs = "Max acc {:.3f}, AUC {:.3f}, thr {:.3f}".format(
                    classifier.thresholds[f].accuracy_abs,
                    classifier.thresholds[f].abs_auroc,
                    classifier.thresholds[f].abs_thr,
                )
                self.bci_panel.update_plot(
                    field_id=str(f),
                    feature='roc_abs',
                    x_data=[classifier.thresholds[f].fpr_abs],
                    y_data=[classifier.thresholds[f].tpr_abs],
                    title=title_abs,
                )

                abs_threshold_x = [classifier.thresholds[f].abs_thr] * 2
                abs_threshold_y = [0, 1]
                self.bci_panel.update_plot(
                    field_id=str(f),
                    feature='fpr_abs',
                    x_data=[classifier.thresholds[f].abs_thrs, abs_threshold_x],
                    y_data=[classifier.thresholds[f].fpr_abs, abs_threshold_y],
                    title=title_abs,
                )

                title_diff = "Max acc {:.3f}, AUC {:.3f}, thr {:.3f}".format(
                    classifier.thresholds[f].accuracy_diff,
                    classifier.thresholds[f].diff_auroc,
                    classifier.thresholds[f].diff_thr,
                )
                self.bci_panel.update_plot(
                    field_id=str(f),
                    feature='roc_diff',
                    x_data=[classifier.thresholds[f].fpr_diff],
                    y_data=[classifier.thresholds[f].tpr_diff],
                    title=title_diff,
                )

                diff_threshold_x = [classifier.thresholds[f].diff_thr] * 2
                diff_threshold_y = [0, 1]
                self.bci_panel.update_plot(
                    field_id=str(f),
                    feature='fpr_diff',
                    x_data=[classifier.thresholds[f].diff_thrs, diff_threshold_x],
                    y_data=[classifier.thresholds[f].fpr_diff, diff_threshold_y],
                    title=title_diff,
                )

    def close(self):
        self.bci_panel.close()


class RawSignalVisualiser(SignalVisualiser):
    """Takes raw signal and filters it for visualisation"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.visual_filter = SamplePacketCascadeFilter(filters_def=self._prepare_filters())

    def _prepare_filters(self):
        nyquist_rate = self.fs / 2

        filters = get_line_filters(nyquist_rate)

        # 125 Hz notch for perun amp
        if nyquist_rate > 125:
            w0 = 125 / nyquist_rate
            Q = w0 / (5 / nyquist_rate)
            b, a = scipy.signal.iirnotch(w0, Q)
            filters.append({'b': b, 'a': a})

        # target_fs filter:
        b, a = scipy.signal.cheby2(2, 12, 30 / nyquist_rate, btype='lowpass')
        filters.append({'b': b, 'a': a})
        return filters

    def handle_packet(self, sample_packet):
        filtered_packet = self.visual_filter.get(sample_packet)
        super().handle_packet(filtered_packet)
