# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import numpy as np
import pytest
import time
import os

from bci_ssvep.logic.ssvep_analysis import SSVEPAnalysis
from braintech.obci.core.broker.messages import TagMsg
from braintech.obci.signal_processing.signal.data_generic_write_proxy import SamplePacket
from bci_ssvep.logic.ssvep_analysis_bci_panel_visualisation import SSVEPAnalysisVisualisation

VISUAL_DEBUG = False


LEFT_FREQ = 38
RIGHT_FREQ = 34


def calibration_type_generator():
    calibration_types = [0, LEFT_FREQ, RIGHT_FREQ]
    i = 0
    while True:
        yield calibration_types[i % len(calibration_types)]
        i += 1


def generate_data():
    """"Generate perfect ssvep signal."""
    Fs = 512.0
    T = 10 * 60
    signal = np.zeros((int(T * Fs), 6))

    np.random.seed(0)
    if VISUAL_DEBUG:
        NOISE_SCALE = 1
    else:
        NOISE_SCALE = 1
    noise = np.random.normal(scale=NOISE_SCALE, size=signal.shape)

    left_stimuli = np.sin(2 * np.pi * LEFT_FREQ * np.arange(0, T, 1 / Fs) + 1.1) / 20
    right_stimuli = np.sin(2 * np.pi * RIGHT_FREQ * np.arange(0, T, 1 / Fs) + 0.6) / 20

    trial_duration = 5

    trial_windows = np.arange(10, T, 20)

    tags = []

    calibration_type_generator_instance = calibration_type_generator()

    for trial in trial_windows:
        calibration_type = next(calibration_type_generator_instance)

        if calibration_type == LEFT_FREQ:
            left_stimuli[int((trial - 1) * Fs):int((trial + trial_duration) * Fs)] *= 4

        if calibration_type == RIGHT_FREQ:
            right_stimuli[int((trial - 1) * Fs):int((trial + trial_duration) * Fs)] *= 4

        tags.append(TagMsg(start_timestamp=float(trial), end_timestamp=float(trial + trial_duration),
                           name='ssvep_trial_start', channels='-1',
                           desc={'calibration_type': str(calibration_type)}))

        tags.append(TagMsg(start_timestamp=float(trial + trial_duration), end_timestamp=float(trial + trial_duration),
                           name='ssvep_trial_end', channels='-1'))

    signal += noise

    signal += left_stimuli[:, None]
    signal += right_stimuli[:, None]

    signal[:, 0:2] = 0  # let our reference channels have nothing
    return signal, tags


if VISUAL_DEBUG:
    classes_to_test = [SSVEPAnalysisVisualisation]
    analysis_fixture_ids = ['SSVEPAnalysisVisualisation']
else:
    classes_to_test = [SSVEPAnalysis, SSVEPAnalysisVisualisation]
    analysis_fixture_ids = ['SSVEPAnalysis', 'SSVEPAnalysisVisualisation']

analysis_fixture_params = classes_to_test


@pytest.fixture(params=analysis_fixture_params, ids=analysis_fixture_ids)
def prepare_analysis_test_data(request):
    cls = request.param
    fs = 512

    analysis = cls(stimulation_frequencies=[LEFT_FREQ, RIGHT_FREQ], sample_rate=512,
                   window=1, feedback_every=0.5, learning_every=1)

    analysis_read_test = cls(stimulation_frequencies=[LEFT_FREQ, RIGHT_FREQ], sample_rate=512,
                             window=1, feedback_every=0.5, learning_every=1)
    try:
        signal, tags = generate_data()

        if VISUAL_DEBUG:
            packet_size = 16
        else:
            packet_size = 256

        for i in range(0, len(signal), packet_size):
            packet = signal[i:i + packet_size]
            ts_start = i / fs
            ts_end = (i + packet_size) / fs
            signal_packet = SamplePacket(packet, np.arange(ts_start, ts_end, 1 / fs))
            analysis.handle_sample_packet(signal_packet)
            if tags:
                if ts_end > tags[0].start_timestamp >= ts_start:
                    tag_to_use = tags.pop(0)
                    analysis.handle_tag_msg(tag_to_use)

            if VISUAL_DEBUG:
                time.sleep(1 / fs * packet_size / 10)

        yield analysis, analysis_read_test
    finally:
        analysis.close()
        analysis_read_test.close()


def test_ssvep_analysis(prepare_analysis_test_data, tmpdir):
    analysis, analysis_read_test = prepare_analysis_test_data
    for f in analysis.stimulation_freqs:
        assert analysis._classifier.thresholds[f].accuracy_diff >= 0.9
        assert analysis._classifier.thresholds[f].accuracy_abs >= 0.9
    analysis_save_file = os.path.join(str(tmpdir), 'classifier.dump')
    analysis.save(analysis_save_file)
    analysis_read_test.load_wisdom(analysis_save_file)
    assert analysis._classifier.thresholds == analysis_read_test._classifier.thresholds
