# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import os
import shutil
import sys
from collections import OrderedDict
from pathlib import Path

from fabric import api
from fabric.context_managers import lcd

try:
    import devops  # noqa
except ImportError:
    print("devops module not found! Not all commands are available!\n"
          "Install devops module using https://pip.braintech.pl/obci/stable", file=sys.stderr)
else:
    from devops.ci import install
    from devops.ci import output
    from devops import devpi

    sys.stdout = output.unbuffered_stdout

APT_REQUIREMENTS_PACKAGES = {
    'common': ['git', 'python3-pip'],
    'codestyle': ([]),
    'test': install.COMMON_APT_DEPENDENCIES[install.Stage.TEST] + [
        # obci dependencies
        'portaudio19-dev',
        # speller and drone control
        'libavdevice-dev',
        'libavfilter-dev',
        'libopus-dev',
        'libvpx-dev',
        'pkg-config',
    ],
}
PIP_REQUIREMENTS = {
    'build': ['setuptools',
              'PySide2'],
    'codestyle': [
        'flake8~=3.0.4',
        'flake8_docstrings~=1.0.2',
        'pydocstyle~=1.0.0',
    ],
    'test': ['pytest~=3.0',
             'pytest-timeout~=1.0',
             'pytest-qt~=2.4',
             'pandas~=0.19',
             'pytest-runner>=2.9',
             'pyqt-distutils~=0.7.3',
             ]
}

PACKAGES = OrderedDict([
    ('bci-panel', 'bci-panel'),
    ('obci-ugm', 'obci-ugm'),
    ('obci-p300', 'obci-p300'),
    ('bci-ssvep', 'bci-ssvep'),
    ('bci-artifacts', 'bci-artifacts'),
    ('bci-study', 'bci-study'),
    ('obci-demo', 'obci-demo'),
    ('bci-demo-ssvep', 'bci-demo-ssvep'),
    ('bci-speller', 'speller'),
    ('bci-lab', 'bci-lab')

])


def install_web_bci_panel():
    install_yarn()
    build_web_bci_panel()


def install_yarn():
    yarn = shutil.which('yarn')
    if yarn:
        print("Yarn installed in ", yarn)
        return
    api.local("curl -sL https://deb.nodesource.com/setup_8.x | sudo -E bash -")
    api.local("curl -sS https://dl.yarnpkg.com/debian/pubkey.gpg | sudo apt-key add -")
    api.local('echo "deb https://dl.yarnpkg.com/debian/ stable main" | sudo tee /etc/apt/sources.list.d/yarn.list')
    api.local("sudo apt-get --allow-unauthenticated update && sudo apt-get install -y --allow-unauthenticated yarn")


def start_webserver():
    with lcd('bci-panel'):
        api.local('yarn run dev')


def build_web_bci_panel():
    with lcd('bci-panel'):
        api.local('yarn')
        api.local('yarn run build')


def build_speller():
    with lcd('speller'):
        api.local("pip3 install -e .")


def start_production_server():
    with lcd('bci-panel'):
        api.local('yarn run serv')


def setup_stage(name):
    api.local('curl -sS https://dl.yarnpkg.com/debian/pubkey.gpg | apt-key add -')
    api.local('pip install -U pip')
    install.setup_stage(name, APT_REQUIREMENTS_PACKAGES, PIP_REQUIREMENTS)
    if name == 'test':
        for package, path in PACKAGES.items():
            if os.path.exists('dist'):  # artifacts from build job
                install.local("pip3 install dist/%s*.tar.gz" % package)
            else:
                install.local("pip3 install -e ./%s/" % path)
        install.local('install_all')


def generate_docker_file(path):
    docker_contents = (Path(__file__).parent / 'scripts' / 'ci' / 'Dockerfile_template').read_text('utf-8')
    aggregator = install.RequirementsAggregator()
    for reqs in APT_REQUIREMENTS_PACKAGES.values():
        aggregator.collect([(install.APT, r) for r in reqs])
    for reqs in PIP_REQUIREMENTS.values():
        aggregator.collect([(install.PIP3, r) for r in reqs])
    uninstall = ['braintech-obci-core', 'braintech-obci-signal-processing',
                 'braintech-obci-lab', 'devops', 'braintech-utils']
    for package, package_path in PACKAGES.items():
        aggregator.collect([(install.PIP3, package)])
        uninstall.append(package)
    docker_contents += """
RUN apt-get -qq update && apt-get install -y {apt_requirements}
RUN pip3 install --index {devpi_index} devops {pip3_requirements}\\
    && install_all \\
    && rm -rf /var/lib/apt/lists/*\\
    && pip3 uninstall -y {uninstall}
    """.format(
        apt_requirements=aggregator.flatten(install.APT, "\\\n\t"),
        pip3_requirements=aggregator.flatten(install.PIP3, "\\\n\t"),
        devpi_index=os.environ.get('DEVPI_URL', '') + '/' + devpi.current_index()['name'],
        uninstall=" ".join(uninstall),
    )

    Path(path).write_text(docker_contents, 'utf-8')


def install_editable():
    for package, dir in PACKAGES.items():
        install.local('pip3 install -e %s/.' % dir)


def manual_upload():
    for package, dir in PACKAGES.items():
        with api.lcd(dir):
            api.local('devpi_upload')


def install_packages():
    for package, dir in PACKAGES.items():
        install.local('pip3 install %s' % package)


def build_packages():
    Path('dist').mkdir(exist_ok=True)
    for package, dir in PACKAGES.items():
        print("Building package", package)
        devpi.upload(dir)
        if dir != '.':
            api.local('mv %s/dist/* ./dist/' % dir)


def deploy_packages():
    for package, dir in PACKAGES.items():
        with lcd(dir):
            api.local("devpi_push --from_tmp")
