import functools
import logging
import os

from braintech.obci.core.broker.message_handler_mixin import subscribe_message_handler
from braintech.obci.core.broker.messages import BaseMessage, Field
from braintech.obci.experiment.driver_utils import driver_discovery
from braintech.obci.core.drivers.eeg import lsl_amplifier
from braintech.obci.core.drivers.eeg import openbci_amplifier
from braintech.obci.core.drivers.eeg import random_amplifier
from braintech.obci.experiment.peers.drivers.amplifiers import amplifier_peer
from obci_p300.common.random_amp_with_sine.random_amplifier_peer_with_sine_generator import \
    BaseRandomAmplifierWithSinePeer

logger = logging.getLogger(__name__)


class AmplifierWithChosenIdMissing(Exception):
    pass


class SignalDescription(BaseMessage):
    sampling_rate = Field(float)
    name = Field(str)
    channels = Field(list)


class SignalDescriptionQuery(BaseMessage):
    pass


AMP_CLASSES = {}


def register_demo_amplifier(name):
    def _register(klass, amp_name=name):
        AMP_CLASSES[amp_name] = klass
        return klass

    if isinstance(name, str):
        return _register
    else:
        klass_ = name
        return _register(klass_, klass_.__name__)


class PreconfiguredBaseAmplifierPeer(amplifier_peer.BaseAmplifierPeer):

    def _set_static_parameters(self, config):
        self.amplifier_id = config['amplifier_id']
        self.channel_names = config['channel_names']
        self.active_channels = self.channel_names
        self.channel_gains = config['gains']
        self.channel_offsets = config['offsets']
        self._sampling_rate = config['sampling_rate']
        self._sampling_rates = [config['sampling_rate']]
        self.samples_per_packet = config['samples_per_packet']
        self.sample_type = config['sample_type']
        self.channels_info = config['channels_info']

    @subscribe_message_handler(SignalDescriptionQuery)
    def handle_signal_description_query(self, _msg):
        return SignalDescription(
            sampling_rate=self.sampling_rate,
            name=str(self._amplifier),
            channels=[
                {'name': name,
                 'gain': gain,
                 'offset': offset} for name, gain, offset in
                zip(self.channel_names, self.channel_gains, self.channel_offsets)
            ]
        )


@register_demo_amplifier
class RandomAmplifierPeer(PreconfiguredBaseAmplifierPeer):
    AmplifierClass = random_amplifier.RandomAmplifier


class OpenBciAmplifierPeer(PreconfiguredBaseAmplifierPeer):
    AmplifierClass = openbci_amplifier.OpenBciAmplifier


@register_demo_amplifier
class RandomAmplifierWithSinePeer(PreconfiguredBaseAmplifierPeer, BaseRandomAmplifierWithSinePeer):
    pass


@register_demo_amplifier
class LslAmplifierAdapterPeer(PreconfiguredBaseAmplifierPeer):
    AmplifierClass = lsl_amplifier.LSLAmplifierAdapter


# @formatter:off
try:
    import braintech.drivers.native_amplifier_lib.dummy_amplifier

    @register_demo_amplifier
    class DummyAmplifierPeer(PreconfiguredBaseAmplifierPeer):
        AmplifierClass = braintech.drivers.native_amplifier_lib.dummy_amplifier.DummyCppBaseAmplifier
except ImportError:
    pass

try:
    from braintech.drivers.perun8.peers.perun_amplifier_peer import *  # noqa
    import braintech.drivers.perun8.amplifiers

    @register_demo_amplifier
    class PerunAmplifierPeer(PreconfiguredBaseAmplifierPeer):
        AmplifierClass = braintech.drivers.perun8.amplifiers.PerunCppAmplifier
except ImportError:
    pass

try:
    from braintech.drivers.tmsi.peers.tmsi_amplifier_peer import *  # noqa
    from braintech.drivers.tmsi.amplifiers import TmsiCppAmplifier

    @register_demo_amplifier
    class TmsiAmplifierPeer(PreconfiguredBaseAmplifierPeer):
        AmplifierClass = TmsiCppAmplifier
except ImportError as e:
    pass

try:
    from braintech.drivers.perun32.amplifier import Perun32Amplifier
    from braintech.drivers.perun32.perun32_peer import *  # noqa

    @register_demo_amplifier
    class Perun32AmplifierPeer(PreconfiguredBaseAmplifierPeer):
        AmplifierClass = Perun32Amplifier
except ImportError:
    import traceback
    traceback.print_exc()
    pass
# @formatter:on


def _get_description(amplifier):
    params = amplifier['amplifier_params']
    return dict(
        amplifier_id=params['additional_params']['amplifier_id'],
        channels=[ch['name'] for ch in params['channels_info']['channels']],
        gains=[ch['gain'] for ch in params['channels_info']['channels']],
        offsets=[ch['offset'] for ch in params['channels_info']['channels']],
        sampling_rates=params['channels_info']['sampling_rates'],
        name=params['channels_info']['name'],
        class_name=amplifier['amplifier_peer_info']['path'].split('.')[-1]
    )


@functools.lru_cache(maxsize=1)
def get_available_amplifiers():
    logger.info('Searching for amplifiers...')
    descriptions = []
    amp_types = os.environ.get("AMPLIFIER_TYPES", 'virtual,usb').split(',')
    for amp_type in amp_types:
        descriptions.extend([_get_description(d) for d in
                             driver_discovery.find_amplifiers(device_type=amp_type)])
    logger.info('{} amplifiers found.'.format(len(descriptions)))
    ret = {d['amplifier_id']: d for d in descriptions if d['class_name'] in AMP_CLASSES}
    return ret


def get_amplifier_class(amplifier_class_name):
    return AMP_CLASSES[amplifier_class_name]
