import zmq

import obci_p300.common.messages
from braintech.obci.core.broker.broker import Broker
from braintech.obci.core.broker.message_handler_mixin import subscribe_message_handler
from braintech.obci.core.utils import message_helpers
from obci_demo import blinking
from obci_demo.p300 import P300Message
from . import settings


def is_any_running():
    ctx = zmq.Context()
    socket = ctx.socket(zmq.REP)
    try:
        socket.bind(settings.BROKER_REP)
    except zmq.error.ZMQError as e:
        message = str(e)
        if message == 'Address already in use':
            return True
        else:
            raise
    else:
        socket.unbind(settings.BROKER_REP)
        return False


class P300GuiBroker(Broker):
    def __init__(self, config, experiment, *args, **kwargs):
        self._blinker = self._create_blinker(config)
        self._experiment = experiment
        super().__init__(*args, **kwargs)

    def _create_blinker(self, config):
        return blinking.P300Blinker(config, self)

    def _set_gui_message(self, text):
        self._logger.info('Setting GUI text: %s', text)

    @subscribe_message_handler(P300Message)
    def handle_p300_blinking_message(self, msg: P300Message):
        if msg.key == 'start_blinking':
            self._logger.info('Start GUI blinking.')
            self._blinker.start_blinking()
        elif msg.key == 'stop_blinking':
            self._logger.info('Stop GUI blinking.')
            self._blinker.stop_blinking()
        elif msg.key == 'instruction':
            self._set_gui_message(msg.value)
        elif msg.key == 'clear':
            self._blinker.clear_gui()

    def send_blink(self, blink_id, blink_time):
        task = message_helpers.send_tag(
            peer=self,
            p_start_timestamp=blink_time,
            p_end_timestamp=blink_time,
            p_tag_name='blink',
            p_tag_desc={'stimuli_id': blink_id},
            p_tag_channels='',
        )
        self.create_task(task)

    @subscribe_message_handler(obci_p300.common.messages.DecisionVerboseMsg)
    def handle_p300_decision_message(self, message: obci_p300.common.messages.DecisionVerboseMsg):
        choice = int(message.score)
        self._logger.info('Got decision to select in GUI: %s', choice)

        is_learning = message.decision_type == 'LEARN'
        if is_learning:
            targets = int(message.additional_info['n_targets'])
            auc = float(message.additional_info['roc_auc'])
            finish = targets >= 40 or (targets >= 15 and auc >= 0.95)
            self._logger.info('Calibration finish (AUC > 0.95 and targets >= 15 or targets >= 40)')
            if finish:
                self._experiment.show_calibration_finished()
