# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved
import pathlib
import pickle
import threading
from concurrent import futures
from concurrent.futures import ThreadPoolExecutor
from functools import wraps
from logging import getLogger

from bci_panel.web_bci_panel import WebBCIPanelPeer
from braintech.obci.core.broker.base_peer import PeerState
from braintech.obci.experiment.peers.drivers.amplifiers.amplifier_peer import BaseAmplifierPeer  # noqa:F401
from obci_demo import p300, config_generation
from obci_demo.amplifiers import get_amplifier_class
from obci_demo.brokers import P300GuiBroker
from obci_demo.simple_experiment import SimpleExperiment

LOGGER = getLogger(__name__)
AMPLIFIER_ID = 'amplifier'
ANALYSIS_ID = 'analysis'
LOGIC_ID = 'logic'


def synchronized(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        with self._lock:
            return func(self, *args, **kwargs)

    return wrapper


class InvalidState(Exception):
    pass


class OBCIDemoExperiment(SimpleExperiment):
    BrokerClass = P300GuiBroker

    def _create_broker(self):
        return self.BrokerClass(self._config['p300']['blinker'], self)

    def __init__(self, config, subprocess=True):
        self._config = config
        super(OBCIDemoExperiment, self).__init__()
        self.bci_panel = self._create_bci_panel()
        self.bci_panel.run_in_thread()
        self._running_calibration = False
        self.calibration_done = False
        self._amplifier = None
        self._analysis = None
        self._logic = None
        self._lock = threading.Lock()
        self._as_subprocess = subprocess
        self._executor = ThreadPoolExecutor()
        self._main_peers = []

    @synchronized
    def start(self, config):
        if self._amplifier:
            raise InvalidState()
        self._config = config
        amplifier_config = config['amplifier']
        AmplifierClass = get_amplifier_class(amplifier_config['amplifier_class'])
        peer_features = []
        peer_features.append(
            self._executor.submit(self._add_peer, self._as_subprocess, AmplifierClass, peer_id=AMPLIFIER_ID,
                                  config=amplifier_config))
        peer_features.append(self._executor.submit(self._create_analysis, config))
        peer_features.append(self._executor.submit(self._create_logic, config))
        futures.wait(peer_features)
        peers = [f.result() for f in peer_features]
        self._main_peers = list(filter(None, peers))
        self.wait_for(self._main_peers, PeerState.ready)

        self._amplifier = peers[0]  # type:BaseAmplifierPeer
        self._analysis = peers[1]  # type:p300.PreconfiguredP300AnalysisPeer
        self._logic = peers[2]  # type:p300.P300LogicPeer

    def _create_bci_panel(self):
        return self.add_peer(WebBCIPanelPeer, peer_id='bci_panel')  # type: WebBCIPanelPeer

    def _create_analysis(self, config):
        self._calibration_file_path = self._config['p300']['wisdom_path']
        config['p300']['wisdom_path'] = self._calibration_file_path
        config['p300']['wisdom_load'] = pathlib.Path(self._calibration_file_path).exists()
        return self._add_peer(self._as_subprocess, p300.PreconfiguredP300AnalysisPeer,
                              peer_id=ANALYSIS_ID, config=config['p300'],
                              amplifier_config=config['amplifier'])

    def _create_logic(self, config):
        return self.add_peer(p300.P300LogicPeer, peer_id=LOGIC_ID,
                             config=config)

    @synchronized
    def stop(self):
        for p in self._main_peers:
            self.remove_peer(p)
        self._analysis = None
        self._amplifier = None
        self._logic = None

    @synchronized
    def start_calibration(self):
        if self._amplifier is None:
            raise InvalidState()
        self._analysis.set_calibration(True)
        if self._logic:
            self._logic.start_calibration()
        self._running_calibration = True

    @synchronized
    def start_bci(self):
        if self._amplifier is None:
            raise InvalidState()
        self._analysis.set_calibration(False)
        if self._logic:
            self._logic.start_bci()

    @synchronized
    def finish(self):
        if self._logic:
            self._logic.stop()
        if self._running_calibration:
            self._running_calibration = False
            if self.is_calibrated():
                LOGGER.info("Calibration done successfully!")
            else:
                LOGGER.info("Calibration not finished!")

    def is_calibrated(self):
        p300_wisdom_file = pathlib.Path(self._calibration_file_path)
        if p300_wisdom_file.exists():
            try:
                with p300_wisdom_file.open('rb') as f:
                    p300_easy_classifier = pickle.load(f)
            except EOFError:
                return False
            else:
                return p300_easy_classifier.calibrated
        else:
            return False

    def calibration_finished(self):
        LOGGER.info('Calibration finished! You can stop the calibration now.')

    def key_pressed(self, key_name):
        if self._logic:
            self._logic.key_pressed(key_name)

    def generate_config(self, config):
        return config_generation.generate(
            wisdom_path=self._config['p300']['wisdom_path'],
            **config
        )
