# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved
import time

import pytest
from socketIO_client import SocketIO

from bci_panel.test_utils import TestNamespace, process_socket
from bci_panel.web_bci_panel import RemoteWebBCIPanel, WebBCIPanelPeer, CHARTS
from braintech.obci.core.broker.base_peer import PeerState
from braintech.obci.core.broker.peer import Peer
from obci_demo import config_generation
from obci_demo.amplifiers import RandomAmplifierPeer
from obci_demo.obci_demo_experiment import OBCIDemoExperiment
from obci_demo.simple_experiment import SimpleExperiment


class BinarySocketIO(SocketIO):
    def _process_packet(self, packet):
        try:
            super(BinarySocketIO, self)._process_packet(packet)
        except ValueError:
            pass

    def _process_packets(self):
        try:
            super(BinarySocketIO, self)._process_packets()
        except ValueError:
            pass


@pytest.fixture()
def experiment():
    exp = SimpleExperiment()
    yield exp
    exp.shutdown()


class _TestCrash(Exception):
    pass


class CrashingPeer(Peer):

    async def _async_crash(self):
        raise _TestCrash

    def async_crash(self):
        self.create_task(self._async_crash())

    def sync_crash(self):
        raise _TestCrash

    def finish_with_panic(self):
        self.panic(_TestCrash())


def test_simple_experiment(experiment):
    config = {
        'amplifier_id': 'random',
        'channel_names': [],
        'gains': [],
        'offsets': [],
        'sampling_rate': 256,
        'samples_per_packet': 64,
        'sample_type': 'FLOAT',
        'channels_info': None,
    }
    p1 = experiment.add_peer(RandomAmplifierPeer, peer_id="amp_1", config=config)
    p2 = experiment.add_subprocess_peer(RandomAmplifierPeer, peer_id="amp_2", config=config)
    crashing = experiment.add_subprocess_peer(CrashingPeer, peer_id="crashing")
    assert p2.id == 'amp_2'
    experiment.wait_for([p1, p2, crashing], PeerState.ready)
    assert p1.is_running and p2.is_running
    with pytest.raises(AttributeError):
        p2.bad_attribute()
    p2.stop_sampling()
    with pytest.raises(_TestCrash):
        crashing.sync_crash()
    crashing.async_crash()
    crashing.finish_with_panic()
    experiment.remove_peer(crashing)
    crashing = experiment.add_subprocess_peer(CrashingPeer, peer_id="crashing")
    experiment.remove_peer(crashing)
    experiment.remove_peer(p2)


def test_web_bci_panel_peer(experiment: SimpleExperiment):
    panel = RemoteWebBCIPanel(experiment.broker)
    bci_panel_peer = experiment.add_peer(WebBCIPanelPeer, peer_id='bci_panel')  # type: WebBCIPanelPeer
    bci_panel_peer.run_in_thread()
    while not bci_panel_peer.is_ready:
        time.sleep(0.1)
    socket = SocketIO(bci_panel_peer.server_host, bci_panel_peer.server_port)
    test_namespace = socket.define(TestNamespace, CHARTS)  # type:TestNamespace
    process_socket(socket)
    test_namespace.connected.wait()
    panel.update_text(feature='AUC', value=1.0)
    test_namespace.received_data.wait()
    assert test_namespace.event[0] == 'AUC'


MONTAGE_CHANNELS = ["Random1", "Random2"]
ANALYSIS_CHANNELS = ["Random3", "Random4", "Random5"]


class BCIGui():
    def __init__(self, peer):
        self._peer = peer


def test_demo_experiment():
    config = config_generation.generate(
        amplifier_id='RandomAmplifier',
        analysis_channels=ANALYSIS_CHANNELS,
        montage_channels=MONTAGE_CHANNELS,
        exclude_from_montage=ANALYSIS_CHANNELS[-1:],
        sampling_rate=128,
        filter_frequency=50,
        wisdom_path=None
    )
    config['amplifier']['amplifier_class'] = 'RandomAmplifierWithSinePeer'
    experiment = OBCIDemoExperiment(config, True)
    bci_panel_peer = experiment.bci_panel
    socket = BinarySocketIO(bci_panel_peer.server_host, bci_panel_peer.server_port)
    test_namespace = socket.define(TestNamespace, CHARTS)  # type:TestNamespace
    process_socket(socket)
    test_namespace.connected.wait()
    experiment.start(config)
    test_namespace.received_data.wait()
    while test_namespace.event[0] != 'impedances':
        test_namespace.received_data.clear()
        test_namespace.received_data.wait()
    experiment.start_calibration()
    experiment.finish()
    experiment.start_bci()
    experiment.finish()
    experiment.stop()
    experiment.start(config)
