import logging
import time
from collections import defaultdict, deque
from collections import namedtuple

import numpy as np

from braintech.obci.signal_processing.buffers.auto_blink_buffer import AutoBlinkBuffer
from .p300_classm import P300EasyClassifier, feature_extraction_singular


class AnalysisStateVariables:
    def __init__(self):
        self.calibration_targets_buffor = []
        self.calibration_nontargets_buffor = []
        self.last_decision_timestamp = 0

        self.clear_visualisations_planned = False
        self.bci_panel_visualiser = None  # type: BCIPanelVisualiser
        self.signal_visualiser = None  # type: SignalVisualiser
        self.signal_fft_visualiser = None  # type: SignalSpectrumVisualiser
        self.signal_impedance_visualiser = None  # type: SignalImpedanceVisualiser

        self.eeg_artifact_detector = None  # type: EEGArtifactsDetector
        self.camera_aggregator = None  # type: aggregators.StateAggregator

    def reset_buffers(self, decision_stop: int):
        # probabilities of selected input for single epochs
        self.singular_attempts_buffer = defaultdict(list)
        # probabilities of selected input for cumulative mean
        self.averaged_attempts_buffer = defaultdict(list)
        # probabilities of selected input for weighted cumulative mean
        self.weighted_averaged_attempts_buffer = defaultdict(list)
        # features buffor to classify on single epochs or cumulative mean
        self.features = defaultdict(list)
        # epochs buffer - mainly for visualisation
        self.epochs = defaultdict(list)
        # final decision buffor
        self.decision_buffer = deque(maxlen=decision_stop)


class BaseP300Analysis:
    def __init__(self, baseline, window, wisdom_path,
                 offline_learning_dataset_path, channel_names,
                 channels_for_classification, learn_every, targetFs,
                 sampling_rate, use_linear_proba, ignored_blink_ids,
                 training_index, state: AnalysisStateVariables,
                 decision_stop,
                 decision_callback,
                 cv_start_callback,
                 proba_used_for_internal_decision,
                 allowed_classifier_FPR, maximum_to_average):
        self._logger = logging.getLogger(self.__class__.__name__)
        self.baseline = baseline
        self.window = window
        self.wisdom_path = wisdom_path
        self.offline_learning_dataset_path = offline_learning_dataset_path
        self.channel_names = channel_names
        self.channels_for_classification = channels_for_classification
        self.learn_every = learn_every
        self.targetFs = targetFs
        self.sampling_rate = sampling_rate
        self.use_linear_proba = use_linear_proba
        self.ignored_blink_ids = ignored_blink_ids
        self.training_index = training_index
        self.state = state
        self.decision_stop = decision_stop
        self._decision_callback = decision_callback
        self._cv_start_callback = cv_start_callback
        self.proba_used_for_internal_decision = proba_used_for_internal_decision
        self.allowed_classifier_FPR = allowed_classifier_FPR
        self.maximum_to_average = maximum_to_average

    @staticmethod
    def blink_converter_function(t):
        raise NotImplementedError

    def add_result(self, blink, probabilities, classifier):
        raise NotImplementedError

    def create_buffer(self, channel_count: int, ret_func):
        raise NotImplementedError

    def create_classifier(self):
        raise NotImplementedError

    def identify_blink(self, blink):
        raise NotImplementedError

    def classify(self, classifier, chunk, blink):
        raise NotImplementedError

    def learn(self, classifier, chunk, target, blink):
        raise NotImplementedError


class P300Analysis(BaseP300Analysis):

    def create_buffer(self, channel_count, ret_func):
        return AutoBlinkBuffer(
            from_blink=int(self.sampling_rate * self.baseline),
            samples_count=int((self.window - self.baseline) * self.sampling_rate),
            sampling=self.sampling_rate,
            num_of_channels=channel_count,
            ret_func=ret_func,
            copy_on_ret=True,
        )

    def create_classifier(self) -> P300EasyClassifier:
        self._logger.info('Creating new, untrained classifier')
        return P300EasyClassifier(learn_every=self.learn_every,
                                  targetFs=self.targetFs,
                                  use_linear_proba=self.use_linear_proba)

    @staticmethod
    def blink_converter_function(tag):
        return int(tag['desc']['stimuli_id'])

    def identify_blink(self, blink):
        blink_in_ignored = int(blink.index) in self.ignored_blink_ids
        if self.in_training_mode and not blink_in_ignored:
            if int(blink.index) == self.training_index:
                return 'target'
            else:
                return 'nontarget'
        else:
            return None

    @property
    def in_training_mode(self):
        return self.training_index is not None

    def _prepare_chunk(self, chunk, blink):
        """
        Performs feature extraction.

        :param chunk: numpy 2D data array (channels × samples)
        """
        chunk_ready = feature_extraction_singular(
            chunk,
            self.sampling_rate,
            self.baseline,
            self.window,
            self.targetFs,
        )
        return chunk_ready

    def classify(self, classifier, chunk, blink):
        """Compute set of classification probabilities for a given blink.

        May be re-implemented in subclass to perform some pre-processing
        or feature extraction on the signal (chunk).
        This method will be run by a separate thread.

        :param (AbstractClassifier): classifier instance to be used for classification
        :param chunk: numpy 2D data array (channels × samples)
        :param blink: information contained in a single blink

        Returns:
            dict. dictionary of {target: probability},
            or None if classification could not be performed

        """

        blink_in_ignored_blinks = int(blink.index) in self.ignored_blink_ids

        if blink_in_ignored_blinks:
            return None
        chunk_ready = self._prepare_chunk(chunk, blink)
        self.state.features[blink.index].append(chunk_ready)
        self.state.epochs[blink.index].append(chunk)

        probabilities = {}
        probabilities['targetSingle'] = classifier.classify(chunk_ready)
        chunk_mean = np.array(self.state.features[blink.index]).mean(axis=0)
        probabilities['targetCMean'] = classifier.classify(chunk_mean)

        weights = self.get_decision_averaging_weights(blink.index)
        chunk_weighted_mean = np.average(self.state.features[blink.index], axis=0, weights=weights)
        probabilities['targetWCMean'] = classifier.classify(chunk_weighted_mean)

        if self.state.clear_visualisations_planned:
            self.state.bci_panel_visualiser.clear(
                features_ids=self.state.features.keys(),
            )
            self.state.clear_visualisations_planned = False

        self.state.bci_panel_visualiser.visualise_classification(
            features=self.state.features,
            epochs=self.state.epochs,
            index=blink.index,
        )
        return probabilities

    def get_decision_averaging_weights(self, index):
        feature_len = len(self.state.features[index])
        weights = []
        for i in range(feature_len):
            if i < self.decision_stop:
                weights.append(1.0)
            else:
                renumbered = i - self.decision_stop + 1
                weight = 1 / (2 ** renumbered)
                weights.append(weight)
        return list(reversed(weights))

    def _run_through_signal_goodness_detectors(self, blink):
        return True

    def add_result(self, blink, probabilities, classifier):
        """
        Sends decision if some last decisions from cumulative mean are the same or a lot of averaging was done.

        :param blink: information contained in a single blink
        :param probabilities 'dict': dictionary of {target: probability}
        """
        # planning for future
        if probabilities:
            self.state.singular_attempts_buffer[blink.index].append(probabilities['targetSingle'])
            self.state.averaged_attempts_buffer[blink.index].append(probabilities['targetCMean'])
            self.state.weighted_averaged_attempts_buffer[blink.index].append(probabilities['targetWCMean'])
            last_single = [blink.index, probabilities['targetSingle']]
            last_mean = [blink.index, probabilities['targetCMean']]
            last_weighted_mean = [blink.index, probabilities['targetWCMean']]
            self._logger.info('Last single dec: {}, proba: {:.2f}'
                              .format(last_single[0], last_single[1]))
            self._logger.info('Last mean dec: {}, proba: {:.2f}'
                              .format(last_mean[0], last_mean[1]))
            self._logger.info('Last weighted mean dec: {}, proba: {:.2f}'
                              .format(last_weighted_mean[0],
                                      last_weighted_mean[1]))

            decision_timestamp = time.time()
            common_decision_info = {
                'start_timestamp': decision_timestamp,
                'end_timestamp': decision_timestamp,
                'additional_info': {
                    'blink_counts': {k: len(v) for k, v in self.state.singular_attempts_buffer.items()},
                    'singular_proba': self.state.singular_attempts_buffer,
                    'averaged_proba': self.state.averaged_attempts_buffer,
                    'weighted_proba': self.state.weighted_averaged_attempts_buffer,
                }
            }

            used_for_dec = {
                'single': last_single,
                'mean': last_mean,
                'weighted_mean': last_weighted_mean,
            }[self.proba_used_for_internal_decision]

            if used_for_dec[1] > classifier.threshold_for_fpr(self.allowed_classifier_FPR):
                self.state.decision_buffer.append(blink.index)

            self.state.bci_panel_visualiser.visualise_classification_confidence(
                classifier=classifier,
                allowed_classifier_FPR=self.allowed_classifier_FPR,
                singular_attempts=self.state.singular_attempts_buffer,
                average_attempts=self.state.averaged_attempts_buffer,
                weighted_averaged_attempts=self.state.weighted_averaged_attempts_buffer,
            )

            # enough of the same decisions condition
            one_decision = (len(set(self.state.decision_buffer)) == 1)
            buffer_full = (len(self.state.decision_buffer) == self.decision_stop)
            is_enough_of_the_same_decision = one_decision and buffer_full
            # number of averaged epochs condition
            # ensure all buttons have been averaged self.maximum_to_average number of time
            minimum_averaged = min(len(self.state.averaged_attempts_buffer[i])
                                   for i in self.state.averaged_attempts_buffer.keys())
            is_decided_by_average = minimum_averaged > self.maximum_to_average
            if is_enough_of_the_same_decision:
                self._send_decision_by_full_support(common_decision_info)
            elif is_decided_by_average:
                self._send_decision_by_average(common_decision_info, classifier)

    def _send_decision_by_full_support(self, common_decision_info):
        decision = self.state.decision_buffer[-1]
        self._logger.info('Decision by decision_stop {}'.format(decision))
        msg_decision = {'decision': 'CLASSIFICATION',
                        'decision_type': 'one_decision',
                        'decision_averaging_type': self.proba_used_for_internal_decision,
                        'score': decision,
                        **common_decision_info}
        self._send_decision(msg_decision, plan_bci_vis_clearing=True)

    def _send_decision_by_average(self, common_decision_info, classifier):
        most_confident_decision, confidences = max(self.state.averaged_attempts_buffer.items(),
                                                   key=lambda key: key[1][-1])
        self._logger.info('Decision by max_avr {}'.format(most_confident_decision))
        if confidences[-1] > classifier.threshold_for_fpr(self.allowed_classifier_FPR):
            decision_type = 'minimum_averaged_sure'
        else:
            decision_type = 'minimum_averaged_not_sure'
        msg_decision = {'decision': 'CLASSIFICATION',
                        'decision_type': decision_type,
                        'decision_averaging_type': 'mean',
                        'score': most_confident_decision,
                        **common_decision_info}
        self._send_decision(msg_decision, plan_bci_vis_clearing=True)

    def learn(self, classifier, chunk, target, blink):
        """Learn that given chunk represents given target.

        May be re-implemented in subclass to perform some pre-processing
        or feature extraction on the signal (chunk).
        This method will be run by a separate thread.

        :param classifier (AbstractClassifier): classifier instance to be used for learning
        :param chunk: numpy 2D data array (channels × samples)
        :param target: name of the target
        """
        chunk_ready = self._prepare_chunk(chunk, blink)
        if target == 'target':
            self.state.calibration_targets_buffor.append(chunk)
        elif target == 'nontarget':
            self.state.calibration_nontargets_buffor.append(chunk)

        score = classifier.learn(chunk_ready, target, self._cv_start_callback)
        self.state.bci_panel_visualiser.visualise_calibration(
            classifier=classifier,
            targets=self.state.calibration_targets_buffor,
            non_targets=self.state.calibration_nontargets_buffor,
        )

        if score is not None:
            self._logger.info('Classifier is calibrated.')
            now = time.time()
            decision = {
                'decision': 'DONE LEARNING',
                'score': score,
                'decision_type': 'LEARN',
                'start_timestamp': now,
                'end_timestamp': now,
                'additional_info': {
                    'n_targets': int(classifier.history_n[-1]),
                    'roc_auc': float(classifier.roc_auc),
                },
            }
            self.state.bci_panel_visualiser.update_plots(classifier, self.allowed_classifier_FPR)
            self._send_decision(decision)

    def _send_decision(self, decision, plan_bci_vis_clearing=False):
        self.state.last_decision_timestamp = time.time()
        if self._decision_callback:
            self._decision_callback(**decision)
        if decision['decision'] != "DONE LEARNING":
            self.state.bci_panel_visualiser.save_plots(dir_name=str(int(self.state.last_decision_timestamp)))
        self.clear(plan_bci_vis_clearing)

    def clear(self, plan_bci_vis_clearing):
        if plan_bci_vis_clearing:
            self.state.clear_visualisations_planned = True
        self.state.reset_buffers(self.decision_stop)


Band = namedtuple('Band', ['center', 'width'])
