# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import time
from typing import Optional

import numpy
from numpy.matlib import repmat

from braintech.obci.core.drivers.eeg.eeg_amplifier import AmplifierDescription
from braintech.obci.core.drivers.eeg.random_amplifier import RandomAmplifier
from braintech.obci.signal_processing.signal.containers import SamplePacket


class RandomAmplifierWithSine(RandomAmplifier):
    name = "RandomAmplifierWithSine"
    _description = AmplifierDescription(name, AmplifierDescription.ALL, AmplifierDescription.UNKNOWN)

    @classmethod
    def get_description(cls, id: Optional[str] = None) -> AmplifierDescription:
        if id is None or id == cls.name:
            return cls()._description
        raise Exception('Invalid amplifier id.')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sine_generator_freq = 0
        self._sine_generator_amp = 10
        self._noise_std = 1
        self._p300_amplitude = 0.5
        self.p300_queue = numpy.array([], ndmin=1)

    def set_sine_freq(self, f):
        self._sine_generator_freq = f

    def set_noise_std(self, std):
        if std > 0:
            self._noise_std = std

    def set_sine_amplitude(self, a):
        self._sine_generator_amp = a

    def set_p300_amplitude(self, a):
        self._p300_amplitude = a

    def generate_p300(self):
        self.p300_queue = self._p300_amplitude * numpy.hanning(0.3 * self._sampling_rate)

    def _get_samples(self, samples_per_packet) -> SamplePacket:
        all_channel_count = len(self._description.channels)
        assert all_channel_count >= 2

        t0 = time.time()
        ts = numpy.linspace(t0, t0 + 1 / self.sampling_rate * samples_per_packet, samples_per_packet)
        samples = numpy.random.normal(loc=0, scale=self._noise_std, size=(samples_per_packet, all_channel_count))

        p300_part = self.p300_queue[:samples_per_packet]
        self.p300_queue = self.p300_queue[samples_per_packet:]

        if p300_part.shape[0] < samples_per_packet:
            pad = samples_per_packet - p300_part.shape[0]
            p300_part = numpy.pad(p300_part, (0, pad), 'constant', constant_values=0)

        samples += p300_part[:, numpy.newaxis]
        if self._sine_generator_freq > 0:
            samples_ch = self._sine_generator_amp * numpy.sin(2 * numpy.pi * self._sine_generator_freq * ts)
            samples += repmat(samples_ch, all_channel_count, 1).T

        for i in range(samples_per_packet):
            samples[i][-2] = self.next_saw_value  # saw
            if self.next_saw_value == 100:
                self.next_saw_value = 0
            else:
                self.next_saw_value += 1
            self.last_counter_value += 1
            samples[i][-1] = self.last_counter_value  # sample counter
        samples = self._reindex_samples(samples)

        impedance = self._get_impedance(samples_per_packet)
        return SamplePacket(ts=ts, samples=samples, impedance=impedance)
