# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved
import numpy as np
from scipy.signal import iirnotch, cheby2

from braintech.obci.signal_processing.signal.containers import SamplePacket
from obci_p300.utils import ArrayCascadeFilter


class SignalPreprocessor:
    def __init__(self, all_channels_names,
                 classification_channel_names,
                 montage_channels_names,
                 sampling_rate,
                 excluded_channel_names=(),
                 notch_filters=None,
                 filters=None,
                 target_fs=None,
                 channels_gains=None,
                 channels_offsets=None):
        self._sampling_rate = sampling_rate
        self._prepare_montage(all_channels_names, classification_channel_names, montage_channels_names,
                              excluded_channel_names)
        if channels_gains is None:
            self.channels_gains = np.ones(len(all_channels_names))
        else:
            self.channels_gains = np.array(channels_gains)

        if channels_offsets is None:
            self.channels_offsets = np.zeros(len(all_channels_names))
        else:
            self.channels_offsets = np.array(channels_offsets)

        if filters:
            self.filters = filters
        elif notch_filters:
            self.filters = self._get_filters(notch_filters, target_fs or self._sampling_rate)
        else:
            self.filters = None

    def _prepare_montage(self, all_channels_names, classification_channel_names, montage_channels_names,
                         excluded_channel_names):
        if all_channels_names is None:
            all_channels_names = []
        if classification_channel_names is None:
            classification_channel_names = []
        if montage_channels_names is None:
            montage_channels_names = []
        if excluded_channel_names is None:
            excluded_channel_names = []
        self._all_channels_names = all_channels_names

        def _get_indexes(names):
            return [self._all_channels_names.index(name) for name in names]

        self._classification_channel_indexes = _get_indexes(classification_channel_names)
        self._montage_channel_indexes = _get_indexes(montage_channels_names)
        self._excluded_channel_indexes = _get_indexes(excluded_channel_names)
        self._montage_matrix = self._get_montage_matrix()

    def _get_montage_matrix(self):
        montage = [
        ]
        for i in self._classification_channel_indexes:
            row = np.zeros(len(self._all_channels_names))
            row[i] = 1
            if i not in self._excluded_channel_indexes:
                for m in self._montage_channel_indexes:
                    row[m] = -1.0 / len(self._montage_channel_indexes)
            montage.append(row)
        return np.array(montage)

    def _montage_data(self, raw_data):
        return np.dot(self._montage_matrix, raw_data.T).T

    def get_preprocessed_packet(self, raw_packet: SamplePacket):
        raw_data = raw_packet.samples
        data = self.get_preprocessed_signal(raw_data)
        packet = SamplePacket(data, raw_packet.ts, raw_packet.impedance)
        return packet

    def get_preprocessed_signal(self, raw_data):
        microvolt_data = self._convert_to_microvolts(raw_data)
        montaged_data = self._montage_data(microvolt_data)
        if self.filters:
            return self.filters.get(montaged_data)
        else:
            return montaged_data

    def _convert_to_microvolts(self, samples):
        return samples * self.channels_gains + self.channels_offsets

    def _get_filters(self, notch_filters, target_fs):
        filters = []
        nyquist_rate = self._sampling_rate / 2
        for freq in notch_filters:
            if nyquist_rate > freq:
                w0 = freq / nyquist_rate
                Q = w0 / (5 / nyquist_rate)
                b, a = iirnotch(w0, Q)
                filters.append({'b': b, 'a': a})

        if target_fs != self._sampling_rate:
            # target_fs filter:
            b, a = cheby2(2, 12, target_fs / nyquist_rate, btype='lowpass')
            filters.append({'b': b, 'a': a})

        return ArrayCascadeFilter(filters_def=filters)
