# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

import numpy as np
import scipy.signal

from bci_panel import bci_panels
from braintech.obci.signal_processing.buffers.auto_ring_buffer import AutoRingBuffer
from obci_p300.helper_functions import get_line_filters
from obci_p300.utils import prepare_x_y, SamplePacketCascadeFilter


def decimate(decimation, fs, chunk):
    decimation = int(fs / decimation)
    decimated_fs = fs / decimation

    detrended = scipy.signal.detrend(chunk, axis=-1, type='linear')
    decimated_chunk = scipy.signal.decimate(detrended, q=decimation, n=2, axis=-1, zero_phase=False)
    return decimated_chunk, decimated_fs


class SignalVisualiser:
    """Visualises signal through BCIPanel."""

    def __init__(self, bci_panel=None, last_s=10, every=1, fs=1024, chnames=None,
                 decimation=128, feature_name='signal', line_filters=False):
        """Shows last seconds."""
        assert chnames is not None
        self.decimation = decimation
        self.feature_name = feature_name
        self.last_s = last_s
        self.fs = fs
        self.every = every
        self.chnames = chnames

        self.use_line_filters = line_filters
        self.line_filters = SamplePacketCascadeFilter(filters_def=get_line_filters(fs / 2))
        self.buffer = AutoRingBuffer(**self._get_buffer_kwargs())

        if bci_panel is not None:
            self.bci_panel = bci_panel
        else:
            self.bci_panel = bci_panels.BCIPanel()

    def _get_buffer_kwargs(self):
        last_s_in_samples = int(self.fs * self.last_s)
        every_samples = int(self.every * self.fs)
        return dict(
            from_sample=last_s_in_samples,
            samples_count=last_s_in_samples,
            every=every_samples,
            num_of_channels=len(self.chnames),
            ret_func=self._handle_buffer,
            copy_on_ret=False,
            return_impedance=False,
        )

    def _decimate(self, chunk):
        return decimate(self.decimation, self.fs, chunk)

    def _prepare_signal_buffor(self, chunk):
        decimated_chunk, decimated_fs = self._decimate(chunk)
        x, y = prepare_x_y(decimated_chunk, times=decimated_fs, separation=5, clip=True,
                           separation_type=self.bci_panel.SEPARATION_TYPE)
        return x, y, decimated_fs

    def _handle_buffer(self, chunk):
        x, y, fs = self._prepare_signal_buffor(chunk)
        self.bci_panel.update_plot(field_id='all', feature=self.feature_name,
                                   x_data=x, y_data=y,
                                   labels=self.chnames,
                                   title=self.feature_name,
                                   sampling_rate=fs)

    def handle_packet(self, sample_packet):
        if self.use_line_filters:
            sample_packet = self.line_filters.get(sample_packet)
        self.buffer.handle_sample_packet(sample_packet)

    def handle_sample(self, sample):
        self.buffer.add_sample_to_buffer(sample)


class SignalSpectrumVisualiser(SignalVisualiser):
    def __init__(self, *args, freq_mask=None, **kwargs):
        super().__init__(*args, **kwargs)
        if freq_mask is None:
            self.freq_mask = [5, 40]
        else:
            self.freq_mask = freq_mask

    def _handle_buffer(self, chunk):
        detrended = scipy.signal.detrend(chunk, axis=-1, type='linear')
        fft_window = scipy.signal.get_window('hamming', Nx=detrended.shape[1])
        fft = np.fft.fft(detrended * fft_window)
        freq = np.fft.fftfreq(n=detrended.shape[1], d=1 / self.fs)
        usable_freq_mask = (freq >= self.freq_mask[0]) * (freq <= self.freq_mask[1])
        freqs = [freq[usable_freq_mask]] * len(self.chnames)
        spectrum = np.abs(fft[:, usable_freq_mask]) ** 2
        self.bci_panel.update_plot(field_id='all', feature=self.feature_name,
                                   x_data=freqs,
                                   y_data=spectrum,
                                   labels=self.chnames,
                                   xlabel='Frequency [Hz]',
                                   ylabel='Power Density',
                                   title="Signal spectrum")


class SignalImpedanceVisualiser(SignalVisualiser):
    """For now - only from 50 Hz."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        b, a = self._prepare_line_peak()
        self.sample_filter = SamplePacketCascadeFilter(filters_def=[{'a': a, 'b': b}])

    def _get_buffer_kwargs(self):
        kwargs = super()._get_buffer_kwargs()
        kwargs['ret_func'] = self._handle_samples_chunk_with_impedance
        kwargs['return_impedance'] = True
        return kwargs

    def _prepare_line_peak(self):
        nyquist_rate = self.fs / 2
        w0 = 50 / nyquist_rate
        Q = w0 / (2 / nyquist_rate)
        b, a = scipy.signal.iirpeak(w0, Q)
        return b, a

    def handle_packet(self, sample_packet):
        filtered_packet = self.sample_filter.get(sample_packet)
        super().handle_packet(filtered_packet)

    def _handle_samples_chunk_with_impedance(self, chunk, impedance):
        """chunk array (channels x samples)"""
        window = int(self.fs * 0.5)
        not_present_mask = impedance.flags != impedance.PRESENT
        if not_present_mask.all():
            impedances = self._get_pseudo_impedance_from_50Hz_for_TMSI(chunk, window)
            assert len(impedances) == len(impedance.flags)
            not_applicable_mask = impedance.flags == impedance.NOT_APPLICABLE
            impedances[not_applicable_mask, :] = np.nan
            decimated_fs = self.fs / window
            times = [np.arange(0, impedances.shape[1] / decimated_fs, 1 / decimated_fs)] * len(self.chnames)
        else:
            impedances = impedance.all_channels_data.T
            times = [np.linspace(0, impedances.shape[1] / self.fs, impedances.shape[1])] * len(self.chnames)
        self.bci_panel.update_plot(field_id='all', feature='signal impedance',
                                   x_data=times,
                                   y_data=impedances,
                                   labels=self.chnames,
                                   title="Signal impedance",
                                   ylim=self._get_bci_panel_ylim_from(impedances))

    @staticmethod
    def _get_bci_panel_ylim_from(impedances: np.array):
        try:
            return [0, impedances[np.isfinite(impedances)].max()]
        except ValueError:  # all impedance values are infinite
            return [0, 1]

    @staticmethod
    def _get_pseudo_impedance_from_50Hz_for_TMSI(chunk, window):
        rms = []
        for i in range(0, chunk.shape[1], window):
            rms_window = np.sqrt(np.sum(chunk[:, i:i + window] ** 2, axis=-1) / window)
            rms.append(rms_window)
        rms = np.array(rms).T
        amplitude = rms * np.sqrt(2)
        impedance = 2 * amplitude * 10e-6 / 10e-7
        impedance[impedance < 10] = np.inf
        return impedance
