#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

import asyncio
import os
import pickle
import time
from queue import Queue, Empty
from bci_panel import bci_panels
from bci_panel import web_bci_panel
from braintech.obci.core.broker import peer, messages
from braintech.obci.core.broker.broker import ObciException
from braintech.obci.experiment.peer.configured_peer import ConfiguredPeer
from braintech.obci.core.broker.message_handler_mixin import subscribe_message_handler
from braintech.obci.signal_processing.tags import tag_utils
from braintech.obci.core.utils.openbci_logging import log_crash
from braintech.obci.core.utils.properties import bool_property, param_property, split_floats, join_strings, split_string
from . import visualisers, analysis
from .common.bci_reporter import ReporterSummaryMsg
from .common.blink_message import BlinkMsg, BlinkConverter
from .common.messages import CalibrationIndexUpdate, DecisionVerboseMsg, CVStartedMsg
from .common.signal_visualiser import SignalVisualiser, SignalSpectrumVisualiser, SignalImpedanceVisualiser
from obci_p300.common.signal_preprocessor import SignalPreprocessor

__all__ = ('P300Peer',)


class BaseBaseP300Peer(peer.Peer):
    analysis_class = analysis.BaseP300Analysis

    def __init__(self, *args, config=None, amplifier_config=None, **kwargs):
        self.bci_panel = bci_panels.DummyBCIPanelProcess()
        self._analysis = None  # type: analysis.P300Analysis
        self.state = analysis.AnalysisStateVariables()

        self.buffer = None
        self.classification_future = None
        self.learning_future = None
        self._tasks_starting_mutex = asyncio.Lock()

        self.power_line_freq = 50
        self.exclude_from_montage = []
        if config is not None and amplifier_config is not None:
            self._set_static_parameters(config, amplifier_config)
        super().__init__(*args, **kwargs)

    def _set_static_parameters(self, config: dict, amplifier_config: dict):
        # timeout after decision when nothing should be accepted and no decision sent.
        self._post_decision_blinks_ignore_time = float(config["hold_after_dec"])
        # confidence (probability) level for internal decision to be counted to decision stop:
        # decision False Positive Rate - maximal level of error for internal decision.
        self.allowed_classifier_FPR = float(config['allowed_classifier_fpr'].split(';')[0])
        # available channels
        self.channel_names = amplifier_config['channel_names'].split(';')
        # used channels
        self.channels_for_classification = config['channels_for_classification'].split(';')
        self.sampling_rate = float(amplifier_config['sampling_rate'])
        # channel montage:
        # montage channels
        self.montage_channels = [itm for itm in config["channels_for_montage"].strip().split(';') if itm]

        # learn every n targets
        self.learn_every = int(config["learn_every"])

        # pre event baseline in seconds (negative) seconds
        self.baseline = float(config['baseline'])

        # post event window to consider seconds
        self.window = float(config['window'])

        # maximum averaged epochs
        # can be inf
        self.maximum_to_average = float(config['maximum_to_average'])
        # identical decisions to get final answer
        self.decision_stop = int(config['decision_stop'])
        # downsample to
        self.targetFs = float(config['downsample_to'])

        self.ignored_blink_ids = [int(i) for i in config['ignored_blink_ids'].strip().split(';')]

        self.wisdom_load = bool(int(config['wisdom_load']))
        self.epochs_dump_path = config['epochs_dump_path']

        if config['calibration_field_index']:
            self.training_index = int(config['calibration_field_index'])
        else:
            self.training_index = None
        self.proba_used_for_internal_decision = config['proba_used_for_internal_decision']
        self.use_linear_proba = bool(int(config['use_linear_proba']))
        self.channel_gains = split_floats(amplifier_config['channel_gains'])
        self.channel_offsets = split_floats(amplifier_config['channel_offsets'])
        self.use_web_bci_panel = bool(int(config['use_web_bci_panel']))
        self.use_windowed_bci_panel = bool(int(config['use_windowed_bci_panel']))
        self.visualise_classification_epochs_bool = bool(int(config['visualise_classification_epochs']))
        self.offline_learning = bool(int(config['offline_learning']))
        self.offline_learning_dataset_path = config['offline_learning_dataset_path']
        self.plots_image_path = config['plots_image_path']
        self.wisdom_path = config['wisdom_path']

    async def _connections_established(self):
        await super()._connections_established()

        # load classifier from file, if requested
        if self.wisdom_path:
            self.wisdom_path = os.path.expanduser(self.wisdom_path)

        await self._setup_before_running()

    async def _setup_before_running(self):
        # initialize parameters of the subclass
        self.init_params()
        self._analysis = self.create_analysis()
        channel_count = len(self.channels_for_classification)
        self.buffer = self._analysis.create_buffer(channel_count, self._buffer_ret_func)
        self._classifier = self._get_classifier()
        # array of requests to classification thread
        self.queue_classification = Queue()
        # array of requests to learning thread
        self.queue_learning = Queue()

    @log_crash
    def init_params(self):
        """Initializes paramters."""
        self._logger.info('Initialising parameters')
        self._signal_preprocessor = SignalPreprocessor(all_channels_names=self.channel_names,
                                                       classification_channel_names=self.channels_for_classification,
                                                       montage_channels_names=self.montage_channels,
                                                       sampling_rate=self.sampling_rate,
                                                       excluded_channel_names=self.exclude_from_montage,
                                                       notch_filters=[self.power_line_freq, ],
                                                       filters=None,
                                                       target_fs=self.targetFs,
                                                       channels_gains=self.channel_gains,
                                                       channels_offsets=self.channel_offsets,
                                                       )

        self._logger.info('Initialising buffers')
        self.state.reset_buffers(self.decision_stop)

    def create_analysis(self) -> analysis_class:
        return self.analysis_class(**self.get_analysis_kwargs())

    def get_analysis_kwargs(self) -> dict:
        return dict(
            baseline=self.baseline,
            window=self.window,
            wisdom_path=self.wisdom_path,
            offline_learning_dataset_path=self.offline_learning_dataset_path,
            channel_names=self.channel_names,
            channels_for_classification=self.channels_for_classification,
            learn_every=self.learn_every,
            targetFs=self.targetFs,
            sampling_rate=self.sampling_rate,
            use_linear_proba=self.use_linear_proba,
            ignored_blink_ids=self.ignored_blink_ids,
            training_index=self.training_index,
            state=self.state,
            decision_stop=self.decision_stop,
            decision_callback=self.send_decision,
            cv_start_callback=self.send_cv_start,
            proba_used_for_internal_decision=self.proba_used_for_internal_decision,
            allowed_classifier_FPR=self.allowed_classifier_FPR,
            maximum_to_average=self.maximum_to_average,
        )

    def send_cv_start(self):
        self.send_message(CVStartedMsg())

    def send_decision(self, **decision):
        msg = DecisionVerboseMsg(**decision)
        self.send_message(msg)

    async def _start(self):
        await super()._start()
        await self._start_tasks()

    async def _start_tasks(self):
        with (await self._tasks_starting_mutex):
            await self._await_tasks()
            self.classification_future = self._loop.run_in_executor(None, self._run_classification)
            self.learning_future = self._loop.run_in_executor(None, self._run_learning)
            self.create_task(self._classification_watchdog(self.classification_future))
            self.create_task(self._classification_watchdog(self.learning_future))
            # we don't need stop, as classifier will react to messages only when PeerState == running

    @property
    def in_training_mode(self):
        return self.training_index is not None

    def _buffer_ret_func(self, blink, chunk=None):
        """
        Internal function for the buffer object.

        :param blink: information contained in a single blink
        :param chunk: numpy 2D data array (channels × samples)
        """
        if chunk is None:
            chunk = blink
            blink = None
        request = {'chunk': chunk, 'blink': blink}
        target = self._analysis.identify_blink(blink)

        if target is None:
            self.queue_classification.put(request)
        else:
            request['target'] = target
            self.queue_learning.put(request)

    def _get_classifier(self):
        """
        Provide a classifier object from file or create a new one.

        :return AbstractClassifier: a classifier object
        """
        if self.wisdom_load:
            if os.path.isfile(self.wisdom_path):
                with open(self.wisdom_path, 'rb') as wisdom_file:
                    return pickle.load(wisdom_file)
            else:
                msg = 'Classifier does not exist, please run calibration first.'
                self.create_task(self.panic(Exception(msg)))
        else:
            return self._analysis.create_classifier()

    async def _classification_watchdog(self, future: asyncio.Future):
        await future

    def _run_classification(self):
        """This method will be run as a separate thread."""
        while self._should_run:
            try:
                request = self.queue_classification.get(timeout=0.1)
            except Empty:
                continue
            probabilities = self._analysis.classify(self._classifier, request['chunk'], request['blink'])
            if probabilities is not None:
                self._analysis.add_result(request['blink'], probabilities,
                                          self._classifier)
        while not self.queue_classification.empty():
            self.queue_classification.get_nowait()

    @property
    def _should_run(self):
        return self.is_running

    def _run_learning(self):
        """This method will be run as a separate thread."""
        while self._should_run:
            requests = self._collect_learning_requests()
            if len(requests) < 1:
                continue
            clone = self._classifier.clone()
            for request in requests:
                self._analysis.learn(clone, request['chunk'], request['target'], request['blink'])

            if self.wisdom_path is not None:
                dir_path = os.path.dirname(self.wisdom_path)
                if dir_path:
                    os.makedirs(dir_path, exist_ok=True)

                with open(self.wisdom_path, 'wb') as wisdom_file:
                    pickle.dump(clone, wisdom_file)

            self._classifier = clone
        while not self.queue_learning.empty():
            self.queue_learning.get_nowait()

    def _collect_learning_requests(self):
        requests = []
        try:
            requests.append(self.queue_learning.get(timeout=0.5))
            while True:
                requests.append(self.queue_learning.get_nowait())
        except Empty:
            pass
        return requests

    def _filter_signal_and_add_to_buffer(self, signal_message):
        self.buffer.handle_sample_packet(signal_message)  # SamplePacket

    @subscribe_message_handler(messages.SignalMessage)
    async def handle_signal(self, msg: messages.SignalMessage):
        """
        Internal method, fired every time signal message is received.

        :param msg: signal message data
        """
        if self.buffer is not None and self.is_running:
            self._raw_signal_sidechannel(msg.data)
            filtered_sample_packet = self._signal_preprocessor.get_preprocessed_packet(msg.data)
            self._filter_signal_and_add_to_buffer(filtered_sample_packet)
            self._filtered_signal_sidechannel(filtered_sample_packet)

    def _filtered_signal_sidechannel(self, sample_packet):
        pass

    def _raw_signal_sidechannel(self, sample_packet):
        pass

    @subscribe_message_handler(BlinkMsg)
    async def handle_blink(self, msg: BlinkMsg):
        """Internal method, fired every time blink message is received.

        :param  msg: blink message data
        """
        self._handle_blink(msg.get_blink())

    def _handle_blink(self, blink):
        if self.buffer is not None and self.is_running:
            self.buffer.handle_blink(blink)

    @subscribe_message_handler(messages.TagMsg)
    async def handle_tag_to_blink_converter(self, msg: messages.TagMsg):
        if self.is_running:
            tag = tag_utils.pack_tag_to_dict(msg.start_timestamp, msg.end_timestamp,
                                             msg.name, msg.desc, msg.channels)
            if tag['name'] == 'blink':
                blink = BlinkConverter(index_lambda=self._analysis.blink_converter_function)
                blink_msg = blink.get_message(tag)
                await self.handle_blink(blink_msg)

    async def _shutting_down(self):
        await self._await_tasks()
        await super()._shutting_down()

    async def _await_tasks(self):
        for future in [self.classification_future, self.learning_future]:
            if future is not None:
                try:
                    await future
                except Exception as e:
                    self._logger.exception("Error while cleaning up: {}".format(e))


class BaseP300Peer(BaseBaseP300Peer):
    graphs_creating_class = visualisers.BCIPanelVisualiser
    analysis_class = analysis.P300Analysis

    def reset_to_blank_state(self):
        async def _reset_to_blank_state():
            try:
                os.remove(self._analysis.wisdom_path)
            except FileNotFoundError:
                pass
            self.state = analysis.AnalysisStateVariables()
            self._connect_bci_panel_to_analysis_state()
            await self._setup_before_running()

        future = self.create_task(_reset_to_blank_state())
        future.result()  # block until done
        self.state.bci_panel_visualiser.reset_auc()
        self.state.bci_panel_visualiser.reset_histogram()
        self.state.bci_panel_visualiser.reset_calibration_erp()

    @property
    def _channel_labels_for_visualization(self):
        return [ch for ch in self.channel_names if ch in self.channels_for_classification]

    @subscribe_message_handler(ReporterSummaryMsg)
    async def handle_reporter_summary(self, msg: ReporterSummaryMsg):
        self._logger.info("Reporter summary: {}".format(msg))
        if self.state.bci_panel_visualiser is not None:
            self.state.bci_panel_visualiser.update_reporter_summary(data=msg.dict_data)

    @subscribe_message_handler(CalibrationIndexUpdate)
    def handle_training_index_change(self, msg: CalibrationIndexUpdate):
        index_string = msg.calibration_index
        self.training_index = int(index_string)
        if self._analysis is not None:
            self._analysis.training_index = self.training_index

    def _handle_blink(self, msg):
        time_from_last_decision = time.time() - self.state.last_decision_timestamp
        if time_from_last_decision > self._post_decision_blinks_ignore_time:
            super()._handle_blink(msg)

    def _filtered_signal_sidechannel(self, sample_packet):
        super()._filtered_signal_sidechannel(sample_packet)
        self.state.signal_visualiser.handle_packet(sample_packet)

    def _raw_signal_sidechannel(self, sample_packet):
        super()._raw_signal_sidechannel(sample_packet)
        self.state.signal_fft_visualiser.handle_packet(sample_packet)
        self.state.signal_impedance_visualiser.handle_packet(sample_packet)

    def _create_bci_panel(self):
        if self.use_web_bci_panel:
            self.bci_panel = web_bci_panel.WebBCIPanel()
        elif self.use_windowed_bci_panel:
            self.bci_panel = bci_panels.BCIPanelProcess()
        else:
            self.bci_panel = bci_panels.DummyBCIPanelProcess()

    async def _connections_established(self):
        await super()._connections_established()
        assert not (self.use_windowed_bci_panel and self.use_web_bci_panel)
        if self.offline_learning:
            learn_offline = self._loop.run_in_executor(None, self._analysis.learn_offline())
            try:
                await asyncio.wait_for(learn_offline, timeout=None)
            except ObciException as e:
                self._logger.error(str(e))
            await self.async_shutdown()
        else:
            self._create_bci_panel()
            self._connect_bci_panel_to_analysis_state()

    def _connect_bci_panel_to_analysis_state(self):
        self.state.signal_visualiser = SignalVisualiser(
            bci_panel=self.bci_panel,
            fs=self.sampling_rate,
            every=self.bci_panel.SIGNAL_UPDATE_RATE,
            last_s=self.bci_panel.SIGNAL_CHUNK_LENGTH,
            chnames=self.channels_for_classification,
            feature_name="raw_signal",
            line_filters=True,
        )
        self.state.signal_fft_visualiser = SignalSpectrumVisualiser(
            bci_panel=self.bci_panel,
            fs=self.sampling_rate,
            chnames=self.channel_names,
            feature_name="raw_signal_fft",
        )
        self.state.signal_impedance_visualiser = SignalImpedanceVisualiser(
            bci_panel=self.bci_panel,
            fs=self.sampling_rate,
            chnames=self.channel_names,
        )

        self.state.bci_panel_visualiser = self.graphs_creating_class(
            bci_panel=self.bci_panel,
            channel_labels_getter=lambda: self._channel_labels_for_visualization,
            visualise_last_and_epochs_getter=lambda: self.visualise_classification_epochs_bool,
            window=self.window,
            baseline=self.baseline,
            sampling_rate=self.sampling_rate,
            base_save_directory=self.plots_image_path or '',
        )

    async def _shutting_down(self):
        self._save_metadata()
        if self.state.bci_panel_visualiser is not None:
            self.state.bci_panel_visualiser.save_plots(dir_name='exit')
            self._logger.info('Closing BCI Panel')
            await self._loop.run_in_executor(executor=None, func=self.bci_panel.close)
            self._logger.info('BCI Panel closed')
        await super()._shutting_down()

    def _save_metadata(self):
        if self.epochs_dump_path:
            epochs = dict(targets=self.state.calibration_targets_buffor,
                          nontargets=self.state.calibration_nontargets_buffor)
            with open(self.epochs_dump_path, 'wb') as file:
                pickle.dump(epochs, file)

    def _buffer_ret_func(self, blink, chunk=None):
        if blink.timestamp < self._post_decision_blinks_ignore_time + self.state.last_decision_timestamp:
            return

        blink_in_ignored = int(blink.index) in self.ignored_blink_ids
        if not blink_in_ignored:
            if self._analysis._run_through_signal_goodness_detectors(blink):
                super()._buffer_ret_func(blink, chunk)


class P300Peer(BaseP300Peer, ConfiguredPeer):
    """P300 classifier master peer."""
    # config option which shows which option is now selected as "target"
    training_index = param_property('calibration_field_index', deserializer=int, serializer=str)
    proba_used_for_internal_decision = param_property('proba_used_for_internal_decision')
    use_linear_proba = bool_property('use_linear_proba')
    channel_gains = param_property('channel_gains', split_floats, join_strings)
    channel_offsets = param_property('channel_offsets', split_floats, join_strings)
    use_web_bci_panel = bool_property('use_web_bci_panel')
    use_windowed_bci_panel = bool_property('use_windowed_bci_panel')
    visualise_classification_epochs_bool = bool_property('visualise_classification_epochs')
    offline_learning = bool_property('offline_learning')
    offline_learning_dataset_path = param_property('offline_learning_dataset_path')
    plots_image_path = param_property('plots_image_path')
    wisdom_path = param_property('wisdom_path')

    _post_decision_blinks_ignore_time = param_property('hold_after_dec',
                                                       deserializer=float)
    allowed_classifier_FPR = param_property('allowed_classifier_fpr',
                                            deserializer=lambda s: float(s.split(';')[0]))
    channel_names = param_property('channel_names', deserializer=split_string)
    channels_for_classification = param_property('channels_for_classification', deserializer=split_string)
    sampling_rate = param_property('sampling_rate', deserializer=float)
    montage_channels = param_property(
        'channels_for_montage',
        deserializer=lambda s: [itm for itm in s.strip().split(';') if itm],
    )
    learn_every = param_property('learn_every', deserializer=int)
    baseline = param_property('baseline', deserializer=float)
    window = param_property('window', deserializer=float)
    maximum_to_average = param_property('maximum_to_average', deserializer=float)
    decision_stop = param_property('decision_stop', deserializer=int)
    targetFs = param_property('downsample_to', deserializer=float)
    ignored_blink_ids = param_property(
        'ignored_blink_ids',
        deserializer=lambda s: [int(i) for i in s.strip().split(';')],
    )
    wisdom_load = bool_property('wisdom_load')
    epochs_dump_path = param_property('epochs_dump_path')
