# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import numpy as np
import scipy.signal
from scipy.signal import lfiltic, lfilter, freqz

from braintech.obci.signal_processing.signal.data_generic_write_proxy import SamplePacket
from braintech.obci.core.utils.properties import split_string


def declutter_chunk(data, separation=2, separation_type='std'):
    """Declutter data chunk.

    :param data: numpy array (chanels x samples)
    """
    offset = 0
    data_new = []
    for data in data:
        data_new.append(np.array(data) + offset)
        if separation_type == 'std':
            offset -= separation * np.std(data)
        else:
            offset -= separation
    return np.array(data_new)


def prepare_x_y(chunk, times, detrend=True, declutter=True, separation_type='std', separation=2, clip=False,
                baseline_correct=True, scale_signal=0.05):
    """
    :param chunk: ndarray (nchannels, nsamples)
    :param times: - np.array of times or sample frequency
    :param clip: - False or a absolute number - for signal clipping (after detrending)
    """
    if not isinstance(times, np.ndarray):
        times = np.arange(0, chunk.shape[1] / times, 1 / times)

    x = []
    y = []
    for channel in range(chunk.shape[0]):
        x.append(times)
        y_ = _prepare_channel_chunk(baseline_correct, channel, chunk, detrend)
        y.append(y_ * scale_signal)  # for webbci panel - tmsi

    if clip:
        if separation_type == 'std':
            clip = separation * np.median(np.std(y, axis=-1))
        else:
            clip = separation
        for channel in y:
            channel[channel > clip / 2] = clip / 2
            channel[channel < -clip / 2] = -clip / 2

    if declutter:
        if clip is not False:
            y = declutter_chunk(y, separation_type='abs', separation=clip)
        else:
            y = declutter_chunk(y, separation_type=separation_type, separation=separation)
    return x, y


def _prepare_channel_chunk(baseline_correct, channel, chunk, detrend):
    if baseline_correct:
        y_ = chunk[channel] - np.mean(chunk[channel])
    else:
        y_ = chunk[channel]
    if detrend:
        y_ = scipy.signal.detrend(y_, axis=-1, type='linear')
    return y_


class ArrayCascadeFilter:
    def __init__(self, filters_def=None, init_length=0):
        """Init sample packet filter.
        This filter assumes that channel number will not change.
        :param filter_def: list of dicts of type {'b':[1, 2, 3], 'a':[1,3,4]}
        """
        if filters_def is None:
            filters_def = []
        self.filters = filters_def
        self.ch_num = None
        self._filtered_samples = 0
        self._initialization_length = init_length

    def _prepare_zi(self, x: np.ndarray, filter):
        zis = []
        for i in range(self.ch_num):
            zis.append(lfiltic(filter['b'], filter['a'], x[:, i], x[:, i]))
        filter['zis'] = zis

    def get(self, x: np.ndarray):
        """x array (samples, channels)"""
        if self.ch_num is None:
            self.ch_num = x.shape[1]
        filtered = x.copy()
        for i in self.filters:
            try:
                zis = i['zis']
            except KeyError:
                self._prepare_zi(x, i)
                zis = i['zis']
            for ch in range(self.ch_num):
                filtered[:, ch], zi = lfilter(i['b'], i['a'], filtered[:, ch], zi=zis[ch])
                zis[ch] = zi
        self._filtered_samples += filtered.shape[0]
        return filtered

    def get_initialized(self, x: np.ndarray):
        number_of_samples_to_forget = self._initialization_length - self._filtered_samples
        filtered = self.get(x)
        if number_of_samples_to_forget > 0:
            return filtered[number_of_samples_to_forget:]
        return filtered

    def filter_response(self, worN=10000):
        first = True
        for i in self.filters:
            if first:
                w, h = freqz(b=i['b'], a=i['a'], worN=worN)
                first = False
            else:
                w_, h_ = freqz(b=i['b'], a=i['a'], worN=worN)
                h *= h_
        return w, h


class SamplePacketCascadeFilter(ArrayCascadeFilter):
    def get(self, x: SamplePacket):
        data = x.samples.copy()
        filtered = super().get(data)

        return SamplePacket(samples=filtered, ts=x.ts, impedance=x._impedance)


def strip_split_string(string):
    return [s.strip() for s in split_string(string.strip()) if s.strip()]
