import pathlib
import typing

import numpy as np
import scipy
from scipy.stats import scoreatpercentile

from bci_panel import web_bci_panel, bci_panels
from obci_p300.utils import prepare_x_y


class BCIPanelVisualiser:
    def __init__(self, bci_panel: typing.Union[bci_panels.BCIPanelProcess, web_bci_panel.WebBCIPanel],
                 channel_labels_getter, visualise_last_and_epochs_getter,
                 window, baseline, sampling_rate, base_save_directory: str):
        self.bci_panel = bci_panel
        self._channel_labels_getter = channel_labels_getter
        self._visualise_last_and_epochs_getter = visualise_last_and_epochs_getter
        # to be extracted elsewhere
        self.window = window
        self.baseline = baseline
        self.sampling_rate = sampling_rate
        self._base_save_dir = pathlib.Path(base_save_directory)
        self.web_bcipanel_epoch_scale = 10,

    @property
    def is_web_bci_panel(self):
        return isinstance(self.bci_panel, web_bci_panel.WebBCIPanelMixin)

    @property
    def _channel_labels(self) -> typing.List[str]:
        return self._channel_labels_getter()

    def update_reporter_summary(self, data):
        name = 'Accuracy and itr'
        self.bci_panel.update_text(field_id='None', feature=name, data=data, title=name)

    @property
    def _visualise_last_and_epochs(self) -> bool:
        return self._visualise_last_and_epochs_getter()

    def visualise_classification(self, features, epochs, index):
        self._visualise_classification_features(
            features=features,
            index=index,
        )
        if self._visualise_last_and_epochs:
            self._visualise_classification_last_feature(
                features=features,
                index=index,
            )
            self._visualise_classification_epochs(
                epochs=epochs,
                index=index,
            )
            self._visualise_classification_last_epoch(
                epochs=epochs,
                index=index,
            )

    def visualise_classification_confidence(self, classifier, allowed_classifier_FPR: float,
                                            singular_attempts: typing.DefaultDict,
                                            average_attempts: typing.DefaultDict,
                                            weighted_averaged_attempts: typing.DefaultDict):
        for blink_index in singular_attempts.keys():
            averaged_prob_values = average_attempts[blink_index][:]
            y_data = [
                singular_attempts[blink_index][:],
                averaged_prob_values,
                weighted_averaged_attempts[blink_index][:]
            ]
            x_data = [np.arange(1, 1 + len(yd), 1) for yd in y_data]

            threshold_line_x = [min(x_data[0]), max(x_data[0])]
            threshold = classifier.threshold_for_fpr(allowed_classifier_FPR)
            threshold_line_y = [threshold, threshold]

            self.bci_panel.update_plot(
                field_id=blink_index,
                feature='classifier_confidence',
                x_data=x_data + [threshold_line_x],
                y_data=y_data + [threshold_line_y],
                labels=['single', 'mean', 'weighted mean', 'threshold'],
                plot_type='scatter',
                title="Confidence",
                ylim=[0, 1],
            )
        if self.is_web_bci_panel:
            batch_values = {
                blink_idx: singular_attempts[blink_idx][:]
                for blink_idx in singular_attempts.keys()
            }
            if batch_values:
                threshold = classifier.threshold_for_fpr(allowed_classifier_FPR)
                self.bci_panel.update_plot(
                    feature='classifier_confidence_batch',
                    threshold=threshold,
                    choices_values=batch_values,
                )

    def visualise_calibration(self, classifier, targets, non_targets):
        self._visualise_calibration_features(
            classifier=classifier,
            targets=targets,
            non_targets=non_targets,
        )
        self._visualise_calibration_epochs(
            targets=targets,
            non_targets=non_targets,
        )

    def reset_auc(self):
        self.bci_panel.update_plot(feature='AUC', value=0.0)

    def reset_histogram(self):
        self.bci_panel.update_hist(
            field_id='None',
            feature='prediction hist',
            data=[
                [],
                [],
            ],
            targets_histogram=[],
            nontargets_histogram=[],
            bins=[],
            bins_centers=[],
            threshold=0.5,
            title='Target cross validation prediction histograms',
            labels=['targets', 'nontargets'],
            normed=True,
            vline=0.5,
        )

    def reset_calibration_erp(self):
        self.bci_panel.update_plot(
            field_id="target field",
            feature='mean epochs calib',
            x_data=[[], []],
            y_data=[[], []],
            labels=[i + ' tgt' for i in self._channel_labels] + [i + ' ntgt' for i in self._channel_labels],
            title="Mean epochs TGT={}, NTGT={}".format(
                0, 0),
            channels=self._channel_labels
        )

    def update_plots(self, classifier, allowed_classifier_FPR):
        self._update_classifier_coefficients_plot(classifier)
        self._update_auc_plot(classifier)
        self._update_auc_per_average(classifier)
        self._update_auc_history_plot(classifier)
        self._update_MannWhitneyU_plot(classifier)
        self._update_cross_validation_prediction_histogram(classifier, allowed_classifier_FPR)
        self._update_FPR_vs_classifier_certainty_plot(classifier, allowed_classifier_FPR)

    def clear(self, features_ids):
        for feature_id in features_ids:
            self.bci_panel.update_plot(
                field_id=feature_id,
                feature='features mean',
                x_data=[[]],
                y_data=[[]],
                title='Mean features',
            )
            self.bci_panel.update_plot(
                field_id=feature_id,
                feature='classifier_confidence',
                x_data=[[]] * 2,
                y_data=[[]] * 2,
                title='Confidence',
            )
            if self._visualise_last_and_epochs:
                self.bci_panel.update_plot(
                    field_id=feature_id,
                    feature='epochs mean',
                    x_data=[[]] * len(self._channel_labels),
                    y_data=[[]] * len(self._channel_labels),
                    labels=self._channel_labels,
                    title='Mean epochs',
                )
                self.bci_panel.update_plot(
                    field_id=feature_id,
                    feature='epoch',
                    x_data=[[]] * len(self._channel_labels),
                    y_data=[[]] * len(self._channel_labels),
                    labels=self._channel_labels,
                    title='Last epoch',
                )
                self.bci_panel.update_plot(
                    field_id=feature_id,
                    feature='feature',
                    x_data=[[]],
                    y_data=[[]],
                    title='Last feature',
                )

    def save_plots(self, dir_name: str):
        plots_dir = self._base_save_dir / dir_name
        self.bci_panel.save_plots(dir_path=str(plots_dir))

    def _visualise_classification_epochs(self, epochs, index):
        """Online session."""
        mean_epochs = scipy.signal.detrend(
            np.array(epochs[index], ndmin=3), axis=-1, type='linear'
        ).mean(axis=0)
        median_mean_epoch_std = np.median(mean_epochs.std(axis=-1))
        mean_epochs_x, mean_epochs_y = self._prepare_x_y(
            mean_epochs, declutter=True, detrend=True, separation=median_mean_epoch_std * 4, separation_type='abs')
        self.bci_panel.update_plot(
            field_id=index,
            feature='epochs mean',
            x_data=mean_epochs_x,
            y_data=mean_epochs_y,
            labels=self._channel_labels,
            title="Mean epochs N={}".format(len(epochs[index])),
        )

    def _visualise_classification_last_epoch(self, epochs, index):
        last_epoch = epochs[index][-1]
        last_epoch_x, last_epoch_y = self._prepare_x_y(last_epoch, declutter=True, detrend=True)
        self.bci_panel.update_plot(
            field_id=index,
            feature='epoch',
            x_data=last_epoch_x,
            y_data=last_epoch_y,
            labels=self._channel_labels,
            title="Last epoch",
        )

    def _visualise_classification_features(self, features, index):
        """Online session."""
        mean_features = np.array(features[index], ndmin=2).mean(axis=0)
        median_mean_feature_std = np.median(mean_features.std(axis=-1))

        mean_features_x, mean_features_y = self._prepare_x_y(
            mean_features.reshape((len(self._channel_labels), -1)),
            declutter=True,
            detrend=False,
            separation=self.bci_panel.get_separation(median_mean_feature_std),
            separation_type='abs',
            times=[0, self.window],
        )
        self.bci_panel.update_plot(
            field_id=index,
            feature='features mean',
            x_data=mean_features_x,
            y_data=mean_features_y,
            title="Mean features N={}".format(len(features[index])),
            labels=self._channel_labels,
        )

    def _visualise_classification_last_feature(self, features, index):
        mean_features = np.array(features[index], ndmin=2).mean(axis=0)
        median_mean_feature_std = np.median(mean_features.std(axis=-1))
        last_feature = features[index][-1]
        last_feature_x, last_feature_y = self._prepare_x_y(
            last_feature.reshape((len(self._channel_labels), -1)),
            declutter=True,
            detrend=False,
            separation=self.bci_panel.get_separation(median_mean_feature_std),
            separation_type='abs',
            times=[0, self.window],
        )
        self.bci_panel.update_plot(
            field_id=index,
            feature='feature',
            x_data=last_feature_x,
            y_data=last_feature_y,
            title="Last feature",
            labels=self._channel_labels,
        )

    def _visualise_calibration_features(self, classifier, targets, non_targets):
        """For calibration"""
        classes = np.array(classifier.learning_buffor_classes, ndmin=1)
        features = np.array(classifier.learning_buffor_features, ndmin=2)
        at_least_one_target_available = np.sum(classes == 1) > 0
        at_least_one_non_target_available = np.sum(classes == 0) > 0
        if at_least_one_target_available and at_least_one_non_target_available:
            mean_target_feature_flat = features[classes == 1].mean(axis=0)
            mean_non_target_feature_flat = features[classes == 0].mean(axis=0)
            mean_target_feature = mean_target_feature_flat.reshape((len(self._channel_labels), -1))
            mean_non_target_feature = mean_non_target_feature_flat.reshape((len(self._channel_labels), -1))
            common_x, common_epochs = self._prepare_epoch_pairwise_for_visualisation(
                mean_target_feature, mean_non_target_feature, detrend=False, times=[0, self.window])
            self.bci_panel.update_plot(
                field_id="target field",
                feature='features mean calib',
                x_data=common_x,
                y_data=common_epochs,
                labels=[i + ' tgt' for i in self._channel_labels] + [i + ' ntgt' for i in self._channel_labels],
                title="Mean feature vectors TGT={}, NTGT={}".format(
                    len(targets), len(non_targets)),
            )
        try:
            last_target_feature = features[classes == 1][-1].reshape((len(self._channel_labels), -1))
            last_non_target_feature = features[classes == 0][-1].reshape((len(self._channel_labels), -1))
        except IndexError:
            pass
        else:
            common_x, common_epochs = self._prepare_epoch_pairwise_for_visualisation(
                last_target_feature, last_non_target_feature, detrend=False, times=[0, self.window])
            self.bci_panel.update_plot(
                field_id="target field",
                feature='features calib',
                x_data=common_x,
                y_data=common_epochs,
                labels=[i + ' tgt' for i in self._channel_labels] + [i + ' ntgt' for i in self._channel_labels],
                title="Last feature vector",
            )

    def _visualise_calibration_epochs(self, targets, non_targets):

        target = np.array(targets, ndmin=3).mean(axis=0)
        non_target = np.array(non_targets, ndmin=3).mean(axis=0)
        if target.shape[0] == len(self._channel_labels) and non_target.shape[0] == len(self._channel_labels):
            common_x, common_epochs = self._prepare_epoch_pairwise_for_visualisation(target, non_target)
            self.bci_panel.update_plot(
                field_id="target field",
                feature='mean epochs calib',
                x_data=common_x,
                y_data=common_epochs,
                labels=[i + ' tgt' for i in self._channel_labels] + [i + ' ntgt' for i in self._channel_labels],
                title="Mean epochs TGT={}, NTGT={}".format(
                    len(targets), len(non_targets)),
                channels=self._channel_labels
            )
        try:  # I want to update when at least one target and one non-target is available
            last_target = targets[-1]
            last_non_target = non_targets[-1]
        except IndexError:
            pass
        else:
            common_x, common_epochs = self._prepare_epoch_pairwise_for_visualisation(last_target, last_non_target)
            self.bci_panel.update_plot(
                field_id="target field",
                feature='last target calib',
                x_data=common_x,
                y_data=common_epochs,
                labels=[i + ' tgt' for i in self._channel_labels] + [i + ' ntgt' for i in self._channel_labels],
                title="Last epochs",
            )

    def _update_classifier_coefficients_plot(self, classifier):
        classifier_coefs = classifier.clf.coef_.flatten()
        classifier_coefs_x = [range(len(classifier_coefs)), ]
        classifier_coefs_y = [classifier_coefs, ]
        y_lim = [scoreatpercentile(classifier_coefs, 5),
                 scoreatpercentile(classifier_coefs, 95)]
        self.bci_panel.update_plot(
            field_id='target field   ',
            feature='features Classifier coefficients',
            x_data=classifier_coefs_x,
            y_data=classifier_coefs_y,
            title='Classifier coefficients',
            plot_type='scatter',
            ylim=y_lim,
        )

    def _update_auc_plot(self, classifier):
        roc_fprs = [i[1] for i in classifier.rocs][::2]
        roc_tprs = [i[0] for i in classifier.rocs][::2]
        roc_averages = ['N_avr = {}'.format(i) for i in classifier.n_averaged_rocs][::2]
        self.bci_panel.update_plot(
            field_id='None',
            feature='ROC',
            x_data=roc_fprs,
            y_data=roc_tprs,
            labels=roc_averages,
            title='AUC: {}, MannWhitU p = {:0.3}'.format(classifier.roc_auc, classifier.p_val),
            legend_loc=4,
        )
        if self.is_web_bci_panel:
            self.bci_panel.update_plot(feature='AUC', value=classifier.roc_auc)

    def _update_auc_per_average(self, classifier):
        self.bci_panel.update_plot(
            field_id='None',
            feature='AUC per average',
            x_data=[classifier.n_averaged_rocs],
            y_data=[classifier.roc_aucs],
            xlabel='N averaged',
            title='AUC per average',
            labels=['AUC'],
            ylim=[0.45, 1.01],
            plot_type='scatter',
        )

    def _update_auc_history_plot(self, classifier):
        self.bci_panel.update_plot(
            field_id='None',
            feature='AUC History',
            x_data=[classifier.history_n, classifier.history_n],
            y_data=[classifier.AUC_history, classifier.crit_AUC_history],
            xlabel='N targets',
            title='AUC History',
            labels=['AUC', 'AUC crit. theoretical p=0.01'],
            ylim=[0, 1.01],
            plot_type='scatter',
        )

    def _update_MannWhitneyU_plot(self, classifier):
        """
        Tracks changes of MannWhitneyU test p value during calibration.
        Test is done on classifiers rankings on collected epochs.
        """
        self.bci_panel.update_plot(
            field_id='None',
            feature='p val History',
            x_data=[classifier.history_n, [min(classifier.history_n),
                                           max(classifier.history_n)]],
            y_data=[classifier.p_val_history, [0.05, 0.05]],
            xlabel='N targets',
            title='MannWhitU p History',
            ylim=[0, 1],
            labels=['p val', 'crit. p val'],
        )

    def _update_cross_validation_prediction_histogram(self, classifier, allowed_classifier_FPR):
        selected_threshold = classifier.threshold_for_fpr(allowed_classifier_FPR)
        bins = np.linspace(0, 1, 20)
        targets_histogram, _ = np.histogram(
            a=classifier.calib_tgts_prediction,
            bins=bins
        )
        non_targets_histogram, _ = np.histogram(
            a=classifier.calib_ntgts_prediction,
            bins=bins
        )
        bins_centers = np.diff(bins) / 2 + bins[:-1]
        self.bci_panel.update_hist(
            field_id='None',
            feature='prediction hist',
            data=[
                classifier.calib_tgts_prediction,
                classifier.calib_ntgts_prediction,
            ],
            targets_histogram=targets_histogram,
            nontargets_histogram=non_targets_histogram,
            bins=bins,
            bins_centers=bins_centers,
            threshold=selected_threshold,
            title='Target cross validation prediction histograms',
            labels=['targets', 'nontargets'],
            normed=True,
            vline=selected_threshold,
        )

    def _update_FPR_vs_classifier_certainty_plot(self, classifier, allowed_classifier_FPR):
        selected_threshold = classifier.threshold_for_fpr(allowed_classifier_FPR)
        self.bci_panel.update_plot(
            field_id='None',
            feature='False Positive Rate vs classifier certainty',
            x_data=[
                classifier.dec_thresholds,
                [selected_threshold, selected_threshold],
            ],
            y_data=[classifier.thr_fpr, [0, 1]],
            xlabel='Classifier certainty',
            ylabel='FPR',
            title=(
                'FPR vs classifier certainty threshold\n'
                'selected threshold: {:0.3}'
            ).format(selected_threshold),
            labels=['FPR', 'Selected threshold'],
        )

    def _prepare_epoch_pairwise_for_visualisation(self, epoch1, epoch2, detrend=True, times=None):
        """times - none or list of (first time index, last time)"""
        epoch1 = epoch1 * self.web_bcipanel_epoch_scale
        epoch2 = epoch2 * self.web_bcipanel_epoch_scale
        separation = self.bci_panel.get_separation(lambda: np.median(
            np.concatenate((epoch1, epoch2)).std(axis=-1)))
        x_t, y_t = self._prepare_x_y(epoch1, declutter=True, detrend=detrend,
                                     separation=separation,
                                     separation_type='abs', times=times)
        x_nt, y_nt = self._prepare_x_y(epoch2, declutter=True, detrend=detrend,
                                       separation=separation,
                                       separation_type='abs', times=times)
        common_epochs = np.concatenate((y_t, y_nt))
        common_x = np.concatenate((x_t, x_nt))
        return common_x, common_epochs

    def _prepare_x_y(self, chunk, declutter=False, detrend=False, separation=4, separation_type='std', times=None):
        assert times is None or len(times) == 2, "times is None - automatic timescale, or first and last scale"
        if times is None:
            times = np.linspace(self.baseline, self.baseline + chunk.shape[1] / self.sampling_rate, chunk.shape[1])
        else:
            times = np.linspace(times[0], times[1], chunk.shape[1])
        if self.is_web_bci_panel:
            scale_signal = 0.02
            separation_type = self.bci_panel.SEPARATION_TYPE
            separation = 4
        else:
            scale_signal = 1
        return prepare_x_y(chunk, times, detrend=detrend, declutter=declutter,
                           separation_type=separation_type, separation=separation, clip=False,
                           scale_signal=scale_signal)
