# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.
import sys
import os
import numpy as np
import pylab as pb
import scipy.signal
from braintech.obci.signal_processing.read_manager import ReadManager
import argparse

parser = argparse.ArgumentParser(description="Process EEG data to create FFT images")
parser.add_argument('-b', help='Baseline .bin/raw/xml filename, will be used with all stimulation files')
parser.add_argument('-a', help='Alpha (eyes closed signal) .bin/raw/xml filename, will be used with all stimulation')
parser.add_argument('-c', help='Channel name to use, if not provided all will be used, this argument can be repeated',
                    type=str, action='append')
parser.add_argument('files', help='The stimulation filenames .bin/raw/xml', nargs="+")
args = parser.parse_args()

# import IPython; IPython.embed()


def _draw_fft(channels, ax, freqs, spectrum, channel_prefix=''):
    for nr, chnl in enumerate(mgr.get_param('channels_names')):
        if len(channels) > 0:
            if chnl not in channels:
                continue
        ax.plot(freqs, spectrum[nr], label=channel_prefix + chnl)


def _get_spectrum(mgr):
    chunk = mgr.get_microvolt_samples()
    fft_window = scipy.signal.get_window('hamming', Nx=chunk.shape[1])
    fft = np.fft.fft(chunk * fft_window)
    freq = np.fft.fftfreq(n=chunk.shape[1], d=1 / float(mgr.get_param("sampling_frequency")))
    usable_freq_mask = (freq >= 8) * (freq <= 40)
    freqs = freq[usable_freq_mask]
    spectrum = np.abs(fft[:, usable_freq_mask]) ** 2
    return freqs, spectrum


def draw_fft(mgr, baseline=None, alpha=None, channels=tuple()):
    freqs, spectrum = _get_spectrum(mgr)

    fig = pb.figure(figsize=(15, 10))
    ax = fig.add_subplot(111)
    ax.set_ylabel('Power spectrum density')
    ax.set_xlabel('Frequency, Hz')
    ax.set_title("Power spectrum, signal Hamming windowed")
    _draw_fft(channels, ax, freqs, spectrum, channel_prefix='Stim ')

    if baseline is not None:
        baseline_mgr = load_read_manager(baseline)
        freqs, spectrum = _get_spectrum(baseline_mgr)
        _draw_fft(channels, ax, freqs, spectrum, channel_prefix='Baseline ')

    if alpha is not None:
        alpha_mgr = load_read_manager(alpha)
        freqs, spectrum = _get_spectrum(alpha_mgr)
        _draw_fft(channels, ax, freqs, spectrum, channel_prefix='Alpha ')

    pb.legend()
    fig.tight_layout()
    return fig


def load_read_manager(file):

    ext = file[-3:]

    if ext in 'xml raw bin'.split():
        file = file[:-4]

    try:
        print("Trying .raw")
        mgr = ReadManager(p_info_source=file + '.xml', p_data_source=file + '.raw', p_tags_source=None)
        print("Ok")
    except FileNotFoundError:
        print("Trying .bin")
        mgr = ReadManager(p_info_source=file + '.xml', p_data_source=file + '.bin', p_tags_source=None)
        print("Ok")
    print("File read: ", file)
    return mgr


files = args.files
if not files:
    print("Please provide obci datasets for analysis (.raw .xml or .bin files, duplicates will be inteligently ignored)"
          "\nFFT will be generated in 'images' subdirectory in the files dir"
          "\n"
          "\nExample:"
          "\npython3 draw_fft_freqs.py ./data_files*")
    sys.exit(1)
print("Got files:", files)
files = set([i.rsplit('.', maxsplit=1)[0] for i in files])
print("\nAfter duplicate removal:", [i + '.xml' for i in files])
for file in files:
    workdir = os.path.dirname(file)
    filename = os.path.basename(file)
    mgr = load_read_manager(file)
    fig = draw_fft(mgr, baseline=args.b, alpha=args.a, channels=args.c)

    imagedir = os.path.join(workdir, 'images')
    os.makedirs(imagedir, exist_ok=True)
    fig.savefig(os.path.join(imagedir, filename + '.png'))
