import pickle
import numpy
import numpy as np
import pylab as pb
import warnings
import os
import sys
import tqdm
from sklearn.model_selection import LeaveOneOut
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.neighbors import KNeighborsClassifier
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from sklearn.svm import SVC
from sklearn.pipeline import Pipeline
from sklearn.feature_selection import RFECV
from sklearn.decomposition import FastICA
from sklearn.metrics import roc_auc_score, roc_curve
from mne.decoding import CSP
from mne.preprocessing import Xdawn  # needs epochs
from imblearn.under_sampling import RandomUnderSampler
from mne import EpochsArray, create_info

warnings.filterwarnings("ignore")

if len(sys.argv) == 1:
    print("Loading default dump")
    dumpfile = "./data_subj3/classifier.dump"
else:
    dumpfile = sys.argv[-1]

c = pickle.load(open(dumpfile, 'rb'))


# removing artifacts (by amplitude criterion)
clean_f = []
clean_c = []
for f, cl in zip(c.learning_buffor_features, c.learning_buffor_classes):
    if np.all(np.abs(f) < 1000):
        clean_f.append(f)
        clean_c.append(cl)
    else:
        print("Dropping feature", cl)

c.learning_buffor_classes = clean_c
c.learning_buffor_features = clean_f

imagedir = os.path.join(os.path.dirname(dumpfile), 'img')


def calc_accuracy(preds, true_class):
    accuracies = []
    thrs = []
    for thr in np.arange(min(preds) - 1, max(preds) + 1, (max(preds) - min(preds)) / 10000):
        targets_preds = (preds > thr)
        accuracy = np.mean(targets_preds == true_class) * 100
        thrs.append(thr)
        accuracies.append(accuracy)
    return accuracies, thrs


def test_classifier(l, draw=True, l_test=None, fig_nr=0):
    if l_test:
        l_test.fit(c.learning_buffor_features, c.learning_buffor_classes)

    loo = LeaveOneOut()

    data = numpy.array(c.learning_buffor_features)
    classes = numpy.array(c.learning_buffor_classes)

    # balancing
    balancer = RandomUnderSampler(random_state=42)

    data, classes = balancer.fit_sample(data, classes)

    leave_one_out_predictions = []
    leave_one_out_true_classes = []

    for train_index, test_index in tqdm.tqdm(loo.split(data)):
        X_train, X_test = data[train_index], data[test_index]
        y_train, y_test = classes[train_index], classes[test_index]
        l.fit(X_train, y_train)
        predictions = l.predict_proba(X_test)[:, 1]
        # import IPython
        # IPython.embed()
        leave_one_out_predictions.extend(predictions)
        leave_one_out_true_classes.extend(y_test)

    leave_one_out_predictions = np.array(leave_one_out_predictions)
    leave_one_out_true_classes = np.array(leave_one_out_true_classes)

    loo_roc_fpr, loo_roc_tpr, _ = roc_curve(leave_one_out_true_classes, leave_one_out_predictions)
    loo_auc = roc_auc_score(leave_one_out_true_classes, leave_one_out_predictions)

    if draw:

        os.makedirs(imagedir, exist_ok=True)

        loo_tgt_mask = np.array(leave_one_out_true_classes) == 1
        loo_ntgt_mask = np.array(leave_one_out_true_classes) == 0
        glob_tgt_mask = np.array(c.learning_buffor_classes) == 1
        glob_ntgt_mask = np.array(c.learning_buffor_classes) == 0
        glob_tgt = np.array(c.learning_buffor_features)[glob_tgt_mask]
        glob_ntgt = np.array(c.learning_buffor_features)[glob_ntgt_mask]

        if l_test:
            fig = pb.figure(figsize=(15, 10))
            pb.plot(l_test.transform(glob_tgt).mean(axis=0), label="Target features mean")
            pb.plot(l_test.transform(glob_ntgt).mean(axis=0), label="NONTarget features mean")
            pb.legend()
            pb.title(str(l))
            fig.tight_layout()
            fig.savefig(os.path.join(imagedir, str(fig_nr).rjust(3, '0') + "_transformed_features.png"))

        fig = pb.figure(figsize=(15, 10))
        pb.hist([leave_one_out_predictions[loo_ntgt_mask],
                 leave_one_out_predictions[loo_tgt_mask]],
                label=['nontarget N={}'.format(np.sum(loo_ntgt_mask)), 'target N={}'.format(np.sum(loo_tgt_mask))],
                normed=True)
        pb.legend()
        pb.title(str(l))
        fig.tight_layout()
        fig.savefig(os.path.join(imagedir, str(fig_nr).rjust(3, '0') + "_loo_predictions.png"))

        fig = pb.figure(figsize=(15, 10))
        pb.plot(loo_roc_fpr, loo_roc_tpr)
        pb.xlabel("FPR")
        pb.ylabel("TPR")
        pb.title("ROC, AUC = {:0.3}\n{}".format(loo_auc, str(l)))
        fig.tight_layout()
        fig.savefig(os.path.join(imagedir, str(fig_nr).rjust(3, '0') + "_loo_ROC.png"))

        accuracy, tresholds = calc_accuracy(leave_one_out_predictions, leave_one_out_true_classes)

        fig = pb.figure(figsize=(15, 10))
        pb.title(str(l) + '\n max ACC: {}'.format(max(accuracy)))
        pb.xlabel('Classifier threshold')
        pb.ylabel('Classifier accuracy at threshold')
        pb.plot(tresholds, accuracy)
        pb.ylim([0, 100])
        fig.tight_layout()
        fig.savefig(os.path.join(imagedir, str(fig_nr).rjust(3, '0') + "_loo_accuracies.png"))

    return loo_auc, str(l)


class ReprMixin:
    def __repr__(self):
        return self.__class__.__name__


class DeVectorizer(ReprMixin):
    n = 4

    def fit(self, X, y=None):
        pass
        return self

    def fit_transform(self, X, y=None):
        X = np.array(X)
        self.fit(X, y)
        return self.transform(X, y)

    def transform(self, X, y=None):
        X = np.array(X)
        X_epochs = np.reshape(X, (X.shape[0], self.n, int(X.shape[1] / self.n)))
        X_epochs_concat = np.concatenate([i.T for i in X_epochs])
        return X_epochs_concat


class Vectorizer(ReprMixin):
    n = 4
    length = 22

    def fit(self, X, y=None):
        pass
        return self

    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X, y)

    def transform(self, X, y=None):
        out = []
        for i in range(0, X.shape[0], self.length):
            out.append(X[i:i + self.length, :].T)
        out = np.array(out)
        return np.reshape(out, (out.shape[0], self.n * self.length))


class VectorToEpochs(ReprMixin):
    n = 4

    def fit(self, X, y=None):
        pass
        return self

    def fit_transform(self, X, y=None):
        X = np.array(X)
        self.fit(X, y)
        return self.transform(X, y)

    def transform(self, X, y=None):
        X = np.array(X)
        X_epochs = np.reshape(X, (X.shape[0], self.n, int(X.shape[1] / self.n)))
        return X_epochs


class EpochsToVector(ReprMixin):
    def fit(self, X, y=None):
        pass
        return self

    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X, y)

    def transform(self, X, y=None):
        return np.reshape(X, (X.shape[0], -1))


class EpochsToMNEEpochs(ReprMixin):
    def fit(self, X, y=None):
        self.events = np.array(y)
        return self

    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X, y)

    def transform(self, X, y=None):
        fs = 128.0
        channels = ["C{}".format(i) for i in range(X.shape[1])]
        chtypes = ['eeg' for i in channels]
        info = create_info(channels, fs, chtypes)

        if y is not None:
            events = self.events
            events_array = np.zeros((X.shape[0], 3), dtype=int)
            events_array[:, 0] = np.arange(0, X.shape[0], dtype=int)
            events_array[:, 2] = events

            epochs = EpochsArray(X, info, events=events_array, verbose=False)
        else:
            epochs = EpochsArray(X, info, verbose=False)
        return epochs


class MNEEpochsToEpochs(ReprMixin):
    def fit(self, X, y=None):
        pass
        return self

    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X, y)

    def transform(self, X, y=None):
        return X.get_data()


lda = LinearDiscriminantAnalysis(solver='lsqr', shrinkage='auto')
rfecv = RFECV(lda, scoring='roc_auc', cv=2)
pipeline = Pipeline([('rfecv', rfecv), ('lda', lda)])

lda2 = LinearDiscriminantAnalysis(solver='lsqr', shrinkage='auto')
ica = FastICA()
pipeline2 = Pipeline([('DeVectorizer', DeVectorizer()), ('ica', ica), ('Vectorizer', Vectorizer()), ('lda2', lda2)])
pipeline2_test = Pipeline(
    [('DeVectorizer', DeVectorizer()), ('ica', ica), ('Vectorizer', Vectorizer()), ('lda2', None)])

lda3 = LinearDiscriminantAnalysis(solver='lsqr', shrinkage='auto')
csp = CSP(reg='oas', transform_into='csp_space')
pipeline3 = Pipeline(
    [('DeVectorizer', VectorToEpochs()), ('csp', csp), ('Vectorizer', EpochsToVector()), ('lda2', lda3)])
pipeline3_test = Pipeline(
    [('DeVectorizer', VectorToEpochs()), ('csp', csp), ('Vectorizer', EpochsToVector()), ('lda2', None)])

lda4 = LinearDiscriminantAnalysis(solver='lsqr', shrinkage='auto')
xdawn = Xdawn(correct_overlap=False, n_components=4, reg='oas')
pipeline4 = Pipeline([('DeVectorizer', VectorToEpochs()),
                      ('EpochstoMNE', EpochsToMNEEpochs()),
                      ('xdawn', xdawn),
                      ('Vectorizer', EpochsToVector()),
                      ('lda', lda4)])
pipeline4_test = Pipeline([('DeVectorizer', VectorToEpochs()),
                           ('EpochstoMNE', EpochsToMNEEpochs()),
                           ('xdawn', xdawn),
                           ('Vectorizer', EpochsToVector()),
                           ('lda', None)])

classifiers = [
    LinearDiscriminantAnalysis(solver='eigen', shrinkage='auto'),
    LinearDiscriminantAnalysis(solver='lsqr', shrinkage='auto'),

    LinearDiscriminantAnalysis(solver='svd'),
    LinearDiscriminantAnalysis(solver='lsqr'),

    KNeighborsClassifier(3),

    QuadraticDiscriminantAnalysis(),

    DecisionTreeClassifier(max_depth=5),
    RandomForestClassifier(max_depth=5, n_estimators=10, max_features=1),
    AdaBoostClassifier(),
    SVC(kernel="linear", C=0.025, probability=True),
    SVC(gamma=2, C=1, probability=True),
    (pipeline2, pipeline2_test),
    (pipeline3, pipeline3_test),
    (pipeline4, pipeline4_test),
    pipeline
]
classifier_auc = []
for nr, classifier in enumerate(classifiers):

    if isinstance(classifier, tuple):
        print("Testing", classifier[0])
        classifier_auc.append((test_classifier(classifier[0], l_test=classifier[1], fig_nr=nr), nr))
    else:
        print("Testing", classifier)
        classifier_auc.append((test_classifier(classifier, fig_nr=nr), nr))

for classifier_auc_s in sorted(classifier_auc, key=lambda x: x[0][0]):
    print()
    print("Leave One Out crossvaldation AUC:", classifier_auc_s[0][0], "\nPipeline nr{}:\n".format(classifier_auc_s[1]),
          classifier_auc_s[0][1])

tgt_mask = np.array(c.learning_buffor_classes) == 1
ntgt_mask = np.array(c.learning_buffor_classes) == 0
tgt = np.array(c.learning_buffor_features)[tgt_mask]
ntgt = np.array(c.learning_buffor_features)[ntgt_mask]
fig = pb.figure(figsize=(15, 10))

pb.bar(range(tgt.shape[1]), pipeline.named_steps['rfecv'].support_ * 100, label='mask', alpha=0.1)
pb.bar(range(tgt.shape[1]), pipeline.named_steps['rfecv'].ranking_, label='rank', alpha=0.1)
pb.plot(tgt.mean(axis=0), label='target')
pb.plot(ntgt.mean(axis=0), label='nontarget')
pb.legend()
fig.savefig(os.path.join(imagedir, 'RFECV.png'))
# pb.show()
