#!/usr/bin/env python3
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

import os
import sys
import time

import requests

from devops import devpi

GITLAB_ADRESS = os.environ.get('GITLAB_ADRESS')
API_ACCESS_TOKEN = os.environ.get('API_ACCESS_TOKEN')
DEPENDANT_PROJECTS_IDS = os.environ.get('DEPENDANT_PROJECTS_IDS')
DEPENDANT_PROJECTS_TRIGGERS = os.environ.get('DEPENDANT_PROJECTS_TRIGGERS')

# own variables
CI_PROJECT_ID = os.environ.get('CI_PROJECT_ID')
CI_PIPELINE_ID = os.environ.get('CI_PIPELINE_ID')
CI_COMMIT_REF_NAME = os.environ.get('CI_COMMIT_REF_NAME')

# DEFAULTS
DEFAULTREF = 'development'
CHECK_INTERVAL = 2
OBCI_DEB_BUILD_JOB_NAME = 'builddebobci'


class GitLabUrlsFactory:
    def api_get_jobs_eldest_first(self):
        return ('{}/api/v4/projects/{}/pipelines/{}/jobs?per_page=100'
                .format(GITLAB_ADRESS, CI_PROJECT_ID, CI_PIPELINE_ID))

    def api_trigger_pipeline(self, dependant_id):
        return ('{}/api/v4/projects/{}/trigger/pipeline'
                .format(GITLAB_ADRESS, dependant_id))

    def api_get_pipeline(self, project_id, pipeline_id):
        return ('{}/api/v4/projects/{}/pipelines/{}'
                .format(GITLAB_ADRESS, project_id, pipeline_id))

    def browser_get_pipeline(self, project_id, pipeline_id):
        return '{}/pipelines/{}'.format(self.browser_get_project(project_id),
                                        pipeline_id)

    def browser_get_project(self, project_id):
        api_get_url = self.api_get_project(project_id)
        response_json = _send_gitlab_api_request(api_get_url).json()
        return response_json['web_url']

    def api_get_project(self, project_id):
        return '{}/api/v4/projects/{}'.format(GITLAB_ADRESS, project_id)


gitlab_urls = GitLabUrlsFactory()


def run():
    assert GITLAB_ADRESS
    assert API_ACCESS_TOKEN
    assert DEPENDANT_PROJECTS_IDS
    assert DEPENDANT_PROJECTS_TRIGGERS
    assert CI_PROJECT_ID
    assert CI_PIPELINE_ID
    assert CI_COMMIT_REF_NAME
    projects_pipelines = _trigger_children_rebuilding()
    _wait_for_results(projects_pipelines)


def _trigger_children_rebuilding():
    projects_pipelines = {}
    build_debs_job_id = _get_latest_successful_deb_build_job_id()
    projects_ids = DEPENDANT_PROJECTS_IDS.strip(';').split(';')
    trigger_tokens = DEPENDANT_PROJECTS_TRIGGERS.strip(';').split(';')
    for project_id, trigger_token in zip(projects_ids, trigger_tokens):
        pipeline_id = _trigger_pipeline(project_id, build_debs_job_id, trigger_token,
                                        CI_COMMIT_REF_NAME)
        projects_pipelines[project_id] = pipeline_id
    return projects_pipelines


def _get_latest_successful_deb_build_job_id():
    response = _send_gitlab_api_request(gitlab_urls.api_get_jobs_eldest_first())
    jobs = response.json()
    jobs_latest_first = reversed(jobs)
    for job in jobs_latest_first:
        if job['name'] == OBCI_DEB_BUILD_JOB_NAME and job['status'] == 'success':
            return job['id']


def _trigger_pipeline(project_id, build_debs_job_id, trigger_token, refname):
    data = {'token': trigger_token,
            'ref': refname,
            r'variables[ARTIFACTS_JOB_ID]': build_debs_job_id,
            r'variables[ARTIFACTS_PROJECT_ID]': CI_PROJECT_ID,
            r'variables[TRIGGERED_AS_DEPENDANT]': 'true',
            r'variables[TRIGGERING_REF_NAME]': refname,
            r'variables[DEVPI_INDEX_NAME]': devpi.temp_name(devpi.current_index()['name']),
            }
    print('Triggering pipeline of project {}, giving build job id {}, on ref {} {}'
          .format(project_id, build_debs_job_id, refname, trigger_token))
    url = gitlab_urls.api_trigger_pipeline(project_id)
    response = requests.post(url, data=data)
    failed_to_find_ref = response.status_code == 400
    if failed_to_find_ref:
        if refname == DEFAULTREF:
            print("\n\n\nWARNING! BAD REQUEST ON DEFAULT REF!")
            print("Refname: {}, response.text: {}, response.status_code: {}"
                  .format(refname, response.text, response.status_code))
            sys.exit(1)
        else:
            return _trigger_pipeline(project_id, build_debs_job_id, trigger_token,
                                     DEFAULTREF)
    else:
        assert response.status_code < 400, (url, response.status_code, response.text)
        return response.json()['id']


def _wait_for_results(initial_projects_pipelines: dict):
    print("Waiting for pipelines: {}".format(initial_projects_pipelines))
    projects_pipelines = initial_projects_pipelines.copy()
    while projects_pipelines:
        for project_id, pipeline_id in list(projects_pipelines.items()):
            status = _get_pipeline_status(project_id, pipeline_id)
            pipeline_url = gitlab_urls.browser_get_pipeline(project_id, pipeline_id)
            print('Pipeline: {} Status: {}'.format(pipeline_url, status))
            if status == 'success':
                del projects_pipelines[project_id]
            elif status in ['failed', 'canceled']:
                print("\n\n\nWARNING! PIPELINE OF DEPENDANT PROJECT FAILED!")
                print('Pipeline: {} Status: {}'.format(pipeline_url, status))
                sys.exit(1)
            time.sleep(CHECK_INTERVAL)


def _get_pipeline_status(project_id, pipeline_id):
    url = gitlab_urls.api_get_pipeline(project_id, pipeline_id)
    response_json = _send_gitlab_api_request(url).json()
    return response_json['status']


def _send_gitlab_api_request(url):
    response = requests.get(url, headers={'PRIVATE-TOKEN': API_ACCESS_TOKEN})
    assert response.status_code < 400, (url, response.status_code, response.text)
    return response


if __name__ == '__main__':
    run()
