#!/usr/bin/env python3
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

import abc
import argparse
import os.path
from typing import List  # noqa

import numpy as np

import braintech.obci.core.utils.signal_generators as signal_generators
from braintech.obci.signal_processing.signal.info_file_proxy import InfoFileWriteProxy

DEFAULT_DTYPE = np.float32

seconds_per_unit = {'s': 1, 'm': 60, 'h': 3600, 'd': 86400, 'w': 604800}


def convert_to_seconds(s: str) -> int:
    """
    Convert string such as '123s', '123d' or '123w' to number of seconds.
    Available modifiers: ``s`` - seconds, ``m`` - minutes, ``h`` - hours,
    ``d`` - days, ``w`` - weeks.
    """
    return int(s[:-1]) * seconds_per_unit[s[-1]]


class SignalGenerator(metaclass=abc.ABCMeta):
    def __init__(self, dtype=DEFAULT_DTYPE):
        super().__init__()
        self.dtype = dtype

    @abc.abstractmethod
    def get_samples(self, n: int) -> np.ndarray:
        pass


class ConstSignalGenerator(SignalGenerator):
    def __init__(self, value: float, dtype=DEFAULT_DTYPE):
        super().__init__(dtype=dtype)
        self.value = float(value)

    def get_samples(self, n: int) -> np.ndarray:
        return np.full((n,), self.value, dtype=self.dtype)


class ZeroSignalGenerator(ConstSignalGenerator):
    def __init__(self, dtype=DEFAULT_DTYPE):
        super().__init__(value=0.0, dtype=dtype)


class OnesSignalGenerator(ConstSignalGenerator):
    def __init__(self, dtype=DEFAULT_DTYPE):
        super().__init__(value=1.0, dtype=dtype)


class RandomSignalGenerator(SignalGenerator):
    def __init__(self, scale: float=1.0, offset: float=0.0, dtype=DEFAULT_DTYPE):
        super().__init__(dtype=dtype)
        self.scale = float(scale)
        self.offset = float(offset)

    def get_samples(self, n: int) -> np.ndarray:
        return self.scale * np.array(np.random.rand(n), dtype=self.dtype) + self.offset


class SawSignalGenerator(SignalGenerator):
    def __init__(self, max_value: int=100, dtype=DEFAULT_DTYPE):
        super().__init__(dtype=dtype)
        self.generator = signal_generators.saw_generator(max_value)

    def get_samples(self, n: int) -> np.ndarray:
        return np.fromiter(self.generator, dtype=self.dtype, count=n)


def parse_length(length_str: str, sampling_frequency: int):
    try:
        return int(length_str)
    except ValueError:
        return convert_to_seconds(length_str) * sampling_frequency


def sizeof_fmt(num, suffix='B'):
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f %s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f %s%s" % (num, 'Yi', suffix)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Generate signal file.')
    parser.add_argument('-l', '--length', help='length is samples or in [s,m,h] (1234, 2h, 30m, 30s)', required=True)
    parser.add_argument('-f', '--sampling-frequency', help='sampling frequency (default: 512 Hz)', default=512)
    parser.add_argument('-c', '--channels', default=32, help='number of channels (default: 32)')
    parser.add_argument('file_name', help="output file name (without '.raw' extension)")
    args = parser.parse_args()

    file_name = args.file_name

    sampling_frequency = int(args.sampling_frequency)
    total_samples = parse_length(args.length, sampling_frequency)

    channel_generators = []  # type: List[SignalGenerator]
    for _ in range(int(args.channels)):
        channel_generators.append(RandomSignalGenerator())

    channel_names = ['Ch{}'.format(i) for i in range(len(channel_generators))]

    if len(channel_generators) >= 2:
        channel_generators[-1] = SawSignalGenerator()
        channel_names[-1] = 'Saw'

    if len(channel_generators) >= 4:
        channel_generators[0] = ZeroSignalGenerator()
        channel_generators[1] = OnesSignalGenerator()
        channel_names[0] = 'Zero'
        channel_names[1] = 'One'

    if DEFAULT_DTYPE == np.float32:
        final_file_size = 4 * len(channel_generators) * total_samples
    elif DEFAULT_DTYPE == np.float64:
        final_file_size = 8 * len(channel_generators) * total_samples
    else:
        raise Exception('Unsupported dtype.')

    print('Signal length (samples): {}'.format(total_samples))
    print('Signal length (seconds): {} seconds'.format(total_samples / sampling_frequency))
    print('Signal length (hours): {:.2f} hours'.format((total_samples / sampling_frequency) / (60 * 60)))
    print('Sampling frequency: {} Hz'.format(sampling_frequency))
    print('Number of channels: {}'.format(len(channel_generators)))
    print('Output file size: {}'.format(sizeof_fmt(final_file_size)))

    info_writer = InfoFileWriteProxy(file_name + '.xml')
    info_writer.set_attributes({
        'number_of_samples': str(total_samples),
        'sampling_frequency': str(sampling_frequency),
        'number_of_channels': str(len(channel_generators)),
        'channels_names': channel_names,
        'sample_type': 'FLOAT' if DEFAULT_DTYPE == np.float32 else 'DOUBLE',
        'byte_order': 'LITTLE_ENDIAN',
        'blocks_per_page': '1',
        'page_size': '20.0',
        'export_file_name': file_name,
        'file': os.path.basename(file_name)
    })
    info_writer.save()

    fragment_samples = 100000
    fragments_number = total_samples // fragment_samples
    fragment = np.zeros((fragment_samples, len(channel_generators)), dtype=DEFAULT_DTYPE)

    if total_samples % fragment_samples != 0:
        fragments_number += 1

    with open(file_name + '.raw', 'w+b') as f:
        for _ in range(fragments_number):
            for i in range(len(channel_generators)):
                samples = channel_generators[i].get_samples(fragment_samples)
                fragment[:, i] = samples
            fragment.tofile(f)
        f.truncate(final_file_size)
