#!/usr/bin/env python3
# Copyright (c) 2016-2018 Braintech Sp. z o.o. [Ltd.] <http://www.braintech.pl>
# All rights reserved.

import sys
import os.path
import argparse
from math import isclose
from typing import Tuple, Optional

import numpy as np

from braintech.obci.signal_processing.signal.read_info_source import FileInfoSource

TOLERANCE = 1e-6
BYTES_PER_SAMPLE = 4  # float32 size in bytes
FRAGMENT_SIZE = 100


def print_bad_sample(global_index: int,
                     description: str
                     ) -> None:
    print('{} - {}'.format(global_index, description), file=sys.stderr)


def show_bad_sample(global_index: int,
                    fragment: np.ndarray,
                    fragment_offset: int,
                    description: str
                    ) -> None:
    import matplotlib.pyplot as plt
    plt.plot(fragment)
    plt.title('Error at {} ({})'.format(global_index, description))
    plt.xlabel('samples')
    plt.ylabel('values')
    plt.axvline(fragment_offset, color='r')
    plt.show()


def verify_data_file(  # noqa: C901
        filename: str,
        number_of_samples: int,
        number_of_channels: int,
        saw_channel_index: int,
        plotting_enabled: bool,
        all_errors: bool
        ) -> int:
    data = np.memmap(filename,
                     mode='r',
                     dtype=np.float32,
                     shape=(number_of_samples, number_of_channels))

    errors_count = 0

    def get_fragment(index: int) -> Tuple[np.ndarray, int]:
        index_from = index - FRAGMENT_SIZE
        if index_from < 0:
            index_from = 0
            fragment_offset = index
        else:
            fragment_offset = FRAGMENT_SIZE
        index_to = index + FRAGMENT_SIZE
        if index_to > data.shape[0]:
            index_to = data.shape[0]
        return data[index_from:index_to, saw_channel_index], fragment_offset

    def bad_sample(index: int, description: str):
        nonlocal errors_count
        errors_count += 1
        print_bad_sample(i, description)
        if plotting_enabled:
            fragment, fragment_offset = get_fragment(index)
            show_bad_sample(i, fragment, fragment_offset, description)

    a = None
    min_val = None
    max_val = None
    last_max_idx = None
    maximums_period = None
    for i in range(1, number_of_samples):
        curr, prev = data[i, saw_channel_index], data[i - 1, saw_channel_index]
        delta = curr - prev
        if delta > 0:
            if a is None:
                a = delta
            if not isclose(delta, a, abs_tol=TOLERANCE):
                bad_sample(i, 'delta != a ({} != {})'.format(delta, a))
        elif delta < 0:
            if min_val is None:
                min_val = curr
            if max_val is None:
                max_val = prev
            if last_max_idx is not None:
                period = i - last_max_idx
                if maximums_period is None:
                    maximums_period = period
                elif period != maximums_period:
                        bad_sample(i, 'invalid period (current: {}, should be: {}, lost samples: {})'
                                   .format(period, maximums_period, maximums_period - period))
            last_max_idx = i
            if not isclose(prev, max_val, abs_tol=TOLERANCE):
                bad_sample(i, 'prev != max_val, ({} != {})'.format(prev, max_val))
            if not isclose(curr, min_val, abs_tol=TOLERANCE):
                bad_sample(i, 'curr != min_val ({} != {})'.format(curr, min_val))
        else:
            bad_sample(i, 'delta == 0')

        if not all_errors and errors_count > 10:
            raise Exception('Too many errors! Interrupting file analysis.', file=sys.stderr)

    return errors_count


def autodetect_signal_params(xml_filename: str) -> Tuple[int, int, Optional[int]]:
    if not os.path.isfile(xml_filename):
        raise Exception("XML file not found: '{}'".format(xml_filename))
    info = FileInfoSource(xml_filename)
    number_of_samples = int(info.get_param('number_of_samples'))
    channel_names = info.get_param('channels_names')
    number_of_channels = len(channel_names)
    try:
        saw_channel_index = channel_names.index('Saw')
    except ValueError:
        saw_channel_index = None
    return number_of_samples, number_of_channels, saw_channel_index


def verify(filename: str, args) -> int:  # noqa: C901
    plotting_enabled = args.plot

    if filename.endswith('.raw'):
        raw_filename = filename
        filename = filename[:-4]
    else:
        raw_filename = filename + '.raw'
    xml_filename = filename + '.xml'

    if not os.path.isfile(raw_filename):
        raise Exception("'{}' is not proper file name.".format(raw_filename))

    print('Checking: {}'.format(raw_filename))

    file_size = os.path.getsize(raw_filename)

    if file_size == 0:
        raise Exception('Input file is empty (file size is zero bytes).')

    if args.number_of_channels is None:
        # autodetect mode
        try:
            number_of_samples, number_of_channels, saw_channel_index = autodetect_signal_params(xml_filename)
        except Exception as ex:
            raise Exception("Error during signal parameters autodetection: '{}'.\n"
                            "You still can manually specify required parameters (see documentation for "
                            "'number-of-channels' and 'saw-channel-index' command line parameters).".format(ex))

        if number_of_channels <= 0:
            raise Exception("Invalid 'number-of-channels' detected: {}.".format(number_of_channels))

        if args.saw_channel_index is not None:
            if args.saw_channel_index < 0 or args.saw_channel_index >= number_of_channels:
                raise Exception("Specified 'saw-channel-index' parameter is invalid (must be {} <= n <= {})"
                                .format(0, number_of_channels - 1))
            saw_channel_index = args.saw_channel_index

        if saw_channel_index is None:
            raise Exception("Cannot find 'Saw' channel. You can manually specify 'saw-channel-index' parameter.",
                            file=sys.stderr)

        if saw_channel_index < 0 or saw_channel_index >= number_of_channels:
            raise Exception("Autodetected 'saw-channel-index' parameter is invalid (must be {} <= n <= {})"
                            .format(0, number_of_channels - 1))
    else:
        if args.saw_channel_index is None:
            raise Exception("'saw-channel-index' parameter is required when specifying number-of-channels.")
        saw_channel_index = args.saw_channel_index

        if args.number_of_channels <= 0:
            raise Exception("Invalid 'number-of-channels' specified: {}.".format(args.number_of_channels))
        number_of_channels = args.number_of_channels

        if saw_channel_index < 0 or saw_channel_index >= number_of_channels:
            raise Exception("'saw-channel-index' parameter is invalid (must be {} <= n <= {})"
                            .format(0, number_of_channels - 1))

        number_of_samples = (file_size // BYTES_PER_SAMPLE) // args.number_of_channels

    if file_size != BYTES_PER_SAMPLE * number_of_channels * number_of_samples:
        print('WARNING: File is corrupted (file_size != 4 * number_of_channels * number_of_samples).\n'
              '         Verifier will still try to verify this file assuming number_of_samples = {}'
              .format(number_of_samples), file=sys.stderr)

    return verify_data_file(raw_filename,
                            number_of_samples,
                            number_of_channels,
                            saw_channel_index,
                            plotting_enabled,
                            args.all_errors)


def main() -> None:
    parser = argparse.ArgumentParser(description='Saw verifier, prints indexes of corrupted samples',)
    parser.add_argument('filename', nargs='+', help='Input filename(s), without .xml or .raw')
    parser.add_argument('-s', '--saw-channel-index', default=None, type=int,
                        help='Index of channel with Saw')
    parser.add_argument('-n', '--number-of-channels', default=None, type=int,
                        help='If no xml metadata is available you can manually provide number of channels'
                             ', saw-channel-index must be provided')
    parser.add_argument('-p', '--plot', action='store_true', help='Show signal fragments with error.')
    parser.add_argument('-a', '--all-errors', action='store_true',
                        help='Show all errors (by default only first 10 are shown).')

    args = parser.parse_args()

    all_ok = True
    for filename in args.filename:
        try:
            errors_count = verify(filename, args)
            if errors_count == 0:
                print('No errors found.\n')
            else:
                print('Number of errors in file: {}\n'.format(errors_count))
        except Exception as ex:
            all_ok = False
            print('Error: {}\n'.format(ex))
    sys.exit(0 if all_ok else 1)


if __name__ == '__main__':
    main()
